package com.monica.interfaces;

import com.monica.util.configuration.ConfigurationException;

import java.io.InputStream;
import java.io.OutputStream;

/** This interface is responsible for the configuration of objects in the
    TM layer hierarchy. Configuration sources can be of any kind, 
    the main thing is that the classes can get configuration data 
    of common way.
*/    

public interface Configuration extends Cloneable {

  /** Returns the internal name of the Configuration.
      @return String: the internal name of the Configuration
  */      
  public String getConfigurationName();
  
  /** Returns the internal version of the Configuration.
      @return String: the internal version of the Configuration
  */      
  public String getConfigurationVersion();
  
  /** Retuns an int configuration value for the specified configuration id.
      @param String id: the required configuration id
      @return int: the configuration int matching the configuration id
      @exception ConfigurationException: if the id is invalid or the id doesn't represent an 
        int value
  */	
  public int getInt(String id) throws ConfigurationException;
  
  /** Returns a boolean configuraton value for the specified configuration id.
      @param String id: the required configuration id
`      @return boolean: the configuration boolean matching the configuration id
      @exception ConfigurationException: if the id is invalid or the id doesn't represent a
        boolean value
  */	
  public boolean getBoolean(String id) throws ConfigurationException;

  /** Returns a String configuraton value for the specified configuration id.
      @param String id: the required configuration id
      @return String: the configuration String matching the configuration id
      @exception ConfigurationException: if the id is invalid or the id doesn't represent a
        String value
  */	
  public String getString(String id) throws ConfigurationException;
  
  /** Returns an int configuraton value for the specified configuration id.
      Should call the getInt(String) method of this interface and if there
      was an exception, returns the specified default value.
      @param String id: the required configuration id
      @param int def: the specified default value
      @return int: the configuration int matching the configuration id
        or if there was an exception then the default int
  */	
  public int getInt(String id, int def);

  /** Returns a boolean configuraton value for the specified configuration id.
      Should call the getBoolean(String) method of this interface and if there
      was an exception, returns the specified default value.
      @param String id: the required configuration id
      @param boolean def: the specified default value
      @return boolean: the configuration boolean matching the configuration id
        or if there was an exception then the default boolean
  */	
  public boolean getBoolean(String id, boolean def);

  /** Returns a String configuraton value for the specified configuration id.
      Should call the getString(String) method of this interface and if there
      was an exception, returns the specified default value.
      @param String id: the required configuration id
      @param boolean String: the specified default value
      @return boolean: the configuration String matching the configuration id
        or if there was an exception then the default String
  */	
  public String getString(String id, String def);

  /** Useable if String possible values in the configuration source 
      should be mapped into ints as configuration values.
      If there is a problem retrieving the int returned, the default
      value will be returned.
      @param String id: the required configuration id
      @param String[] possibilities: possible Strings acceptable in the
        configuration source
      @param int[] values: int values matching the possible Strings
      @param int def: the dafault int that should be returned in case of
        errors
      @return int: the int value matching the found possibility String of
        the def int
  */  
  public int getIntFromStrings(String id, String[] possibilities, int[] values, int def);
  
  /** Defines for the given ID an InputStream stream, whose contents can be retrieved unlimited times from
      clients of this class. This class reads the specied stream to the end and creates a String from 
      the read contents. When using the getStream, a new StringInputStream initialized with this String
      will be returned.
      @param String ID: the ID of the InputStream
      @param InputStream stream: the stream whose contents can be retrieved
  */      
  public void setStream(String ID, InputStream stream);
  
  /** Returns the InputStream identified with the specified ID.
      @param String ID: the ID of the returned stream
      @return InputStream: the InputStream with the specified ID
  */      
  public InputStream getStream(String ID) throws ConfigurationException;
  
  /** Returns a Configuration object that can be created
   *  from the stream with the specified ID. Uses getStream(ID)
   *  to retrieve the stream.
   *  @param String ID: the ID for the stream
   *  @exception ConfigurationException: throws exception if an error occures
   */
  public Configuration getConfiguration(String ID) throws ConfigurationException;
  
  /** Exports this Configuration to the specified OutputStream.
   *  @param OuputStream os: the specified OuputStream to output this Configuration
   *  @return boolean: returns whether the save process has been successfully finished
   */
  public boolean save(OutputStream os);
  
  /** Sets a key with the specified ID to the specified value.
   *  @param String ID: the specified key ID
   *  @param String value: specified value
   */
  public void set(String ID, String value);
  
  /** Removes the specified key ID from this Configuration.
   *  @param String ID: the specified key ID
   */
  public void remove(String ID);

  /** Clones the current Configuration.
   *  @return Object: a clone of the current object
   */
  public Object clone();
  
}
