package com.monica.interfaces;

import com.monica.interfaces.*;

/** Implements the destination part of the write functionality in the
    TM layer hierarchy. Writeable is always the destination of data
    from the Writer. Writing in the TM layer hierarchy is synchronous.
    If the Writer is willing to write, it calls one of the write
    methods of the Writeable. The write will return after the data is 
    submitted to bottom layers. Data can be buffered by the Writeable.
*/    
public interface Writeable {

  /** Returns the internal name of the Writeable.
      @return String: the internal name of the Writeable
  */      
  public String getWriteableName();
  
  /** Returns the internal version of the Writeable.
      @return String: the internal version of the Writeable
  */      
  public String getWriteableVersion();
  
  /** Writes one single char. Writing is synchronous and can be
      buffered.
      @param char c: the char to be written
      @return boolean: wether the writing has been successful
  */      
  public boolean write(char c);
  
  /** Writes a String. Writing is synchronous and can be buffered.
      @param String s: the String to be written
      @return boolean: wether the writing has been successful
  */      
  public boolean write(String s);
  
  /** Writes a char array. Writing is synchronous and can be buffered.
      @param char[] c: the char array to be written
      @return boolean: wether the writing has been successful
  */   
  public boolean write(char[] c);
  
  /** Flushes the contents of the write buffer to bottom layers.
  */
  public void flush();
  
  /** Returns the error message of the last occured error. After
      returning it, the error message will be deleted and cannot
      be queried again.
      @return String: the last error message
  */  
  public String getLastWriteableError();
  
  /** Returns wether the Writeable is connected to its target.
      @return boolean: wether the Writeable is connected to
        its target
  */
  public boolean isWriteableConnected();
  
}    