package com.monica.javacom.ssh2;

import com.monica.javacom.ssh2.*;
import com.monica.interfaces.*;
import com.monica.javacom.interfaces.*;
import com.monica.javacom.ssh2.interfaces.*;
import com.monica.javacom.ssh2.userauthentication.*;
import com.monica.javacom.ssh2.connection.*;

import com.monica.util.debug.*;

import java.util.*;

public class SSH2Facade implements Configureable {
  
  // ----------------------------------------------------------------------
  // Public constants
  // ----------------------------------------------------------------------
    
  // Feature identifiers
  public static final int OPT_KEX_ALGORITHM=0;
  public static final int OPT_SERVER_HOSTKEY_ALGORITHM=1;
  public static final int OPT_CIPHER_ALGORITHM_CTS=2;
  public static final int OPT_CIPHER_ALGORITHM_STC=3;
  public static final int OPT_MAC_ALGORITHM_CTS=4;
  public static final int OPT_MAC_ALGORITHM_STC=5;
  public static final int OPT_COMPRESSION_ALGORITHM_CTS=6;
  public static final int OPT_COMPRESSION_ALGORITHM_STC=7;
  public static final int OPT_USERAUTHENTICATION_LAYER=8;
  public static final int OPT_CONNECTION_LAYER=9;
  
  public static final int OPT_COUNT=10;
  public static final boolean[] OPT_FEATURES={true, true, true, true, true, true, true, true, true, true};
  
  // Return codes
  public static final int REG_OK=0;
  public static final int REG_ALREADY_PRESENT=1;
  public static final int REG_NO_OPT=2;
  public static final int REG_NOT_REGISTERED=3;

  // ----------------------------------------------------------------------
  // Internal constants
  // ----------------------------------------------------------------------
  
  private static final String DP="SSH2Facade";     

  private static final String ERR_LIST_NOT_MAINTAINED="List of such feature is not maintained.";
  private static final String ERR_KEY_NOT_PRESENT="Such key is not present in the maintained features list. ->";
  private static final String ERR_NOT_REGISTERED="Such protocol is not registered. ->";
  
  // ----------------------------------------------------------------------
  // Variables
  // ----------------------------------------------------------------------
  
  private Hashtable features[];
  private Configuration configuration;
  private SSH2IO io;    
  private Vector clients, servers;
  private SSH2ConnectionLayerFactory connFact;

  public SSH2Facade() {
    io=new SSH2IO();
    features=new Hashtable[OPT_COUNT];
    for (int i=0; i<OPT_COUNT; i++) {
      features[i]=new Hashtable();
    }  
    clients=new Vector();
    servers=new Vector();
  }    
  
  public void setConfiguration(Configuration configuration) {
    this.configuration=configuration;
  }
  
  public Configuration getConfiguration() {
    return configuration;
  }        
    
  public synchronized int registerClassForFeature(int opt, String name, String className, Object param) throws Exception {
    if ((opt>=OPT_COUNT) || (opt<0) || (!OPT_FEATURES[opt])) {
      return REG_NO_OPT;
    }
    if (features[opt].get(name)!=null) {
      return REG_ALREADY_PRESENT;
    }
    switch (opt) {
      case OPT_USERAUTHENTICATION_LAYER :
        SSH2UserAuthenticationLayerFactory factUA=(SSH2UserAuthenticationLayerFactory)Class.forName(className).newInstance();
        factUA.setConfiguration(configuration);
        factUA.setFacade(this);
        registerUserAuthenticationLayer(name, factUA);
        break;
      case OPT_CONNECTION_LAYER :
        SSH2ConnectionLayerFactory factC=(SSH2ConnectionLayerFactory)Class.forName(className).newInstance();
        factC.setConfiguration(configuration);
        factC.setFacade(this);
        registerConnectionLayer(factC);
        break;
      default :  
        Object o[]=new Object[2];
        o[0]=className;
        o[1]=param;
        features[opt].put(name, o);
        break;
    }
    return REG_OK;
  }
  
  public synchronized String getFeatureList(int opt) throws Exception{
    if ((opt>=OPT_COUNT) || (opt<0)) {
      throw new Exception(ERR_LIST_NOT_MAINTAINED);      
    }
    return io.createList(features[opt].keys(), SSH2Constants.SEP_COMMA_NA);
  }
  
  public synchronized Object getFeature(int opt, String name) throws Exception {
    if ((opt>=OPT_COUNT) || (opt<0) || (!OPT_FEATURES[opt])) {
      throw new Exception(ERR_LIST_NOT_MAINTAINED);      
    }
    Object o=features[opt].get(name);
    if (o==null) {
      throw new Exception(ERR_KEY_NOT_PRESENT+name);
    }
    String className=(String)(((Object[])(o))[0]);
    return Class.forName(className).newInstance();
  }
  
  public synchronized Object getFeatureParam(int opt, String name) throws Exception {
    if ((opt>=OPT_COUNT) || (opt<0) || (!OPT_FEATURES[opt])) {
      throw new Exception(ERR_LIST_NOT_MAINTAINED);      
    }
    Object o=features[opt].get(name);
    if (o==null) {
      throw new Exception(ERR_KEY_NOT_PRESENT);
    }
    Object param=((Object[])(o))[1];
    return param;
  }    
  
  public synchronized int registerUserAuthenticationLayer(String name, SSH2UserAuthenticationLayerFactory fact) throws Exception {
    if (features[OPT_USERAUTHENTICATION_LAYER].get(name)!=null) {
      return REG_ALREADY_PRESENT;
    }
    features[OPT_USERAUTHENTICATION_LAYER].put(name, fact);
    fact.setFacade(this);
    return REG_OK;
  }

  public synchronized SSH2UserAuthenticationLayer getUserAuthenticationLayer(String name, Object param) throws Exception {
    SSH2UserAuthenticationLayerFactory f;
    Object o=features[OPT_USERAUTHENTICATION_LAYER].get(name);
    if (o==null) {
      throw new Exception(ERR_NOT_REGISTERED+name);
    }
    f=(SSH2UserAuthenticationLayerFactory)o;
    if (param==null) {
      return f.getUserAuthenticationLayer();
    }
    else {
      return f.getUserAuthenticationLayer(param);
    }
  }
  
  public synchronized int registerConnectionLayer(SSH2ConnectionLayerFactory fact) throws Exception {
    connFact=fact;  
    return REG_OK;
  }

  public synchronized SSH2ConnectionLayer getConnectionLayer(Object param, SSH2ChannelManager chManager) throws Exception {
    SSH2ConnectionLayer cl;  
    if (connFact!=null) {
      if (param==null) {
        cl=connFact.getConnectionLayer();
      }
      else {
        cl=connFact.getConnectionLayer(param);
      }
      cl.setChannelManager(chManager);
      return cl;
    }
    else {
      throw new Exception(ERR_NOT_REGISTERED);
    }
  }
    
  public synchronized SSH2ClientSide createClient(LayerBottomView topLevel) throws Exception {
    SSH2ClientSide c=new SSH2ClientSide(this, topLevel);
    c.setConfiguration(configuration);
    clients.addElement(c);
    return c;
  }
    
  public synchronized SSH2ServerSide createServer(SSH2ChannelManager chManager, SSH2ConfigurationManager coManager, Object cParam) throws Exception {
    SSH2ServerSide s=new SSH2ServerSide(this, chManager, coManager, cParam);
    s.setConfiguration(configuration);
    servers.add(s);
    return s;
  } 
  
  public synchronized void clientDone(SSH2ClientSide c) {
    clients.removeElement(c);
  }
  
  public synchronized void serverDone(SSH2ServerSide s) {
    servers.removeElement(s);
  }       
}

