package com.monica.javacom.ssh2;

import com.monica.util.*;

import java.math.*;
import java.util.*;

/** Class for providing functionality either field conversion to String or
    from String according to the SSH2 specification, or some aspects of
    the SSH2 functionality that could not be put into any other class.
*/    
public class SSH2IO {

  Converter converter;
  BigInteger zero;

  public SSH2IO() {
    converter=new Converter();
    zero=new BigInteger("0");
  }

  // *******************************************************************************************
  // SSH Package Write public functions
  // *******************************************************************************************

  /** Creates a one byte String according to the SSH2 (BYTE) specification from an int.
      @param int i: the int to be converted to String
      @return String: the result
  */      
  public String wByte(int i) {
    char c[]=new char[1];
    c[0]=(char)(i&0xFF);
    return new String(c);
  }

  /** Creates a one byte String according to the SSH2 (BOOLEAN) specification from a boolean.
      @param boolean b: the boolean to be converted to String
      @return String: the result
  */       
  public String wBoolean(boolean b) {
    char c[]=new char[1];
    if (b) {
      c[0]=(char)1;
      return new String(c);
    }
    else {
      c[0]=(char)0;
      return new String(c);
    }
  }

  /** Creates a four byte String according to the SSH2 (UINT32) specification from an int.
      @param int i: the int to be converted to String
      @return String: the result
  */      
  public String wUint32(int i) {
    char c[]=new char[4];
    c[0]=(char)((i&0xFF000000)>>>24);
    c[1]=(char)((i&0x00FF0000)>>16);
    c[2]=(char)((i&0x0000FF00)>>8);
    c[3]=(char) (i&0x000000FF);
    return new String(c);
  }

  /** Creates a four byte String from an int by writing the bytes in the low order way.
      @param int i: the int to be converted to String
      @return String: the result
  */      
  public String wLowOrderUint32(int i) {
    char c[]=new char[4];
    c[3]=(char)((i&0xFF000000)>>>24);
    c[2]=(char)((i&0x00FF0000)>>16);
    c[1]=(char)((i&0x0000FF00)>>8);
    c[0]=(char) (i&0x000000FF);
    return new String(c);
  }

  /** Creates an eight byte String according to the SSH2 (UINT64) specification from a long.
      @param long l: the long to be converted to String
      @return String: the result
  */      
  public String wUint64(long l) {
    char c[]=new char[8];

    long k, m;

    for (int i=0; i<8; i++) {
      k=l;
      m=l>>8;
      k-=(m<<8);
      c[7-i]=(char)k;
      l=m;
    }

    return new String(c);
  }

  /** Creates an arbitrary length String according to the SSH2 (STRING) specification from a String.
      @param String s: the String to be converted to String
      @return String: the result
  */      
  public String wString(String s) {
    return wUint32(s.length())+s;
  }

  /** Creates an arbitrary length String according to the SSH2 (MPINT) specification from a BigInteger.
      @param BigInteger bi: the BigInteger to be converted to String
      @return String: the result
  */      
  public String wMpint(BigInteger bi) {
    byte b[]=bi.toByteArray();
    String bs=converter.toString(b);
    while ((bs.length()>0) && (bs.charAt(0)==((char)0))) {
      bs=bs.substring(1);
    }
    if (bs.length()>0) {
      if (bi.min(zero)==zero) {
        if (((int)bs.charAt(0))>=128) {
          bs=(new String(""+(char)0)).concat(bs);
        }
      }
    }
    return (wUint32(bs.length())).concat(bs);
  }

  public String wMpintString(BigInteger bi) {
      String s=bi.toString(16);
      return fromHex(s);
  }
  
  private String fromHex(String f) {
      char s[]=f.toCharArray();
      char t[]=new char[s.length/2];
      int v;
      for (int i=0; i<t.length; i++) {          
          v=_fromHex(s[i*2]);
          t[i]=(char)v;
          t[i]<<=4;
          v=_fromHex(s[i*2+1]);
          t[i]|=(char)((v)&0x0f);
      }
      return new String(t);
  }
          
  private int _fromHex(int v) {
      if (v<65) {
          v-=48;
      }
      else {
          v-=87;
      }
      return v;
  }
  
    
  // *******************************************************************************************
  // SSH Package Read Public functions
  // *******************************************************************************************

  /** Creates a BigInteger from a String ancoded as specified in the SSH2 specification and
      beginning with an SSH2 BYTE.
      @param String s: the String containing the BYTE
      @return BigInteger: the BYTE as BigInteger
  */      
  public BigInteger rByte(String s) {
    return fromString(s,1);
  }

  /** Creates an int from a String encoded as specified int the SSH2 specification and
      beginning with an SSH2 BYTE.
      @param String s: the String containing the BYTE
      @return int: the BYTE as int
  */      
  public int rByteInt(String s) {
    return (rByte(s)).intValue();
  }

  /** Creates a boolean from a String encoded as specified in the SSH2 specification and
      beginning with a BOOLEAN.
      @param String s: the String containing the BYTE
      @return boolean: the BOOLEAN as boolean
  */      
  public boolean rBoolean(String s) {
    if (s.charAt(0)==(char)0) return false;
    return true;
  }

  /** Creates a BigInteger from a String encoded as specified in the SSH2 specification and
      beginning with an UINT32.
      @param String s: the String containing the UINT32
      @return BigInteger: the UINT32 as BigInteger
  */      
  public BigInteger rUint32(String s) {
    return fromString(s,4);
  }

  /** Creates an int from a String encoded as specified in the SSH2 specification and
      beginning with an UINT32.
      @param String s: the String containing the UINT32
      @return int: the UINT32 as int
  */      
  public int rUint32Int(String s) {
    return rUint32(s).intValue();
  }

  /** Creates a BigInteger from a String encoed as specified in the SSH2 specification
      and beginning with an UINT64.
      @param String s: the String containing the UINT64
      @return BigInteger: the UINT64 as BigInteger
  */      
  public BigInteger rUint64(String s) {
    return fromString(s,8);
  }

  /** Creates a long from a String encoed as specified in the SSH2 specification
      and beginning with an UINT64.
      @param String s: the String containing the UINT64
      @return long: the UINT64 as long
  */      
  public long rUint64Long(String s) {
    return fromString(s,8).longValue();
  }

  /** Creates a String from a String encoded as specified in the SSH2 specification
      and beginning with a STRING.
      @param String s: the String containing the STRING
      @return String: the STRING as String
  */
  public String rString(String s) {
    int l=0;
    try {
      l=rUint32Int(s);
    }
    catch (Exception e) {
      return e.toString();
    }
    if (l<0) {
      String ss=new String("------------------------------------------"+((char)13)+((char)10));
      ss=ss.concat(new String("negative length"+((char)13)+((char)10)));
      ss=ss.concat(new String(print(s)+((char)13)+((char)10)));
      ss=ss.concat(new String("------------------------------------------"+((char)13)+((char)10)));
      return ss;
    }
    return s.substring(4,4+l);
  }

  /** Creates a BigInteger from a String encoded as specified in the SSH2 specification
      and beginning with an MPINT.
      @param String s: the String containing the MPINT
      @return BigInteger: the MPINT as BigInteger
  */      
  public BigInteger rMpint(String s) {
    int len=rUint32Int(s);
    String r=s.substring(4,4+len);
    if (r.length()==0) {
      return new BigInteger("0");
    }
    else {
      return new BigInteger(converter.getBytes(r));
    }
  }

  // *******************************************************************************************
  // Functionality
  // *******************************************************************************************

  private BigInteger fromString(String s, int len) {
    char c[]=new char[len*2];
    char cc, cc_low, cc_high;
    for (int i=0; i<len; i++) {
      cc=s.charAt(i);
      cc_low=(char)(cc&0x000F);
      cc_high=(char)((cc&0x00F0)>>>4);
      c[i*2]=toHex(cc_high);
      c[i*2+1]=toHex(cc_low);
    }
    return new BigInteger(new String(c),16);
  }

  private char toHex(char c) {
    if (((int)c)<10) return (char)(((int)c)+48);
    return (char)(((int)c)+55);
  }
  
  // *******************************************************************************************
  // Miscellanous functionality
  // *******************************************************************************************

  /** Parses an unencoded SSH2 payload String and returns a Vector containing
      the elements of the payload as different Objects.
      @param String payload: the full payload String
      @param String[] desc: the String array describing the payload's structure<br>
        <ul>
	  <li>"B" (BYTE), element type BigInteger
	  <li>"3" (UINT32), element type BigInteger
	  <li>"6" (UINT64), element type BigInteger
	  <li>"M" (MPINT), element type BigInteger
	  <li>"S" (STRING), elemnt type String
	  <li>"O" (BOOLEAN), element type Boolean
	  <li>"Uxxx" (---), element type String, means not SSH2 coded xxx length string
	</ul>
      @return Vector: the Vector containing the elements as specified in desc
  */      	  
  public Vector parsePayload(String payload, String[] desc) {
    int pos=0;
    Vector r=new Vector();
    String fS;

    BigInteger bi;
    String s;
    Boolean bo;

    int len;

    for (int i=0; i<desc.length; i++) {
      fS=desc[i].substring(0,1);
      if (fS.equals("B")) {
        bi=rByte(payload.substring(pos,pos+1));
        r.addElement(bi);
        pos++;
      }
      else if (fS.equals("U")) {
        len=Integer.parseInt(desc[i].substring(1));
        s=payload.substring(pos,pos+len);
        r.addElement(s);
        pos+=len;
      }
      else if (fS.equals("M")) {
        len=rUint32Int(payload.substring(pos,pos+4));
        bi=rMpint(payload.substring(pos,pos+4+len));
        r.addElement(bi);
        pos+=4+len;
      }
      else if (fS.equals("S")) {
        len=rUint32Int(payload.substring(pos,pos+4));
        pos+=4;
        s=payload.substring(pos,pos+len);
        pos+=len;
        r.addElement(s);
      }
      else if (fS.equals("O")) {
        bo=new Boolean(rBoolean(payload.substring(pos,pos+1)));
        r.addElement(bo);
        pos++;
      }
      else if (fS.equals("3")) {
        bi=rUint32(payload.substring(pos,pos+4));
        r.addElement(bi);
        pos+=4;
      }
      else if (fS.equals("6")) {
        bi=rUint64(payload.substring(pos,pos+8));
        r.addElement(bi);
        pos+=8;
      }
    }

    return r;
  }

  /** Let s1 and s2 be two comma separated Strings. The result of this
      method r is a comma separated String. r contains elements, which 
      are contained in both s1 and s2. The order of the elements in r
      is either the order of the elements in s1 (pri=1) or the
      order of the elements in s2 (pri=2).
      @param String s1: s1
      @param String s2: s2
      @param int pri: pri
      @return String: r
  */
  public String findMatch(String s1, String s2, int pri) {
    String ss1, ss2;
    if (pri==1) {
      ss1=s1;
      ss2=s2;
    }
    else {
      ss1=s2;
      ss2=s1;
    }

    int pos;
    String ss3;
    String result=new String();

    while (ss1.length()>0) {
      ss3=getNextToken(ss1,SSH2Constants.SEP_COMMA);
      ss1=delNextToken(ss1,SSH2Constants.SEP_COMMA);

      pos=ss2.indexOf(ss3);
      if (pos!=-1) {
        if (result.length()>0) {
          result=result.concat(",");
        }
        result=result.concat(ss3);
      }
    }
    return result;
  }

  /** Returns the first element of the comma separated String s, or throws
      an Exception if s is empty.
      @param String s: the comma separated String
      @return String: the first element of s
      @throws Exception: is s is empty
  */      
  public String chosenMatch(String s) throws Exception {
    if (s.length()==0) {
      throw new Exception("Could not return chosen match. No match available");
    }
    int p=s.indexOf(',');
    if (p==-1) {
      return s;
    }
    else {
      return s.substring(0,p);
    }
  }

  /** Let msg be a String that can contain separating space Strings. The
      possible separating space Strings are described in spaces. This method
      returns the smallest left substring of msg that either ends with
      one of the separating space Strings or if msg does not contain
      any of the separating space Strings then this method returns
      msg.
      @param String msg: msg
      @param String[] spaces: separating space Strings
      @return String: the result as described above
  */      
  public String getNextToken(String msg, String spaces[]) {

    int pos[]=new int[spaces.length];
    int minpos=msg.length();

    for (int i=0; i<spaces.length; i++) {
      pos[i]=msg.indexOf(spaces[i]);
      if ((pos[i]!=-1) && (pos[i]<minpos)) {
        minpos=pos[i];
      }
    }
    return msg.substring(0,minpos);
  }

  /** This method deletes the left substring of msg that would be
      returned by getNextToken(msg, spaces) and the separaing
      space string at the end of the result from getNextToken as well (if there is
      any).
      @param String msg: msg
      @param String[] spaces: separating space Strings
      @return String: the result after the deletion
  */      
  public String delNextToken(String msg, String spaces[]) {
    int pos[]=new int[spaces.length];
    int minpos=msg.length();
    int dist=0;

    for (int i=0; i<spaces.length; i++) {
      pos[i]=msg.indexOf(spaces[i]);
      if ((pos[i]!=-1) && (pos[i]<minpos)) {
        minpos=pos[i];
        dist=spaces[i].length();
      }
    }
    return msg.substring(minpos+dist);
  }

  /** Creates a BigInteger from a String s. s is interpreted as
      as an unsigned number in 256 radix. s.charAt(0) is MSB, s.charAt(s.length()-1) is LSB.
      (s.charAt(i)&0x00ff)>>7 is MSB and s.charAt(i)&0x0001 is LSB.
      @param String s: String to be interpreted as an unsigned number
      @return BigInteger: the result
  */      
  public BigInteger stringAsBigInteger(String s) {
    String r=(new String("0")).concat(print(s));
    return new BigInteger(r,16);
  }

  private String print(int num) {
    char c[]=new char[8];
    int a;
    for (int i=0; i<8; i++) {
      a=num-((num>>4)<<4);
      c[7-i]=toHex((char)a);
      num>>=4;
    }
    return new String(c);
  }

  /** Coverts a String s interpreted as an unsigned number in 256 radix to a String r
      representing an unsigned number in 16 radix with the number signs
      "0"..."9" and "A"..."F". r.length()=s.length()*2. r.charAt(i*2)==HEX((s.charAt(i)&0x00f0)>>4),
      r.charAt(i*2+1)==HEX(s.charAt(i)&0x000f)
      @param String s: the String to be interpreted as an unsigned number in 256 radix
      @return String: the String interpreted as an unsigned number in 16 radix
  */      
  public String print(String s) {
    int len=s.length();
    char c[]=new char[len*2];
    char cc, cc_low, cc_high;

    for (int i=0; i<len; i++) {
      cc=s.charAt(i);
      cc_low=(char)(cc&0x000F);
      cc_high=(char)((cc&0x00F0)>>4);
      c[i*2]=toHex(cc_high);
      c[i*2+1]=toHex(cc_low);
    }

    return new String(c);
  }
  
  public String createList(Enumeration e, String sep) {
    String ret=new String();
    for (;e.hasMoreElements();) {
      if (ret.length()>0) {
        ret=ret.concat(sep);
      }
      ret=ret.concat(e.nextElement().toString());
    }
    return ret;
  }    	
  
  public boolean listContains(String list, String element, String listSep) {
  
    int len=element.length()+listSep.length();
    String con=element.concat(listSep);
  
    // is there a listseparator in the list
    int pos=list.indexOf(listSep);
    if (pos==-1) { // no listseparator, the whole list has to be the element
      return list.equals(element);
    }
    
    // => There are listseparators int the list

    // middle position    
    pos=list.indexOf(listSep.concat(con));
    if (pos!=-1) {
      return true;
    }        
    
    // beginning
    if (list.substring(0,len).equals(con)) {
      return true;
    }
    
    // end    
    if (list.substring(list.length()-len).equals(listSep.concat(element))) {
      return true;
    }  
    
    return false;    
  }
    
}