package com.monica.javacom.ssh2;

import com.monica.javacom.interfaces.*;
import com.monica.interfaces.*;
import com.monica.javacom.ssh2.*;
import com.monica.util.*;
import com.monica.util.hash.*;
import com.monica.javacom.ssh2.cipher.*;
import com.monica.javacom.ssh2.compression.*;
import com.monica.javacom.ssh2.mac.*;
import com.monica.util.configuration.*;
import com.monica.javacom.ssh2.interfaces.*;
import com.monica.util.debug.*;

import java.util.*;
import java.math.*;

public class SSH2Tokenizer implements com.monica.interfaces.Reader, com.monica.interfaces.Writer, Configureable, Runnable, Closeable {

  // *****************************************************************************************************
  // Public Constants
  // *****************************************************************************************************

  public static final int RET_OK=0;

  // *****************************************************************************************************
  // Private Constants
  // *****************************************************************************************************
  private static final String ERR_NO_PACKET="There is no packet to retrieve from the SSH transport layer.";
  private static final String ERR_UA_ALREADY_RUNNING="User Authentication Protocol already running.";
  private static final String ERR_CONN_ALREADY_RUNNING="Connection Layer already running.";
  private static final String ERR_CONN_NOT_RUNNING="Connection layer not running.";
  
  private static final String NAME="SSH2Tokenizer";
  private static final String VERSION="0.31";
  
  private static final String DP_S="SSH2Tokenizer";

  // *****************************************************************************************************
  // Public Constants
  // *****************************************************************************************************

  // *****************************************************************************************************
  // Variables
  // *****************************************************************************************************
  private MACAlgorithm macAlgorithmCTS, macAlgorithmSTC;
  private CipherAlgorithm cipherAlgorithmCTS, cipherAlgorithmSTC;
  private CompressionAlgorithm compressionAlgorithmCTS, compressionAlgorithmSTC;

  private String lastError;
  private String buffer;

  private SSH2IO io;
  private Vector packets;

  private int minProcessSizeCTS, minProcessSizeSTC;
  private int macLengthCTS, macLengthSTC;
  private String IV_CTS, IV_STC, EK_CTS, EK_STC, IK_CTS, IK_STC;
  private int sequenceNumberCTS, sequenceNumberSTC;

  private int status;

  private Random rnd;
  private Configuration configuration;
  private Readable readable;
  private Writeable writeable;
  private SSH2BothSideFeatures owner;
  private SSH2TransportLayer transportLayer;
  private SSH2UserAuthenticationLayer uaLayer;
  private Hashtable uaLayers;
  private int side;
  
  private Object cParam;

  private HashAlgorithm hash;
  private SHA1 sha1;

  // Thread management

  private boolean threadSuspended;
  private boolean threadRunning;
  private Thread myThread;
  
  private boolean connected;
  
  private SSH2ChannelManager channelManager;
  private SSH2ConnectionLayer connLayer;
  
  private boolean connectionLayerReady;
  
  // *****************************************************************************************************
  // Constructor
  // *****************************************************************************************************
  public SSH2Tokenizer(SSH2BothSideFeatures owner, int side) {
    this.owner=owner;
    this.side=side;

//    side=SSH2Constants.SSH_SIDE_NOT_SET;

    io=new SSH2IO();
    packets=new Vector();
    transportLayer=new SSH2TransportLayer();
    transportLayer.setOwner(owner);
    transportLayer.setTokenizer(this);
    transportLayer.setSide(side);

    lastError=new String();
    buffer=new String();
    IV_CTS=new String(); EK_CTS=new String(); IK_CTS=new String();
    IV_STC=new String(); EK_STC=new String(); IK_STC=new String();
    sequenceNumberCTS=0; sequenceNumberSTC=0;
    minProcessSizeCTS=8; minProcessSizeSTC=8;
    macLengthCTS=0; macLengthSTC=0;

    setCipherAlgorithm(new SSH2NoneCipher(),SSH2Constants.SSH_TOKENIZER_CTS);
    setCipherAlgorithm(new SSH2NoneCipher(),SSH2Constants.SSH_TOKENIZER_STC);
    setMACAlgorithm(new SSH2NoneMAC(),SSH2Constants.SSH_TOKENIZER_CTS);
    setMACAlgorithm(new SSH2NoneMAC(),SSH2Constants.SSH_TOKENIZER_STC);
    setCompressionAlgorithm(new SSH2NoneCompression(),SSH2Constants.SSH_TOKENIZER_CTS);
    setCompressionAlgorithm(new SSH2NoneCompression(),SSH2Constants.SSH_TOKENIZER_STC);

    connected=true;
    
    uaLayers=new Hashtable();
    
    sha1=new SHA1();
    
    connectionLayerReady=false;

    threadRunning=true;
    threadSuspended=true;
    myThread=new Thread(this);
    myThread.start();
  }
  
  public void setChannelManager(SSH2ChannelManager chManager) {
      channelManager=chManager;
      try {
        if (side==SSH2Constants.SSH_SIDE_SERVER) {
          connLayer=owner.getConnectionLayer(cParam, chManager);
          connLayer.setTokenizer(this);
          connLayer.setConfiguration(configuration);
          connLayer.setSide(side);
          connLayer.setOwner(owner);
        }          
      }
      catch (Exception e) {
        DebugManager.logException(DP_S, e);  
      }
  }
  
  public SSH2ChannelManager getChannelManager() {
      return channelManager;
  }
  
  public boolean connected() {
    return connected;
  }

  // *******************************************************************************************
  // Runnable
  // *******************************************************************************************

  public String getCloseableName() {
    return NAME;
  }

  public String getCloseableVersion() {
    return VERSION;
  }

  public void closeConnections() {
    if (connected) {  
        DebugManager.logOut(DP(), "Closing connections [closeConnections method invoked] ...");
        String temp=io.wByte(SSH2Constants.SSH_TRANSPORT_MSG_DISCONNECT);
        temp=temp.concat(io.wUint32(SSH2Constants.SSH_TRANSPORT_DISCONNECT_BY_APPLICATION));
        temp=temp.concat(io.wString("Requested by user."));
        temp=temp.concat(io.wString("en"));
        sendP(temp);
        connected=false;
        owner.closeConnections();
    }
  }

  // *******************************************************************************************
  // Runnable
  // *******************************************************************************************

  /** Entry point for the reader thread.
  */
  public void run() {
    while (threadRunning) {
      synchronized(this) {
        while (threadSuspended) {
          try {
            wait();
          }
          catch (Exception e) {
            //--
          }
        }
        threadSuspended=true;
      }
      process(readable.getInput());
    }
  }

  // *******************************************************************************************
  // Configureable
  // *******************************************************************************************

  public void setConfiguration(Configuration configuration) {
    this.configuration=configuration;
    rereadConfig();
  }

  public Configuration getConfiguration() {
    return configuration;
  }

  private void rereadConfig() {      
    transportLayer.setConfiguration(configuration);
    if (connLayer!=null) {
      connLayer.setConfiguration(configuration);
    }
  }

  // *******************************************************************************************
  // Reader
  // *******************************************************************************************

  public String getReaderName() {
    return NAME;
  }

  public String getReaderVersion() {
    return VERSION;
  }

  public synchronized void readEvent(Readable readable) {
    DebugManager.logOut(DP(), "readevent"+readable.getAvailable());  
    if (readable.getAvailable()>0) {
      threadSuspended=false;
      notify();
//      DebugManager.logOut(DP(), "notify sent");
    }
  }

  public synchronized void connectionClosedEvent(Readable readable) {
    connected=false;
    
    if (readable.getAvailable()>0) {
      readEvent(readable);
    }
    DebugManager.logOut(DP(), "Connection closed by other side.");
    
    writeToReader("Connection closed."+SSH2Constants.CRLF);
    if (connLayer!=null) {
      DebugManager.logOut(DP(), "Closing connection layer.");  
      connLayer.closeConnections();
    }  
//    transportProtocolReady();
//    userAuthenticationProtocolReady();
//    connectionProtocolReady();
    threadRunning=false;
    threadSuspended=false;
    notify();
  }

  public synchronized void errorEvent(String errorDesc, Readable readable) {
    connected=false;
    writeToReader("Error occured ("+errorDesc+"). Connection closed."+SSH2Constants.CRLF);
    closeConnections();
    threadRunning=false;
    threadSuspended=false;
    notify();
  }

  public synchronized void setReadable(Readable readable) {
    this.readable=readable;
  }

  public Readable getReadable() {
    return readable;
  }

  // *******************************************************************************************
  // Writer
  // *******************************************************************************************

  public String getWriterName() {
    return NAME;
  }

  public String getWriterVersion() {
    return VERSION;
  }

  public void setWriteable(Writeable writeable) {
    this.writeable=writeable;
  }

  public Writeable getWriteable() {
    return writeable;
  }

  // *******************************************************************************************
  // SSH2
  // *******************************************************************************************

  /** Processes the stream represented as a String received from the upper  layer.
      @param String msg: the stream String
  */
  private void process(String msg) {
      
    DebugManager.logOut(DP(), "process start, inLen="+msg.length());  

    String cur_msg;
    SSH2Packet pack;
    BigInteger msg_type;
    int msg_type_int;
    String temp;
    Vector dVector;    

    switch (status) {

      case SSH2Constants.STAT_INIT:

        temp=transportLayer.processInitMsg(msg);
        if (temp!=null) {
          process(temp);
        }
        break;

      case SSH2Constants.STAT_SSH:

        try {
                        
          arrived(msg);              

          while (processBuffer()) {  
            while (isAvailable()) {
              pack=getNextPacket();
              msg_type=io.rByte(pack.payload);
              msg_type_int=msg_type.intValue();

              DebugManager.logOut(DP(), "in="+msg_type_int);
//              writeToReader();

              switch (msg_type_int) {

                // Transport
	        case SSH2Constants.SSH_TRANSPORT_MSG_IGNORE :
	          DebugManager.logOut(DP(), "ignore: "+pack.payload.substring(1));
	          break;
                case SSH2Constants.SSH_TRANSPORT_MSG_KEXINIT :
                  transportLayer.processKexInitMsg(pack);
                  break;
                case SSH2Constants.SSH_TRANSPORT_MSG_KEXDH_REPLY :
                  transportLayer.processKexDHReplyMsg(pack);
                  break;
                case SSH2Constants.SSH_TRANSPORT_MSG_NEWKEYS :
                  transportLayer.processNewKeysMsg(pack);
                  break;
                case SSH2Constants.SSH_TRANSPORT_MSG_DISCONNECT :
                  transportLayer.processDisconnectMsg(pack);
                  break;
                case SSH2Constants.SSH_TRANSPORT_MSG_KEXDH_INIT :
                  transportLayer.processKexDHInitMsg(pack);
                  break;
                case SSH2Constants.SSH_TRANSPORT_MSG_UNIMPLEMENTED :
                  closeConnections();
                  break;
                                    
                case SSH2Constants.SSH_TRANSPORT_MSG_SERVICE_ACCEPT :
                  processServiceAcceptMsg(pack);
                  break;
                case SSH2Constants.SSH_TRANSPORT_MSG_SERVICE_REQUEST :
                  processServiceRequestMsg(pack);
                  break;  

                // User Authentication
                case SSH2Constants.SSH_USERAUTH_MSG_USERAUTH_SUCCESS :
                  if (uaLayer!=null) {
                    uaLayer.processUserAuthSuccessMsg(pack);
                  }
                  else {
                    sendUnimplementedMsg(pack);
                  }
                  break;
                case SSH2Constants.SSH_USERAUTH_MSG_USERAUTH_PK_OK :
                  if (uaLayer!=null) {
                    uaLayer.processUserAuthPkOkMsg(pack);
                  }
                  else {
                    sendUnimplementedMsg(pack);
                  }
                  break;
                case SSH2Constants.SSH_USERAUTH_MSG_USERAUTH_FAILURE :
                  if (uaLayer!=null) {
                    uaLayer.processUserAuthFailureMsg(pack);
                  }
                  else {
                    sendUnimplementedMsg(pack);
                  }
                  break;
                case SSH2Constants.SSH_USERAUTH_MSG_USERAUTH_BANNER :
		  sendUnimplementedMsg(pack);
                  break;
                case SSH2Constants.SSH_USERAUTH_MSG_USERAUTH_REQUEST :
                  dVector=io.parsePayload(pack.payload,SSH2Constants.USERAUTH_REQUEST_MSG_DATA_DESC);
                  temp=(String)dVector.elementAt(SSH2Constants.USERAUTH_REQUEST_METHOD);                  
		  DebugManager.logOut(DP(), temp);
                  try {
		    getUALayer(temp, null);
                    uaLayer.processUserAuthRequestMsg(pack);
                  }
                  catch (Exception e) {
                    DebugManager.logException(DP_S, e);
		    DebugManager.logOut(DP(), "Request-failure");
                    userAuthFailure();  
                  }
                  break;

                // Connection
                  
                  // GLOBAL                  
                case SSH2Constants.SSH_CONNECTION_MSG_GLOBAL_REQUEST :
                  if (connLayer!=null) {  
                    connLayer.processConnectionGlobalRequest(pack);
                  }
                  else {
                    sendUnimplementedMsg(pack);
                  }
                  break;
                case SSH2Constants.SSH_CONNECTION_MSG_REQUEST_SUCCESS :
                  if (connLayer!=null) {  
                    connLayer.processConnectionRequestSuccess(pack);
                  }
                  else {
                    sendUnimplementedMsg(pack);
                  }
                  break;
                case SSH2Constants.SSH_CONNECTION_MSG_REQUEST_FAILURE :
                  if (connLayer!=null) {  
                    connLayer.processConnectionRequestFauilure(pack);
                  }
                  else {
                   sendUnimplementedMsg(pack);
                  }
                  break;
                      
                // OPEN                                                                        
                case SSH2Constants.SSH_CONNECTION_MSG_CHANNEL_OPEN :
                  if (connLayer!=null) {  
                    connLayer.processConnectionChannelOpen(pack);
                  }
                  else {
                    sendUnimplementedMsg(pack);
                  }
                  break;
                case SSH2Constants.SSH_CONNECTION_MSG_CHANNEL_OPEN_CONFIRMATION :
                  if (connLayer!=null) {  
                    connLayer.processConnectionChannelOpenConfirmation(pack);
                  }
                  else {
                    sendUnimplementedMsg(pack);
                  }
                  break;
                case SSH2Constants.SSH_CONNECTION_MSG_CHANNEL_OPEN_FAILURE :
                  if (connLayer!=null) {  
                    connLayer.processConnectionChannelOpenFailure(pack);
                  }
                  else {
                    sendUnimplementedMsg(pack);
                  }
                  break;
                    
                // REQUEST  
                case SSH2Constants.SSH_CONNECTION_MSG_CHANNEL_REQUEST :
                  DebugManager.logOut(DP_S, "chReq-1");      
                  if (connLayer!=null) {  
                    DebugManager.logOut(DP_S, "chReq-2");    
                    connLayer.processConnectionChannelRequest(pack);
                  }
                  else {
                    DebugManager.logOut(DP_S, "chReq-3");      
                    sendUnimplementedMsg(pack);
                  }
                  break;                                        
                case SSH2Constants.SSH_CONNECTION_MSG_CHANNEL_SUCCESS :
                  DebugManager.logOut(DP_S, "chSucc-1");  
                  if (connLayer!=null) {  
                    DebugManager.logOut(DP_S, "chSucc-2-1");    
                    connLayer.processConnectionChannelSuccess(pack);
                    DebugManager.logOut(DP_S, "chSucc-2-2");    
                  }
                  else {
                    DebugManager.logOut(DP_S, "chSucc-3");    
                    sendUnimplementedMsg(pack);
                  }
                  break;
                case SSH2Constants.SSH_CONNECTION_MSG_CHANNEL_FAILURE :
                  DebugManager.logOut(DP_S, "chFail-1");      
                  if (connLayer!=null) {  
                    DebugManager.logOut(DP_S, "chFail-2");        
                    connLayer.processConnectionChannelFailure(pack);
                  }
                  else {
                    DebugManager.logOut(DP_S, "chFail-3");        
                    sendUnimplementedMsg(pack);
                  }
                  break;
                    
                // DATA                    
                case SSH2Constants.SSH_CONNECTION_MSG_CHANNEL_DATA :
                  if (connLayer!=null) {  
                    connLayer.processConnectionChannelData(pack);
                  }
                  else {
                    sendUnimplementedMsg(pack);
                  }                    
                  break;
                case SSH2Constants.SSH_CONNECTION_MSG_CHANNEL_EXTENDED_DATA :
                  if (connLayer!=null) {  
                    connLayer.processConnectionChannelExtendedData(pack);
                  }
                  else {
                    sendUnimplementedMsg(pack);
                  }                    
                  break;
                    
                // WINDOW ADJUST                      
                case SSH2Constants.SSH_CONNECTION_MSG_CHANNEL_WINDOW_ADJUST :
                  if (connLayer!=null) {  
                    connLayer.processConnectionChannelWindowAdjust(pack);
                  }
                  else {
                    sendUnimplementedMsg(pack);
                  }
                  break;
                    
                // CLOSE                    
                case SSH2Constants.SSH_CONNECTION_MSG_CHANNEL_CLOSE :
                  if (connLayer!=null) {  
                    connLayer.processConnectionChannelClose(pack);
                  }
                  else {
                    sendUnimplementedMsg(pack);
                  }
                  break;
                case SSH2Constants.SSH_CONNECTION_MSG_CHANNEL_EOF :
                  if (connLayer!=null) {  
                    connLayer.processConnectionChannelEOF(pack);
                  }
                  else {
                    sendUnimplementedMsg(pack);
                  }
                  break;

                // Other
                default:
                  sendUnimplementedMsg(pack);
                  break;

              }
            }  
	  }
        }    
        catch (Exception e) {
          writeToReader(SSH2Constants.CRLF);
          writeToReader("Exception occured while SSH2 processing: "+e.getMessage());
	  DebugManager.logException(DP_S, e);;
          closeConnections();
          return;
	}
        break;
    }
  }
  
  private void getUALayer(String name, Object param) throws Exception {
   
    uaLayer=(SSH2UserAuthenticationLayer)uaLayers.get(name);
    if (uaLayer==null) {
      uaLayer=(SSH2UserAuthenticationLayer)owner.getUserAuthenticationLayer(name, param);  
      uaLayer.setTokenizer(this);
      uaLayer.setSide(side);
      uaLayers.put(name, uaLayer);
    }  
    
  }

  public void setStatus(int s) {
    status=s;
  }

  public int getStatus() {
    return status;
  }

  public int startUserAuthentication(SSH2UserAuthenticationLayer layer) throws Exception {
    if (uaLayer!=null) {
      throw new Exception(ERR_UA_ALREADY_RUNNING);
    }
    uaLayer=layer;
    sendServiceRequestMsg("ssh-userauth");
    return RET_OK;
  }
  
  public int startConnection(SSH2ConnectionLayer layer) throws Exception {
    if (connLayer!=null) {
      throw new Exception(ERR_CONN_ALREADY_RUNNING);
    }
    DebugManager.logOut(DP(), "starting connection: "+layer.getClass().toString());    
    connLayer=layer;
//    DebugManager.logOut(DP(), "starting conn layer via start method");
    connLayer.start();
//    DebugManager.logOut(DP(), "returning ok");
    return RET_OK;
  }
  
  public SSH2Channel startNewChannel() throws Exception {
    if (connLayer==null) {
      throw new Exception(ERR_CONN_NOT_RUNNING);
    }
    return connLayer.startNewChannel();
  }
  
  /** Processes the service accept message from the server.
      @param SSHTransportPacket pack: the service accept message from the server
  */
  private void processServiceAcceptMsg(SSH2Packet pack) {
    if (pack.payload.length()>1) {
      String serviceName=io.rString(pack.payload.substring(1));
      if (serviceName.equals("ssh-userauth")) {
        if (uaLayer!=null) {
          try {
            uaLayer.start();
          }
          catch (Exception e) {
            writeToReader("Error occured during service accept msg processing. Exception="+e.toString());
            DebugManager.logException(DP_S, e);
            closeConnections();
          }
        }
      }
      else {
        sendUnimplementedMsg(pack);
        closeConnections();
      }
    }
/*    else {
      loginAttempt(LA_REQ_ACCEPT);
    } */
  }
  
  private void processServiceRequestMsg(SSH2Packet pack) {
    String serviceName;  
    String toSend;
    switch (side) {
      case SSH2Constants.SSH_SIDE_CLIENT :
        sendUnimplementedMsg(pack);
        closeConnections();
        break;
      case SSH2Constants.SSH_SIDE_SERVER :
        serviceName=io.rString(pack.payload.substring(1));
        DebugManager.logOut(DP(), serviceName);
        if (serviceName.equals("ssh-userauth")) {
          toSend=io.wByte(SSH2Constants.SSH_TRANSPORT_MSG_SERVICE_ACCEPT);
          toSend+=io.wString(serviceName);
          sendP(toSend);
        }
        else {
          sendUnimplementedMsg(pack);  
          closeConnections();
        }
        break;
    }
  }
                        
  /** Replies an UNIMPLEMENTED message for a packet that contains a message type that
      is not implemented in this client or is invalid.
      @param SSHTransportPacket pack: the packet that is invalid or unimplemented
  */
  public void sendUnimplementedMsg(SSH2Packet pack) {
    int msg_type_int=io.rByteInt(pack.payload);
    DebugManager.logOut(DP(), "Unknown message type="+msg_type_int);

    String temp=io.wByte(SSH2Constants.SSH_TRANSPORT_MSG_UNIMPLEMENTED);
    temp=temp.concat(io.wUint32(pack.sequenceNumber));
    sendP(temp);
  }


  // *****************************************************************************************************
  // Tokenizer functionality
  // *****************************************************************************************************
  // true=CTS, false=STC
  /** Sets the keys for encryption (iv, key) and integrity. This method
      has to be called before the setCipherAlgorithm, setCompressionAlgorithm
      and setMACAlgorithm methods.
      @param String iv: the initialization vector for ciphering
      @param String ek: the encryption key
      @param String ik: the integrity key (for MAC)
      @param boolean dir: true for client to server, false for server to client
  */
  public void setKeys(String iv, String ek, String ik, boolean dir) {
    if (dir) {
      IV_CTS=iv;
      EK_CTS=ek;
      IK_CTS=ik;
    }
    else {
      IV_STC=iv;
      EK_STC=ek;
      IK_STC=ik;
    }
  }

  // true=CTS, false=STC
  /** Sets the MAC algorithm that has to be used.
      @param SSHMACAlgorithm: the MAC algorithm to be used
      @param boolean dir: true for client to server, false for server to client
  */
  public void setMACAlgorithm(MACAlgorithm ma, boolean dir) {
    if (dir) {
      macAlgorithmCTS=ma;
      macLengthCTS=macAlgorithmCTS.getMACLength();
    }
    else {
      macAlgorithmSTC=ma;
      macLengthSTC=macAlgorithmSTC.getMACLength();
    }
  }

  // true=CTS, false=STC
  /** Sets the Cipheralgorithm that has to be used.
      @param CipherAlgorithm: the Cipher algorithm to be used
      @param boolean dir: true for client to server, false for server to client
  */
  public void setCipherAlgorithm(CipherAlgorithm ca, boolean dir) {
    if (dir) {
      cipherAlgorithmCTS=ca;
      cipherAlgorithmCTS.setKeys(EK_CTS, IV_CTS);
      if (cipherAlgorithmCTS.getCipherBlockSize()>minProcessSizeCTS) {
        minProcessSizeCTS=cipherAlgorithmCTS.getCipherBlockSize();
      }
    }
    else {
      cipherAlgorithmSTC=ca;
      cipherAlgorithmSTC.setKeys(EK_STC, IV_STC);
      if (cipherAlgorithmSTC.getCipherBlockSize()>minProcessSizeSTC) {
        minProcessSizeSTC=cipherAlgorithmSTC.getCipherBlockSize();
      }
    }
  }

  // true=CTS, false=STC
  /** Sets the Compression algorithm that has to be used.
      @param CompressionAlgorithm: the MAC algorithm to be used
      @param boolean dir: true for client to server, false for server to client
  */
  public void setCompressionAlgorithm(CompressionAlgorithm ca, boolean dir) {
    if (dir) {
      compressionAlgorithmCTS=ca;
    }
    else {
      compressionAlgorithmSTC=ca;
    }
  }

  /** This method tells this object that at the to be processsed stream
      new data has arrived.
      @param String s: the arrived data
      @throws Exception: if there has been an Exception during the processing of
        the arrived data
  */
  public synchronized void arrived(String s) throws Exception {
    buffer=buffer.concat(s);
  }

  /** Returns true if there are processed packets that can be retrieved.
      @return boolean: true if there processed packets that can be retrieved
  */
  public synchronized boolean isAvailable() {
    return (packets.size()>0);
  }

  /** Returns the next available packet from the processed packets.
      @return SSH2Packet: the next packet from the processed packets
        or null if there aren't any
  */
  public synchronized SSH2Packet getNextPacket() {
    if (packets.size()>0) {
      SSH2Packet p=((SSH2Packet)packets.elementAt(0));
      packets.removeElementAt(0);
      return p;
    }
    else {
      lastError=ERR_NO_PACKET;
      return null;
    }
  }

  /** Returns the description of the last occured error. By calling this method
      the error description will be lost.
      @return String: the last occured error
  */
  public String getLastError() {
    String temp=lastError;
    lastError=new String();
    return temp;
  }

  private synchronized boolean processBuffer() throws Exception {
      return _us_processBuffer();
  }
  
  private boolean _us_processBuffer() throws Exception {
      
    DebugManager.logOut(DP(), "_us_processBuffer()");
    
    String pckl_e, pckl_d, p_e, p_d;
    BigInteger pckl, padl;
    int pckl_int, padl_int, payload_l;
    SSH2Packet pack;
    String unencryptedPacket;
    Object o;
    
    CompressionAlgorithm coA=null;
    CipherAlgorithm ciA=null;
    MACAlgorithm mA=null;
    int minLen=0;
    int seqNo=0;
    int macLen=0;
    String IK=null;
    
//    DebugManager.logOut(DP(), "bufferlen at startup: "+buffer.length());
        
    switch (side) {
        case SSH2Constants.SSH_SIDE_SERVER :
            coA=compressionAlgorithmCTS;
            ciA=cipherAlgorithmCTS;
            mA=macAlgorithmCTS;
            minLen=minProcessSizeCTS;
            IK=IK_CTS;
            macLen=macLengthCTS;
            break;
        case SSH2Constants.SSH_SIDE_CLIENT :
            coA=compressionAlgorithmSTC;
            ciA=cipherAlgorithmSTC;
            mA=macAlgorithmSTC;
            minLen=minProcessSizeSTC;
            IK=IK_STC;
            macLen=macLengthSTC;
            break;
    }
    
    if (buffer.length()>minLen) {
      pckl_e=buffer.substring(0,minLen);
      o=ciA.getState();

      pckl_d=ciA.decode(pckl_e);
      pckl=io.rUint32(pckl_d);
      pckl_int=pckl.intValue();

//      DebugManager.logOut(DP(), "pckl_e /* buffer.substring(0,minLen) */ ="+pckl_e);
//      DebugManager.logOut(DP(), "pckl_d /* decode(pckl_e) */ ="+pckl_d);
//      DebugManager.logOut(DP(), "pckl_int /* int from pckl_d => packet length */ ="+pckl_int);
                  
      if (pckl_int<0) {
        throw new Exception("Decode error: smaller: ="+io.print(pckl_d.substring(0,4))+", intvalue="+pckl_int);
      }
      
      if (buffer.length()>=(4+pckl_int+macLen)) { // whole packet is here
      
        DebugManager.logOut(DP(), "We have the whole packet. ");
          
        switch (side) {
          case SSH2Constants.SSH_SIDE_SERVER :
            seqNo=sequenceNumberCTS;
            sequenceNumberCTS++;
            break;
          case SSH2Constants.SSH_SIDE_CLIENT :
            seqNo=sequenceNumberSTC;
            sequenceNumberSTC++;
            break;        
        }
        
//        DebugManager.logOut(DP(), "Processing packet ("+seqNo+") with "+ciA.getCipherAlgorithmName());
                  
        p_e=buffer.substring(minLen,4+pckl_int);
        p_d=pckl_d.concat(ciA.decode(p_e)); //decrypted packet data

        pack=new SSH2Packet();

        pack.packet_length=pckl;
        pack.padding_length=io.rByte(p_d.substring(4,5));
        padl_int=pack.padding_length.intValue();
        payload_l=pckl_int-padl_int-1;

        pack.payload=p_d.substring(5,5+payload_l);

        // decompressing payload
        pack.payload=coA.decompress(pack.payload);

        pack.random_padding=p_d.substring(5+payload_l,5+payload_l+padl_int);

        pack.mac=buffer.substring(pckl_int+4,pckl_int+4+macLen);
        
        unencryptedPacket=io.wUint32(seqNo);
        pack.sequenceNumber=seqNo;

        unencryptedPacket=unencryptedPacket.concat(p_d);

//        DebugManager.logOut(DP(), "packet type="+(int)(pack.payload.charAt(0))+", seqNo="+(sequenceNumberSTC-1));
        if (!mA.authenticate(unencryptedPacket,pack.mac,IK)) {
          DebugManager.logOut(DP(), "REC PACKauth Failed!");  
          throw new Exception("REC PACKauth Failed!");
        }

        buffer=buffer.substring(4+pckl_int+macLen);

        packets.addElement(pack);
	
//	DebugManager.logOut(DP(), "bufferLen after packet takeout="+buffer.length());
	
        return true;
      }
      else {
      
        ciA.setState(o);
        
        DebugManager.logOut(DP(), "Waiting for the whole packet.");        
        
        DebugManager.logOut(DP(), "Buffer length should be at least "+(4+pckl_int+macLen));
        DebugManager.logOut(DP(), "Buffer length ="+buffer.length());
        
        int r_len=readable.getAvailable();
                
        if (r_len>0) {
            DebugManager.logOut(DP(), "There is data available from the readable: "+r_len);
            buffer+=readable.getInput();
            return _us_processBuffer();
        }
                          
      }
    }
    return false;

  }

  /** Creates an encrypted, MAC-ed, compressed SSH2 transport packet from the supplied payload
      and returns it as a String.
      @param String payload: the packet payload
      @return String: the full SSH2 transport packet
  */
  public String createPacket(String payload) {
      
    CompressionAlgorithm coA=null;
    CipherAlgorithm ciA=null;
    MACAlgorithm mA=null;
    int minLen=0;
    int seqNo=0;
    String IK=null;
    
    DebugManager.logOut(DP(), "out="+((int)payload.charAt(0)));
    
    switch (side) {
        case SSH2Constants.SSH_SIDE_CLIENT :
            coA=compressionAlgorithmCTS;
            ciA=cipherAlgorithmCTS;
            mA=macAlgorithmCTS;
            minLen=minProcessSizeCTS;
            seqNo=sequenceNumberCTS;
            sequenceNumberCTS++;
            IK=IK_CTS;
            break;
        case SSH2Constants.SSH_SIDE_SERVER :
            coA=compressionAlgorithmSTC;
            ciA=cipherAlgorithmSTC;
            mA=macAlgorithmSTC;
            minLen=minProcessSizeSTC;
            seqNo=sequenceNumberSTC;
            sequenceNumberSTC++;
            IK=IK_STC;
            break;
    }
           
//    DebugManager.logOut(DP(), "Creating packet ("+seqNo+") with "+ciA.getClass().toString());
            
    // compressing payload;
      
    payload=coA.compress(payload);

    int len_without_mac;
    int padding_len;
    String result;
    String mac;
    char padding[];

    len_without_mac=16;
    padding_len=len_without_mac-5-payload.length();
    if (padding_len<4) padding_len=4;
    len_without_mac=padding_len+payload.length()+5;
    while ((len_without_mac%minLen)!=0) len_without_mac++;
    padding_len=len_without_mac-5-payload.length();

    result=io.wUint32(len_without_mac-4);
    result=result.concat(io.wByte(padding_len));
    result=result.concat(payload);

    padding=new char[padding_len];
    for (int i=0; i<padding_len; i++) {
      padding[i]=(char)(Math.random()*255.0);
    }

    result=result.concat(new String(padding));

    // time profiling
    long TP_start, TP_stop, TP_diff; //--
    int TP_bytelen; //--
    
    TP_start=System.currentTimeMillis(); //--
    TP_bytelen=4+result.length(); //--
    
    mac=mA.createMAC((io.wUint32(seqNo)).concat(result),IK);
   
    TP_stop=System.currentTimeMillis(); //--    
    TP_diff=(TP_stop-TP_start)+1;
//    DebugManager.logOut(DP(), "MAC speed="+((TP_bytelen/TP_diff)*1000)+" at the len="+TP_bytelen); //--
    
    TP_start=System.currentTimeMillis(); //--
    TP_bytelen=result.length(); //--
    
    result=ciA.encode(result);
    
    TP_stop=System.currentTimeMillis(); //--    
    TP_diff=(TP_stop-TP_start)+1;
//    DebugManager.logOut(DP(), "Cipher speed="+((TP_bytelen/TP_diff)*1000)+" at the len="+TP_bytelen); //--
    
    result=result.concat(mac);

    return result;
  }

  public void sendP(String pack) {
    synchronized (this) {
      pack=createPacket(pack);
      writeable.write(pack);
      writeable.flush();
    }
  }

  public void sendNP(String pack) {
    synchronized (this) {
      writeable.write(pack);
      writeable.flush();
    }
  }

  public void transportProtocolReady() {
    DebugManager.logOut(DP(), "Transport Protocol ready.");
    owner.event(SSH2BothSideFeatures.WAIT_TR);
  }

  public void userAuthenticationProtocolReady() {
    DebugManager.logOut(DP(), "User Authentication Protocol ready.");
    owner.event(SSH2BothSideFeatures.WAIT_UA);
  }
  
  public void connectionProtocolReady() {
    DebugManager.logOut(DP(), "Connection Protocol ready.");
    connectionLayerReady=true;
    owner.event(SSH2BothSideFeatures.WAIT_CO);
  }
  
  public boolean isConnectionLayerReady() {
      return connectionLayerReady;
  }
            
  public void writeToReader(String s) {
//    DebugManager.logOut(DP(), s);
  }
  
  public void sendVersionString() {
      transportLayer.sendVersionString();
  }
  
 /** Sends a service request message to the server with a specified service name.
      @param String serviceName: the service name to be requested
  */
  private void sendServiceRequestMsg(String payload) {
    String toSend=io.wByte(SSH2Constants.SSH_TRANSPORT_MSG_SERVICE_REQUEST);
    toSend=toSend.concat(io.wString(payload));
    sendP(toSend);
  }
  
  public void userAuthSuccess() {
    String toSend=io.wByte(SSH2Constants.SSH_USERAUTH_MSG_USERAUTH_SUCCESS);
    sendP(toSend);
  }
  
  public void userAuthFailure() throws Exception {
    String possibleContinuation="";  
    String toSend;
    toSend=io.wByte(SSH2Constants.SSH_USERAUTH_MSG_USERAUTH_FAILURE);
    toSend+=io.wString(possibleContinuation());
    toSend+=io.wBoolean(true);
    sendP(toSend);
  }
  
  private String possibleContinuation() throws Exception {
    String originalList=owner.getFeatureList(SSH2Facade.OPT_USERAUTHENTICATION_LAYER);
    String createdList=new String();
    String curItem;
    SSH2UserAuthenticationLayer l;
    while (originalList.length()>0) {
      curItem=io.getNextToken(originalList, SSH2Constants.SEP_COMMA);
      originalList=io.delNextToken(originalList, SSH2Constants.SEP_COMMA);
      l=(SSH2UserAuthenticationLayer)uaLayers.get(curItem);
      if (l!=null) {
        if (l.canContinue()) {
          if (createdList.length()>0) {
            createdList+=SSH2Constants.SEP_COMMA_NA+l.methodName();
          }
          else {
            createdList=l.methodName();
          }
        }
      }
      else {
        if (createdList.length()>0) {
          createdList+=SSH2Constants.SEP_COMMA_NA+curItem;
        }
        else {
          createdList=curItem;
        }
      }
    }	      
    DebugManager.logOut(DP(), "possible continuation="+createdList);
    return createdList;
  }      
  
  public String getSessionID() {
      return transportLayer.getSessionID();
  }
  
  public void performServerCheck() {
      transportLayer.performServerCheck();
  }
  
  public void setCParam(Object cParam) {
      this.cParam=cParam;      
  }
  
  public String getInformation(int opt) {
      
      DebugManager.logOut(DP(), "getInformation : "+opt);
      
      String r;
      r=transportLayer.getInformation(opt);
      if (r!=null) {
          return r;
      }
      if (uaLayer!=null) {
          r=uaLayer.getInformation(opt);
          if (r!=null) {
              return r;
          }
      }
      if (connLayer!=null) {
          r=connLayer.getInformation(opt);
          if (r!=null) {
              return r;
          }
      }
      return null;
  }                    
  
  private String DP() {
      return DP_S+"["+Thread.currentThread().hashCode()+"]";
  }
  
}