package com.monica.javacom.ssh2.connection;

import com.monica.javacom.ssh2.interfaces.*;
import com.monica.javacom.ssh2.*;
import com.monica.interfaces.*;
import com.monica.util.debug.*;

import java.util.*;
import java.math.*;

public abstract class SSH2BasicConnection implements SSH2ConnectionLayer {
    
    private static final String DP="SSH2BasicConnection";
    
    protected static final String NAME="SSH2BasicConnection";
    protected static final String VERSION="0.31";
    
    protected SSH2Tokenizer tokenizer;
    protected SSH2BothSideFeatures owner;
    protected int side;
    protected Configuration configuration;
    protected Object param;
    protected boolean connected;
    protected SSH2IO io;
    protected int curID;
    protected Hashtable channels;
    protected SSH2ChannelManager channelManager;
    
    // --------------------------------------------------------
    // SSH2BasicConnection
    // --------------------------------------------------------

    public void _init() {
        io=new SSH2IO();
        curID=-1;
        channels=new Hashtable();
    }
    
    private void setConnected() {
        if ((tokenizer!=null) && (side!=SSH2Constants.SSH_SIDE_NOT_SET) &&
            (configuration!=null)) {
            connected=true;
            return;
        }
        connected=false;
    }
    
    protected synchronized int getNewChannelID() {
      curID++;
      return curID;
    }
    
    protected void newChannel(int ID, SSH2Channel channel) {
      channels.put(""+ID, channel);
    }
    
    protected SSH2Channel getChannel(int ID) {
      Object o=channels.get(""+ID);
      if (o==null) {
        return null;
      }      
      return (SSH2Channel)o;
    }
    
    protected void removeChannel(int ID) {
        SSH2Channel ch=getChannel(ID);
        if (ch!=null) {
            ch.closeConnections();
            channels.remove(""+ID);
        }
    }
    
    // --------------------------------------------------------
    // SSH2ConnectionLayer
    // --------------------------------------------------------
    
    public String getConnectionLayerName() {
        return NAME;
    }
    
    public String getConnectionLayerVersion() {
        return VERSION;
    }
    
    public void setChannelManager(SSH2ChannelManager chManager) {
      channelManager=chManager;
    }
    
    public SSH2ChannelManager getChannelManager() {
      return channelManager;
    }
    
    public void channelClosed(SSH2Channel channel) {
        channels.remove(""+channel.getOwnID());
    }
    
    public abstract void start();
    
    // GLOBAL REQUEST
    
    public abstract void processConnectionGlobalRequest(SSH2Packet pack);    
    public abstract void processConnectionRequestSuccess(SSH2Packet pack);
    public abstract void processConnectionRequestFauilure(SSH2Packet pack);
    
    // CHANNEL OPEN
    
    public abstract void processConnectionChannelOpen(SSH2Packet pack);    
    public abstract void processConnectionChannelOpenConfirmation(SSH2Packet pack);    
    public abstract void processConnectionChannelOpenFailure(SSH2Packet pack);
    
    // WINDOW ADJUST
    
    public void processConnectionChannelWindowAdjust(SSH2Packet pack) {
      Vector data=io.parsePayload(pack.payload, SSH2Constants.CONN_CHANNEL_WINDOW_ADJUST_MSG_DATA_DESC);
      int channelNo=((BigInteger)data.elementAt(SSH2Constants.DI_CONN_CHANNEL_WINDOW_ADJUST_CHANNELNO)).intValue();
      int bytes=((BigInteger)data.elementAt(SSH2Constants.DI_CONN_CHANNEL_WINDOW_ADJUST_BYTES)).intValue();
      SSH2Channel channel=getChannel(channelNo);
      if (channel!=null) {
        channel.modifyWindowSize(bytes);
      }
      else {
        tokenizer.closeConnections();
      }
    }
    
    // DATA ARRIVED
    
    public void processConnectionChannelData(SSH2Packet pack) {
      Vector data=io.parsePayload(pack.payload, SSH2Constants.CONN_CHANNEL_DATA_MSG_DATA_DESC);
      int channelNo=((BigInteger)data.elementAt(SSH2Constants.DI_CONN_CHANNEL_DATA_CHANNELNO)).intValue();
      String dataS=(String)data.elementAt(SSH2Constants.DI_CONN_CHANNEL_DATA_DATA);
      SSH2Channel channel=getChannel(channelNo);
      if (channel!=null) {
        channel.data(dataS);
      }
      else {
        tokenizer.closeConnections();
      }
    }
    
    public void processConnectionChannelExtendedData(SSH2Packet pack) {
      Vector data=io.parsePayload(pack.payload, SSH2Constants.CONN_CHANNEL_EXTENDED_DATA_MSG_DATA_DESC);
      int channelNo=((BigInteger)data.elementAt(SSH2Constants.DI_CONN_CHANNEL_EXTENDED_DATA_CHANNELNO)).intValue();
      String dataS=(String)data.elementAt(SSH2Constants.DI_CONN_CHANNEL_EXTENDED_DATA_DATA);
      SSH2Channel channel=getChannel(channelNo);
      if (channel!=null) {
        channel.extendedData(dataS);
      }
      else {
        tokenizer.closeConnections();
      }
    }

    // CHANNEL END
    
    public abstract void processConnectionChannelEOF(SSH2Packet pack);
    
    public void processConnectionChannelClose(SSH2Packet pack) {
      int channelNo=io.rUint32Int(pack.payload.substring(1));
      SSH2Channel channel=getChannel(channelNo);
      if (channel!=null) {
        channel.closeConnections();
      }
      else {
        tokenizer.closeConnections();
      }      
    }

    // CHANNEL REQUEST
    
    public void processConnectionChannelRequest(SSH2Packet pack) {
      int channelNo=io.rUint32Int(pack.payload.substring(1));
      SSH2Channel channel=getChannel(channelNo);
      if (channel!=null) {
        channel.channelRequest(pack);
      }
      else {
        tokenizer.closeConnections();
      }      
    }
    
    public void processConnectionChannelSuccess(SSH2Packet pack) {
      int channelNo=io.rUint32Int(pack.payload.substring(1));
      SSH2Channel channel=getChannel(channelNo);
      if (channel!=null) {
        channel.channelSuccess(pack);
      }
      else {
        tokenizer.closeConnections();
      }             
    }
    
    public void processConnectionChannelFailure(SSH2Packet pack) {
      int channelNo=io.rUint32Int(pack.payload.substring(1));
      SSH2Channel channel=getChannel(channelNo);
      if (channel!=null) {
        channel.channelFailure(pack);
      }
      else {
        tokenizer.closeConnections();
      }              
    }
    
    // --------------------------------------------------------
    // SSH2Layer
    // --------------------------------------------------------
        
    public void setTokenizer(SSH2Tokenizer tokenizer) {
        this.tokenizer=tokenizer;
        setConnected();
    }
    
    public SSH2Tokenizer getTokenizer() {
        return tokenizer;
    }
        
    public void setParam(Object o) throws Exception {
        param=o;
        setConnected();
    }
    
    public void setOwner(SSH2BothSideFeatures owner) {
        this.owner=owner;
        setConnected();
    }

    public SSH2BothSideFeatures getOwner() {
        return owner;
    }
    
    public void setSide(int side) {
        this.side=side;
        setConnected();
    }
        
    public int getSide() {
        return side;
    }
    
    public String getInformation(int type) {
	return null;
    }    
                
    // --------------------------------------------------------
    // Closeable
    // --------------------------------------------------------
    
    public String getCloseableName() {
        return NAME;
    }
    
    public String getCloseableVersion() {
        return VERSION;
    }
    
    public void closeConnections() {
        DebugManager.logOut(DP, "Closing connections...");
        Enumeration e=channels.elements();
        SSH2Channel ch;
        while (e.hasMoreElements()) {
            ch=(SSH2Channel)(e.nextElement());
            ch.closeConnections();
            DebugManager.logOut(DP, "...Channel ID="+ch.getOwnID()+" closed.");
        }
        connected=false;
        tokenizer.closeConnections();
    }
        
    // --------------------------------------------------------
    // Configureable
    // --------------------------------------------------------
           
    public void setConfiguration(Configuration configuration) {
        this.configuration=configuration;
        _rereadConfig();
    }
    
    public Configuration getConfiguration() {
        return configuration;
    }
    
    private void _rereadConfig() {
        _rereadConfigStub();
    }
    
    protected void _rereadConfigStub() {
    }
        
}
