package com.monica.javacom.ssh2.userauthentication;

import com.monica.interfaces.*;
import com.monica.javacom.interfaces.*;
import com.monica.javacom.ssh2.*;
import com.monica.javacom.ssh2.interfaces.*;
import com.monica.util.*;
import com.monica.util.hash.*;
import com.monica.util.configuration.*;
// import com.monica.javacom.ssh2.cipher.dss.*;
import com.monica.util.ccdm.dss.*;
import com.monica.util.debug.*;

import java.util.*;

public class SSH2DSSPublicKeyUA implements SSH2UserAuthenticationLayer {
    
  private static final String DP="SSH2DSSPublicKeyUA";  

  private static final String NAME="SSH2DSSPublicKeyUA";
  private static final String VERSION="0.31";
  private static final String METHODNAME="publickey";

  private static final String ERR_NOT_CONNECTED="Not connected.";
  private static final String ERR_NOT_UA_SERVICE_ACCEPT="Not an ssh-userauth service accept msg.";
  
  private Configuration configuration;
  private SSH2Tokenizer tokenizer;
  private int side;
  private boolean connected;
  private String lastError;
  private SSH2IO io;
  private SSH2BothSideFeatures owner;
  private DSS dss;
  private SSHDSS sshdss;
  private SHA1 sha1;
  private Object myParam;
  private String userName;
  private boolean canContinue;
  private SSH2KeyManager keyManager;

  public SSH2DSSPublicKeyUA() {
    lastError=new String();
    connected=false;
    side=SSH2Constants.SSH_SIDE_NOT_SET;
    io=new SSH2IO();
    sha1=new SHA1();
    dss=new DSS();
    sshdss=new SSHDSS();
    sshdss.setDSS(dss);
    canContinue=true;
  }

  // **************************************************************************************
  // Closeable
  // **************************************************************************************

  public String getCloseableName() {
    return NAME;
  }

  public String getCloseableVersion() {
    return VERSION;
  }

  public void closeConnections() {
    connected=false;
    tokenizer.closeConnections();
  }

  // **************************************************************************************
  // Configureable
  // **************************************************************************************

  public void setConfiguration(Configuration configuration) {
    this.configuration=configuration;
    rereadConfig();
  }

  public Configuration getConfiguration() {
    return configuration;
  }

  private void rereadConfig() {      
    try {
      keyManager=new SSH2KeyManager(configuration);  
      dss=keyManager.getOwn();
      sshdss=new SSHDSS();
      sshdss.setDSS(dss);
      userName=keyManager.getOwnInfo();
      setConnected();
    }
    catch (Exception e) {
      DebugManager.logException(DP, e);  
    }        
  }

  // **************************************************************************************
  // SSH2Layer
  // **************************************************************************************

  public void setTokenizer(SSH2Tokenizer tokenizer) {
//    DebugManager.logOut(DP, "setTokenizer");  
    this.tokenizer=tokenizer;
    setConnected();
  }

  public SSH2Tokenizer getTokenizer() {
    return tokenizer;
  }

  public void setSide(int side) {
//    DebugManager.logOut(DP, "setSide");  
    this.side=side;
    setConnected();
  }

  public int getSide() {
    return side;
  }

  public String methodName() {
    return METHODNAME;
  }

  public void setParam(Object o) throws Exception {
    myParam=o;  
    setConnected();
  }

  public void setOwner(SSH2BothSideFeatures owner) {
    this.owner=owner;
  }

  public SSH2BothSideFeatures getOwner() {
    return owner;
  }

  // **************************************************************************************
  // SSH2UserAuthenticationLayer
  // **************************************************************************************

  public String getUserAuthenticationLayerName() {
    return NAME;
  }

  public String getUserAuthenticationLayerVersion() {
    return VERSION;
  }

  public void start() throws Exception {
    if (!connected) {
      throwException(ERR_NOT_CONNECTED);
    }
    switch (side) {
      case SSH2Constants.SSH_SIDE_CLIENT :          
        requestMethod();  
        break;
      case SSH2Constants.SSH_SIDE_SERVER :
        break;
    }
  }
  
  private void requestMethod() throws Exception {
    String req=io.wByte(SSH2Constants.SSH_USERAUTH_MSG_USERAUTH_REQUEST);
    req+=io.wString(userName);
    req+=io.wString("ssh-connection");
    req+=io.wString("publickey");
    req+=io.wBoolean(false);
    req+=io.wString("ssd-dss");
    req+=io.wString(sshdss.ssh2_pubKey());
    tokenizer.sendP(req);
  }

  public void processUserAuthRequestMsg(SSH2Packet pack) throws Exception {
    if (!connected) {
      throwException(ERR_NOT_CONNECTED);
    }
    switch (side) {
      case SSH2Constants.SSH_SIDE_CLIENT :
        tokenizer.sendUnimplementedMsg(pack);
        break;
      case SSH2Constants.SSH_SIDE_SERVER :
        Vector data;
        String userConfName, userPubKey;
        boolean type;
        
        data=io.parsePayload(pack.payload,SSH2Constants.PUBKEY_USERAUTH_REQUEST_MSG_DATA_DESC);                          
        type=((Boolean)data.elementAt(SSH2Constants.PUBKEY_USERAUTH_REQUEST_BOOLEAN)).booleanValue();
        userName=(String)data.elementAt(SSH2Constants.PUBKEY_USERAUTH_REQUEST_USERNAME);
        userPubKey=(String)data.elementAt(SSH2Constants.PUBKEY_USERAUTH_REQUEST_ALG_BLOB);        
        
        dss=keyManager.getUserPublic(userName);
        sshdss.setDSS(dss);
                
        if (type) { // second message
            data=io.parsePayload(pack.payload,SSH2Constants.PUBKEY_USERAUTH_REQUEST_MSG_DATA_DESC_2);                          
            String signature=(String)data.elementAt(SSH2Constants.PUBKEY_USERAUTH_REQUEST_SIGNATURE);
            String toSign=createToSign();
            if (sshdss.ssh2_verify(toSign, signature)) {
                tokenizer.userAuthSuccess();
                tokenizer.userAuthenticationProtocolReady();
            }
            else {
                canContinue=false;
                tokenizer.userAuthFailure();
            }

        }
        else { // first message                        
            if (userPubKey.equals(sshdss.ssh2_pubKey())) {
                String reply=new String();
                reply=io.wByte(SSH2Constants.SSH_USERAUTH_MSG_USERAUTH_PK_OK);
                reply+=io.wString("ssh-dss");
                reply+=io.wString(userPubKey);
                tokenizer.sendP(reply);
            }
            else {
                canContinue=false;
                tokenizer.userAuthFailure();
            }
        }
    }
  }
  
  public void processUserAuthPkOkMsg(SSH2Packet pack) throws Exception {
    if (!connected) {
      throwException(ERR_NOT_CONNECTED);
    }
    switch (side) {
      case SSH2Constants.SSH_SIDE_SERVER :      
        tokenizer.sendUnimplementedMsg(pack);
        break;        
      case SSH2Constants.SSH_SIDE_CLIENT :          
        String toSign=createToSign();           
        String signature=sshdss.ssh2_sign(toSign);
        String req=io.wByte(SSH2Constants.SSH_USERAUTH_MSG_USERAUTH_REQUEST);
        req+=io.wString(userName);
        req+=io.wString("ssh-connection");
        req+=io.wString("publickey");
        req+=io.wBoolean(true);
        req+=io.wString("ssd-dss");
        req+=io.wString(sshdss.ssh2_pubKey());
        req+=io.wString(signature);
        tokenizer.sendP(req);                
        break;
    }
  }

  private String createToSign() {
      String toSign;
      toSign=io.wString(tokenizer.getSessionID());
      toSign+=io.wByte(SSH2Constants.SSH_USERAUTH_MSG_USERAUTH_REQUEST);
      toSign+=io.wString(userName);
      toSign+=io.wString("ssh-connection");
      toSign+=io.wString("publickey");
      toSign+=io.wBoolean(true);
      toSign+=io.wString("ssd-dss");
      toSign+=io.wString(sshdss.ssh2_pubKey());
      return toSign;
  }

  public void processUserAuthSuccessMsg(SSH2Packet pack) throws Exception {
    if (!connected) {
      throwException(ERR_NOT_CONNECTED);
    }
    switch (side) {
      case SSH2Constants.SSH_SIDE_CLIENT :
        tokenizer.userAuthenticationProtocolReady();
        break;
      case SSH2Constants.SSH_SIDE_SERVER :
        tokenizer.sendUnimplementedMsg(pack);
        break;
    }
  }

  public void processUserAuthFailureMsg(SSH2Packet pack) throws Exception {
    String possibleContinuation;
    if (!connected) {
      throwException(ERR_NOT_CONNECTED);
    }
    switch (side) {
      case SSH2Constants.SSH_SIDE_CLIENT :
        DebugManager.logOut(DP, "Connection failed.");
        tokenizer.closeConnections();
        break;
      case SSH2Constants.SSH_SIDE_SERVER :
        tokenizer.sendUnimplementedMsg(pack);
        break;
    }
  }

  public void processUserAuthBannerMsg(SSH2Packet pack) throws Exception {
    String temp;
    if (!connected) {
      throwException(ERR_NOT_CONNECTED);
    }
    switch (side) {
      case SSH2Constants.SSH_SIDE_CLIENT :
        tokenizer.sendUnimplementedMsg(pack);
        break;
      case SSH2Constants.SSH_SIDE_SERVER :
        tokenizer.sendUnimplementedMsg(pack);
        break;
    }
  }
  
  private void setConnected() {
//    DebugManager.logOut(DP, tokenizer.toString());
//    DebugManager.logOut(DP, ""+side);
//    DebugManager.logOut(DP, configuration.toString());
//    DebugManager.logOut(DP, dss.getPrivate().toString());
    if ((tokenizer!=null) && (side!=SSH2Constants.SSH_SIDE_NOT_SET) &&
        (configuration!=null) && (dss!=null)) {
        connected=true;
        return;
    }
    connected=false;
  }

  private void throwException(String s) throws Exception {
    lastError=s;
    throw new Exception(s);
  }

  public boolean canContinue() {
      return canContinue;
  }
  
  public String getInformation(int opt) {
      switch (opt) {
          case SSH2Constants.INFO_USERAUTHENTICATION_PARTNER_USERNAME :
              return userName;
          case SSH2Constants.INFO_USERAUTHENTICATION_METHOD :
              return METHODNAME;
          default :
              return null;
      }
  }
              
}