package com.monica.javacom.terminal;

import com.monica.javacom.interfaces.*;
import com.monica.interfaces.*;
import com.monica.javacom.virtualdisplay.*;
import com.monica.util.debug.*;

/** Class for providing terminal functionality according to the VT 220 standard.
    <br><br>
    This class handles the following incoming character sequences:
    <ul>
      <li><b>0-NULL</b>: nothing
      <li><b>7-BELL</b>: nothing
      <li><b>8-BS</b>: backspace
      <li><b>10-LF</b>: new line
      <li><b>13-CR</b>: home
      <li><b>14-SO</b>: G0 charset
      <li><b>15-SI</b>: G1 charset
      <li><b>27 40 48</b>: G0 charset
      <li><b>27 41 48</b>: G1 charset
      <li><b>27 61</b>: keypad application mode
      <li><b>27 62</b>: keypad numeric mode
      <li><b>27 68</b>: Scroll down
      <li><b>27 69</b>: Scroll up and go home
      <li><b>27 77</b>: Scroll up
      <li><b>27 91 52 105</b>: IRM set to insert
      <li><b>27 91 52 108</b>: IRm set to overwrite
      <li><b>27 91 55 109</b>: set to inverting
      <li><b>27 91 63 49 104</b>: cursor key mode set width
      <li><b>27 91 63 49 108</b>: cursor key mode reset width
      <li><b>27 91 63 55 104</b>: wraparound set width
      <li><b>27 91 63 55 108</b>: wraparound reset width
      <li><b>27 91 Pn 65</b>: Pn Up
      <li><b>27 91 Pn 66</b>: Pn Down
      <li><b>27 91 Pn 67</b>: Pn Right
      <li><b>27 91 Pn 68</b>: Pn Left
      <li><b>27 91 72</b>: go to origo
      <li><b>27 91 Pl 59 Pc 72</b>: Goto Pl line Pc column
      <li><b>27 91 74</b>: clear screen
      <li><b>27 91 75</b>: erase to end of line
      <li><b>27 91 Pn 76</b>: insert Pn lines
      <li><b>27 91 Pn 77</b>: delete Pn lines
      <li><b>27 91 Pn 80</b>: delete Pn char
      <li><b>27 91 109</b>: reset inverting
      <li><b>27 91 Pl 59 Pl 114</b>: set scrolling region between Pl and Pl lines
    </ul>
    This class handles the keys:
    <ul>
      <li><b>ASCII 32-127</b>
      <li><b>TAB</b>
      <li><b>ENTER</b>
      <li><b>BACKSPACE</b>
      <li><b>HOME, END</b>
      <li><b>PGUP, PGDN</b>
      <li><b>INS, DEL</b>
      <li><b>UP, DOWN, LEFT, RIGHT</b>
      <li><b>NUMPAD +,/,*,-</b>
      <li><b>F1-F10</b>
    </ul>
    Configureable values from the Configuration <b>JavaCom.terminal</b>:
    <ul>
      <li><b>terminalReadBuffer</b>: the maximum size of the terminal read buffer
    </ul>
*/
public class VT220Terminal implements Terminal {

  // ***************************************************************************************
  // Variables
  // ***************************************************************************************
  private int tabSize;
  private int[] readBuffer;
  private int readBufferPos;
  private int readBufferSize;
  private int processStatus;

  private int tempInt_aC;
  private int tempInt_gN;
  private int tempInt_rE;
  private int tempInt_DN_i;
  private int tempInt_DN_s;
  private int tempInt_DN_n1;
  private int tempInt_DN_n2;
  private int tempInt_gSP;
  private int tempInt_cRB;

  private String input;

  private Configuration configuration;
  private VirtualDisplay display;
  private Readable readable;
  private Writeable writeable;

  private int charset;
  private int keypad;
  private int IRMstatus;

  private static final int CS_G0=0;
  private static final int CS_G1=1;

  private static final int KP_APP=0;
  private static final int KP_NUM=1;

  private static final int IRM_OVR=0;
  private static final int IRM_INS=1;

  // ***************************************************************************************
  // Constants
  // ***************************************************************************************
  
  private static final String DP="VT220Terminal";
  
  private static final String TYPE="vt220";
  
  private static final String CONF_STREAM_NAME="JavaCom.terminal";

  // Configuration
  private static final String CONF_TABSIZE="displayTabSize";
  private static final String CONF_READ_BUFFER_SIZE="terminalReadBufferSize";

  // Default values
  private static final int DEF_TABSIZE=8;
  private static final int DEF_READ_BUFFER_SIZE=10000;

  private static final String NAME="VT220Terminal";
  private static final String VERSION="0.31";

  private static final int PS_NO=0;
  private static final int PS_DUALNUMBER=1;

  private int[] i27={27};
  private int[] i2791={27, 91};
  private int[] i279155={27, 91, 55};
  private int[] i2740={27, 40};
  private int[] i2741={27, 41};
  private int[] i13={13};
  private int[] i279163={27, 91, 63};
  private int[] i27916349={27, 91, 63, 49};
  private int[] i27916355={27, 91, 63, 55};
  private int[] i279149={27, 91, 49};
  private int[] i279152={27, 91, 52};

  // ***************************************************************************************
  // Constructors
  // ***************************************************************************************
  public VT220Terminal() {
    processStatus=PS_NO;
    charset=CS_G0;
    keypad=KP_NUM;
    IRMstatus=IRM_OVR;
  }

  private void updateConfig() {
    tabSize=configuration.getInt(CONF_TABSIZE, DEF_TABSIZE);
    readBufferSize=configuration.getInt(CONF_READ_BUFFER_SIZE, DEF_READ_BUFFER_SIZE);
    readBuffer=new int[readBufferSize];
    emptyReadBuffer();
  }
  
  // ***************************************************************************************
  // LayerBottomView
  // ***************************************************************************************
  
  public String getLayerBottomViewName() {
    return NAME;
  }
  
  public String getLayerBottomViewVersion() {
    return VERSION;
  }        

  // ***************************************************************************************
  // Configureable
  // ***************************************************************************************

  /** Sets the Configuration for this object.
      @param Configuration configuration: the Configuration
        object for the customization
  */
  public void setConfiguration(Configuration configuration) {
      try {
          this.configuration=configuration.getConfiguration(CONF_STREAM_NAME);          
          updateConfig();
      }
      catch (Exception e) {
      }
  }

  /** Returns the Configuration with which this object is customized.
      @return Configuration: the Configuration with which this
        object is customized
  */
  public Configuration getConfiguration() {
    return configuration;
  }

  // ***************************************************************************************
  // Reader
  // ***************************************************************************************

  public String getReaderName() {
    return NAME;
  }

  public String getReaderVersion() {
    return VERSION;
  }

  public void readEvent(Readable readable) {
    input=readable.getInput();
    for (tempInt_rE=0; tempInt_rE<input.length(); tempInt_rE++) {
      add(input.charAt(tempInt_rE));
    }
    display.refresh();
  }

  public void connectionClosedEvent(Readable readable) {
    display.setInvert(false);
    display.clear();
    display.setCPos(0,0);
    display.addS("Connection closed.");
    display.refresh();
  }

  public void errorEvent(String errorDescription, Readable readable) {
  }

  public void setReadable(Readable readable) {
    this.readable=readable;
  }

  public Readable getReadable() {
    return readable;
  }

  // ***************************************************************************************
  // Writeable
  // ***************************************************************************************

  public String getWriteableName() {
    return NAME;
  }

  public String getWriteableVersion() {
    return VERSION;
  }

  public boolean write(char c) {
    return false;
  }

  public boolean write(String s) {
    switch ((int)s.charAt(0)) {

      case 0 :                                        //ascii 0-127

        switch ((int)s.charAt(1)) {
          case 8 : // BACKSPACE
            if (keypad==KP_NUM) {
              writeable.write((char)127);
            }
            else {
              writeable.write((char)8);
            }
            break;

          case 9 : // Tab
            writeable.write((char)9);
            break;

          case 10 : // Enter
            //writeable.write((char)10);
            writeable.write((char)13);
            break;

          default :
            writeable.write(s.charAt(1));
            break;
        }
        break;

      case 4 :                                        //"control" billentyuk
        switch ((int)s.charAt(1)) {

          case 59 : // f1
            writeable.write((char)27);
            writeable.write('1');
            break;

          case 60 : // f2
            writeable.write((char)27);
            writeable.write('2');
            break;

          case 61 : // f3
            writeable.write((char)27);
            writeable.write('3');
            break;

          case 62 : // f4
            writeable.write((char)27);
            writeable.write('4');
            break;

          case 63 : // f5
            writeable.write((char)27);
            writeable.write('5');
            break;

          case 64 : // f6
            writeable.write((char)27);
            writeable.write('6');
            break;

          case 65 : // f7
            writeable.write((char)27);
            writeable.write('7');
            break;

          case 66 : // f8
            writeable.write((char)27);
            writeable.write('8');
            break;

          case 67 : // f9
            writeable.write((char)27);
            writeable.write('9');
            break;

          case 68 : // f10
            writeable.write((char)27);
            writeable.write('0');
            break;

          case 71 : // home
            writeable.write((char)1);
            break;

          case 72 : // up
            writeable.write((char)27);
            writeable.write((char)91);
            writeable.write('A');
            break;

          case 73 : // pgup
            writeable.write((char)27);
            writeable.write((char)91);
            writeable.write((char)53);
            writeable.write((char)126);
            break;

          case 75 : // left
            writeable.write((char)27);
            writeable.write((char)91);
            writeable.write('D');
            break;

          case 77 : // right
            writeable.write((char)27);
            writeable.write((char)91);
            writeable.write('C');
            break;

          case 79 : // end
            writeable.write((char)5);
            break;

          case 80 : // Down
            writeable.write((char)27);
            writeable.write((char)91);
            writeable.write('B');
            break;

          case 81 : // Pgdn
            writeable.write((char)27);
            writeable.write((char)91);
            writeable.write((char)54);
            writeable.write((char)126);
            break;

          case 82 : // Insert
            writeable.write((char)27);
            writeable.write((char)91);
            writeable.write((char)50);
            writeable.write((char)126);
            break;

          case 83 : // delete
            writeable.write((char)27);
            writeable.write((char)91);
            writeable.write((char)51);
            writeable.write((char)126);
            break;

          case 128 : // +
            writeable.write('+');
            break;

          case 129 : // /
            writeable.write('/');
            break;

          case 130 : // *
            writeable.write('*');
            break;

          case 131 : // /
            writeable.write('-');
            break;

        }
        break;
    }
    return true;
  }

  public boolean write(char[] ca) {
    return false;
  }

  public void flush() {
    if (writeable!=null) {
      writeable.flush();
    }
  }

  public String getLastWriteableError() {
    if (writeable!=null) {
      return writeable.getLastWriteableError();
    }
    return new String();
  }

  public boolean isWriteableConnected() {
    if (writeable!=null) {
      return writeable.isWriteableConnected();
    }
    return false;
  }

  // ***************************************************************************************
  // Writer
  // ***************************************************************************************

  public String getWriterName() {
    return NAME;
  }

  public String getWriterVersion() {
    return VERSION;
  }

  public void setWriteable(Writeable writeable) {
    this.writeable=writeable;
  }

  public Writeable getWriteable() {
    return writeable;
  }

  // ***************************************************************************************
  // Terminal
  // ***************************************************************************************

  public String getTerminalName() {
    return NAME;
  }

  public String getTerminalVersion() {
    return VERSION;
  }

  public void setVirtualDisplay(VirtualDisplay display) {
    this.display=display;
  }

  public VirtualDisplay getVirtualDisplay() {
    return display;
  }

  private void add(char c) {
//    DebugManager.logOut(DP, "readBufferPos="+readBufferPos+", pS="+processStatus+" : "+(int)c+" : "+c);
    if (readBufferPos==0) {
      try {
        switch (c) {
          case 0 : // NULL
            emptyReadBuffer();
            break;
          case 7 : // BELL
            emptyReadBuffer();
//            DebugManager.logOut(DP, "Bell");
            break;
          case 8 : // BS
            emptyReadBuffer();
            display.goLeft();
//            DebugManager.logOut(DP, "backspace");
            break;
	        case 9 : // TAB
	          tempInt_aC=display.getCPosX();
            display.goRight(tabSize-(tempInt_aC%tabSize));
            emptyReadBuffer();
//            DebugManager.logOut(DP, "tab");
            break;
	        case 10 : // LF
            emptyReadBuffer(); //--//
	          display.newLine();
//            DebugManager.logOut(DP, "newline");
	          break;
          case 13 : // CR
            emptyReadBuffer();
            display.setCPosX(0); //--//
//            DebugManager.logOut(DP, "CR - carriage return");
            break;
          case 14 : // SO - invoke G1 charset
            emptyReadBuffer();
//            DebugManager.logOut(DP, "SO");
            charset=CS_G1;
            break;
          case 15 : // SI - invoke G0 charset
            emptyReadBuffer();
//            DebugManager.logOut(DP, "SI");
            charset=CS_G0;
            break;
	        case 27 : // ESC
	          addToReadBuffer(c);
	          break;
	  default :
            emptyReadBuffer();
            switch (charset) {
              case CS_G0 :
  	        addChar(c);
                break;
              case CS_G1 :
                switch (c) {
                  case 'q' : // horizontal line
                    addChar(display.translateSpecial(VirtualDisplayConstants.HORIZ_LINE));
                    break;
                  case 'x' : // vertical line
                    addChar(display.translateSpecial(VirtualDisplayConstants.VERT_LINE));
                    break;
                  case 'j' : // lower right corner
                    addChar(display.translateSpecial(VirtualDisplayConstants.LR_CORNER));
                    break;
                  case 'k' : // upper right corner
                    addChar(display.translateSpecial(VirtualDisplayConstants.UR_CORNER));
                    break;
                  case 'l' : // upper left corner
                    addChar(display.translateSpecial(VirtualDisplayConstants.UL_CORNER));
                    break;
                  case 'm' : // lower left corner
                    addChar(display.translateSpecial(VirtualDisplayConstants.LL_CORNER));
                    break;
                  case 't' : // |-
                    addChar(display.translateSpecial(VirtualDisplayConstants.HORIZ_VERT_LEFT));
                    break;
                  case 'u' : // -|
                    addChar(display.translateSpecial(VirtualDisplayConstants.HORIZ_VERT_RIGHT));
                    break;
                  default:
                    addChar(c);
                    break;
                }
                break;
            }
	    break;
	}
      }
      catch (Exception e) {
        DebugManager.logException(DP, e);;
      }
    }
    else {
      processBuffer(c);
    }
  }

  /** Processes a char if there are ones buffered.
      @param char c: the char to be processed
  */
  private void processBuffer(char c) {
    try {
      addToReadBuffer(c);
    }
    catch (Exception e) {
      DebugManager.logException(DP, e);;
      return;
    }

    switch (processStatus) {
      case PS_NO :
        processPSNo(c);
        break;
      case PS_DUALNUMBER:
        processPSDualNumber(c);
        break;
    }
  }

  /** Processes a char if there hasn't been number in the buffer.
      @param char c: the char to be processed
  */
  private void processPSNo(char c) {
    switch ((int)c) {

      case 40 : // (
        if (!checkReadBuffer(i27)) {
          dumpReadBuffer();
          emptyReadBuffer();
          addChar(c);
        }
        break;

      case 41 : // )
        if (!checkReadBuffer(i27)) {
          dumpReadBuffer();
          emptyReadBuffer();
          addChar(c);
        }
        break;

      case 48 : // 0
        if (checkReadBuffer(i2740)) {
          // Graphics
          emptyReadBuffer();
//          DebugManager.logOut(DP, "Graphics");
          charset=CS_G1;
        }
        else if (checkReadBuffer(i2741)) {
          emptyReadBuffer();
//          DebugManager.logOut(DP, "Graphics G1");
          charset=CS_G1;
        }
        else if ((checkReadBuffer(i2791)) || (wasPrevNumber())) {
          processStatus=PS_DUALNUMBER;
        }
        else {
          dumpReadBuffer();
          emptyReadBuffer();
          addChar(c);
        }
        break;

      case 49 : // 1
        if ((wasPrevNumber()) || (checkReadBuffer(i2791))) {
          processStatus=PS_DUALNUMBER;
        }
        else if (!checkReadBuffer(i279163)) {
          dumpReadBuffer();
          emptyReadBuffer();
          addChar(c);
        }
        break;

      case 52 : // 4
        if (wasPrevNumber()) {
	        processStatus=PS_DUALNUMBER;
	      }
	      else if (!checkReadBuffer(i2791)) {
	        dumpReadBuffer();
	        emptyReadBuffer();
	        addChar(c);
	      }
	      break;

      case 55: // 7
        if (wasPrevNumber()) {
          processStatus=PS_DUALNUMBER;
        }
        else if ((!checkReadBuffer(i2791)) && (!checkReadBuffer(i279163))) {
          dumpReadBuffer();
          emptyReadBuffer();
          addChar(c);
        }
        break;

      case 59: // ;
        if (wasPrevNumber()) {
	        processStatus=PS_DUALNUMBER;
	      }
	      else {
          dumpReadBuffer();
          emptyReadBuffer();
          addChar(c);
        }
        break;

      case 61 : // =
        if (!checkReadBuffer(i27)) {
          dumpReadBuffer();
          emptyReadBuffer();
          addChar(c);
        }
        else {
          // keypad application
          keypad=KP_APP;
//          DebugManager.logOut(DP, "keypad application");
          emptyReadBuffer();
        }
        break;

      case 62 : // >
        if (!checkReadBuffer(i27)) {
          dumpReadBuffer();
          emptyReadBuffer();
          addChar(c);
        }
        else {
          // keypad numeric mode
          keypad=KP_NUM;
//          DebugManager.logOut(DP, "keypad numeric mode");
          emptyReadBuffer();
        }
        break;

      case 63 : // ?
        if (!checkReadBuffer(i2791)) {
          dumpReadBuffer();
          emptyReadBuffer();
          addChar(c);
        }
        break;

      case 64 : // @
        if (checkReadBuffer(i2791)) { // insert character
	        emptyReadBuffer();
	        display.insertChar();
          //	  DebugManager.logOut(DP, "insert character");
	      }
	      else {
	        dumpReadBuffer();
          emptyReadBuffer();
          addChar(c);
        }
        break;

      case 65 : // A
        if (checkReadBuffer(i2791)) {
          emptyReadBuffer();
          display.goUp();
//          DebugManager.logOut(DP, "goup");
        }
        else if (wasPrevNumber()) {
          display.goUp(readBuffer[readBufferPos-2]-48);
          emptyReadBuffer();
//          DebugManager.logOut(DP, "goup No.");
        }
        else {
          dumpReadBuffer();
          emptyReadBuffer();
          addChar(c);
        }
        break;

      case 66 : // B
        if (checkReadBuffer(i2791)) {
          emptyReadBuffer();
          display.goDown();
//          DebugManager.logOut(DP, "godown");
        }
        else if (checkReadBuffer(i2740)) {
          emptyReadBuffer();
//          DebugManager.logOut(DP, "United States G0");
          charset=CS_G0;
        }
        else if (wasPrevNumber()) {
          display.goDown(readBuffer[readBufferPos-2]-48);
          emptyReadBuffer();
//          DebugManager.logOut(DP, "godown No.");
        }
        else {
          dumpReadBuffer();
          emptyReadBuffer();
          addChar(c);
        }
        break;

      case 67 : // C
        if (checkReadBuffer(i2791)) {
          emptyReadBuffer();
          display.goRight();
//          DebugManager.logOut(DP, "goright");
        }
        else if (wasPrevNumber()) {
          display.goRight(readBuffer[readBufferPos-2]-48);
          emptyReadBuffer();
//          DebugManager.logOut(DP, "goright No.");
        }
        else {
          dumpReadBuffer();
          emptyReadBuffer();
          addChar(c);
        }
        break;

      case 68 : // D
        if (checkReadBuffer(i27)) { // IND
          emptyReadBuffer();
          display.scrollDown();
//          DebugManager.logOut(DP, "IND - Scroll down");
        }
        else if (checkReadBuffer(i2791)) {
          emptyReadBuffer();
          display.goLeft();
//          DebugManager.logOut(DP, "goleft");
        }
        else if (wasPrevNumber()) {
          display.goLeft(readBuffer[readBufferPos-2]-48);
          emptyReadBuffer();
//          DebugManager.logOut(DP, "goleft 7");
        }
        else {
          dumpReadBuffer();
          emptyReadBuffer();
          addChar(c);
        }
        break;

      case 69 : // E
        if (checkReadBuffer(i27)) { // NEL
          display.scrollUp();
          display.setCPosX(0);
          emptyReadBuffer();
//          DebugManager.logOut(DP, "Next Line - Scroll up and home");
        }
        else {
          dumpReadBuffer();
          emptyReadBuffer();
          addChar(c);
        }
        break;

      case 72 : // H
        if (!checkReadBuffer(i2791)) {
          dumpReadBuffer();
          addChar(c);
          emptyReadBuffer();
        }
        else {
          emptyReadBuffer();
          display.setCPos(0,0);
//          DebugManager.logOut(DP, "home"); /////
        }
        break;

      case 74: // J
        if (checkReadBuffer(i2791)) {
          emptyReadBuffer();
//          System.out.println("clear I");
          display.clear();
//          DebugManager.logOut(DP, "clear");
        }
        else {
          dumpReadBuffer();
          emptyReadBuffer();
          addChar(c);
        }
        break;

      case 75: // K
        if (checkReadBuffer(i2791)) {
          emptyReadBuffer();
          display.eraseEOL();
//          DebugManager.logOut(DP, "erase to eol");
        }
        else {
          dumpReadBuffer();
          emptyReadBuffer();
          addChar(c);
        }
        break;

      case 77 : // M
        if (checkReadBuffer(i27)) { // reverse index
          emptyReadBuffer();
          display.scrollUp();
//          DebugManager.logOut(DP, "Reverse Index - Scroll up");
        }
        else {
          dumpReadBuffer();
          emptyReadBuffer();
          addChar(c);
        }
        break;

      case 80 : // P
        if (checkReadBuffer(i2791)) { // delete character
	        emptyReadBuffer();
	        display.deleteChar();
          //	  DebugManager.logOut(DP, "delete character");
	      }
	      else {
	        dumpReadBuffer();
          emptyReadBuffer();
          addChar(c);
        }
        break;

      case 91: // [
        if (!checkReadBuffer(i27)) {
          dumpReadBuffer();
          emptyReadBuffer();
          addChar(c);
        }
        break;

      case 104 : // h
        if (checkReadBuffer(i27916349)) {
          emptyReadBuffer();
//          DebugManager.logOut(DP, "cursor key mode set width");
          //  cursor key mode set width
        }
        else if (checkReadBuffer(i27916355)) {
          emptyReadBuffer();
          display.moveCursor(true);
//          DebugManager.logOut(DP, "wraparound set width");
          // wraparound set width
        }
	      else if (checkReadBuffer(i279152)) { // IRM INS
	        emptyReadBuffer();
	        IRMstatus=IRM_INS;
          //	  DebugManager.logOut(DP, "IRM set to INS");
	}
        else {
          dumpReadBuffer();
          emptyReadBuffer();
          addChar(c);
        }
        break;

      case 108 : // l
        if (checkReadBuffer(i27916349)) {
          emptyReadBuffer();
//          DebugManager.logOut(DP, "cursor ket mode reset width");
          //  cursor key mode reset width
        }
        else if (checkReadBuffer(i27916355)) {
          emptyReadBuffer();
          display.moveCursor(false);
//          DebugManager.logOut(DP, "wraparound reset width");
          // wraparound reset width
        }
	      else if (checkReadBuffer(i279152)) { // IRM OVR
	        emptyReadBuffer();
	        IRMstatus=IRM_OVR;
          //	  DebugManager.logOut(DP, "IRM set to OVR");
	      }
        else {
          dumpReadBuffer();
          emptyReadBuffer();
          addChar(c);
        }
        break;

      case 109: // m
        if (checkReadBuffer(i279155)) {
          emptyReadBuffer();
          display.setInvert(true);
//          DebugManager.logOut(DP, "set invert");
        }
        else if (checkReadBuffer(i2791)) {
          emptyReadBuffer();
          display.setInvert(false);
//          charset=CS_G0;
//          DebugManager.logOut(DP, "reset invert");
        }
        else if (checkReadBuffer(i279152)) {
          emptyReadBuffer();
        }
        else {
          dumpReadBuffer();
          emptyReadBuffer();
          addChar(c);
        }
        break;

      default:
        if ((checkReadBuffer(i2791)) || (wasPrevNumber())) {
          if (((int)c>=48) && ((int)c<=57)) { // 0-9
            processStatus=PS_DUALNUMBER;
            return;
          }
        }
        dumpReadBuffer();
        emptyReadBuffer();
        addChar(c);
        break;
    }
  }

  /** Retruns an in from its the string representation in the stream.
      @param int b: the begin index in the buffer inclusive
      @param int e: the end index in the buffer exclusive
      @return int: the int
  */
  private int getNumber(int b, int e) {
    int tempInt_gN=0;
    for (int i=b; i<e; i++) {
      tempInt_gN*=10;
      tempInt_gN+=readBuffer[i]-48;
    }
    return tempInt_gN;
  }

  /** Returns wether the last char added to the buffer was a number.
      @return boolean: wether the last char added to the buffer was a number
  */
  private boolean wasPrevNumber() {
    if ((readBuffer[readBufferPos-2]>=48) && (readBuffer[readBufferPos-2]<=57)) {
      return true;
    }
    return false;
  }

  /** Processes a char if there are numbers in the buffer.
      @param char c: the char to be processed
  */
  private void processPSDualNumber(char c) {
    switch((int)c) {

      case 59 : // ;
        break;

      case 64 : // @
        processStatus=PS_NO;
	      tempInt_DN_s=getNumber(2,readBufferPos-1);
	      display.insertChar(tempInt_DN_s);
	      DebugManager.logOut(DP, "insert char "+tempInt_DN_s);
	      emptyReadBuffer();
	      break;

      case 65 : // A
        processStatus=PS_NO;
        tempInt_DN_s=getNumber(2, readBufferPos-1);
        display.goUp(tempInt_DN_s);
        emptyReadBuffer();
//        DebugManager.logOut(DP, "up "+tempInt_DN_s);
        break;

      case 66 : // B
        processStatus=PS_NO;
        tempInt_DN_s=getNumber(2, readBufferPos-1);
        display.goDown(tempInt_DN_s);
        emptyReadBuffer();
//        DebugManager.logOut(DP, "down "+tempInt_DN_s);
        break;

      case 67 : // C
        processStatus=PS_NO;
        tempInt_DN_s=getNumber(2, readBufferPos-1);
        display.goRight(tempInt_DN_s);
        emptyReadBuffer();
//        DebugManager.logOut(DP, "right "+tempInt_DN_s);
        break;

      case 68 : // D
        processStatus=PS_NO;
        tempInt_DN_s=getNumber(2, readBufferPos-1);
        display.goLeft(tempInt_DN_s);
        emptyReadBuffer();
//        DebugManager.logOut(DP, "left "+tempInt_DN_s);
        break;

      case 72 : // H
        tempInt_DN_s=getSemiColonPos();
        tempInt_DN_n1=getNumber(2,tempInt_DN_s); // row
        tempInt_DN_n2=getNumber(tempInt_DN_s+1, readBufferPos-1); // col
//        DebugManager.logOut(DP, "setPos row="+tempInt_DN_n1+", col="+tempInt_DN_n2);
        display.setCPos(tempInt_DN_n2-1, tempInt_DN_n1-1);
        processStatus=PS_NO;
        emptyReadBuffer();
        break;

      case 74 : // J
//        System.out.println("clear II");
        display.clear();
        processStatus=PS_NO;
        emptyReadBuffer();
//	DebugManager.logOut(DP, "clear");
        break;

      case 76 : // L
        tempInt_DN_n1=getNumber(2,readBufferPos-1);
	      emptyReadBuffer();
	      display.insertLine(tempInt_DN_n1);
//	DebugManager.logOut(DP, "insert line "+tempInt_DN_n1);
       	processStatus=PS_NO;
	      break;

      case 77 : // M
        tempInt_DN_n1=getNumber(2,readBufferPos-1);
	      emptyReadBuffer();
        display.deleteLine(tempInt_DN_n1);
//	DebugManager.logOut(DP, "delete line "+tempInt_DN_n1);
	      processStatus=PS_NO;
	      break;

      case 80 : // P
        processStatus=PS_NO;
	      tempInt_DN_s=getNumber(2,readBufferPos-1);
	      display.deleteChar(tempInt_DN_s);
	      DebugManager.logOut(DP, "delete char "+tempInt_DN_s);
	      emptyReadBuffer();
	      break;

      case 108 : // l
        // set insert width
        emptyReadBuffer();
        processStatus=PS_NO;
//        DebugManager.logOut(DP, "set insert width");
        break;

      case 109 : // m
        emptyReadBuffer();
        display.setInvert(false);
        processStatus=PS_NO;
//        DebugManager.logOut(DP, "bright");
        break;

      case 114 : // r
        tempInt_DN_s=getSemiColonPos();
        tempInt_DN_n1=getNumber(2,tempInt_DN_s); // firstLine
        tempInt_DN_n2=getNumber(tempInt_DN_s+1, readBufferPos-1); // lastLine
//        DebugManager.logOut(DP, "set scrolling region : firstLine="+tempInt_DN_n1+", lastLine="+tempInt_DN_n2);
        display.setScrollingRegion(tempInt_DN_n1-1, tempInt_DN_n2-1);
        processStatus=PS_NO;
        emptyReadBuffer();
        break;

      default :
        if (((int)c<48) || ((int)c>57)) { // 0-9
          dumpReadBuffer();
          emptyReadBuffer();
          processStatus=PS_NO;
          addChar(c);
        }
        break;
    }
  }

  /** Returns the name of the terminal.
      @return String: the type of the terminal ("vt220")
  */
  public String getType() {
    return TYPE;
  }

  /** Dumps the contents of the buffer to the standard error if the buffer cannot be understood.
  */
  private void dumpReadBuffer() {
    for (int i=0; i<readBufferPos; i++) {
      System.err.println("dRB : "+readBuffer[i]+" - "+(char)readBuffer[i]);
    }
  }

  /** Adds a char to the read buffer.
      @param char c: the char to be added to the buffer
      @exception Exception: by buffer overflow
  */
  private synchronized void addToReadBuffer(char c) throws Exception {
    if (readBufferPos<readBufferSize) {
      readBuffer[readBufferPos]=(int)c;
      readBufferPos++;
    }
    else {
      throw new Exception("VT 220 Read Buffer owerflow");
    }
  }

  /** Deletes the contents of the read buffer.
  */
  private synchronized void emptyReadBuffer() {
    // readBuffer=new int[readBufferSize];
    readBufferPos=0;
    processStatus=PS_NO;
  }

  /** Returns the first occurence of a semicolon in the read buffer.
      @return int: the first occurence of a semicolon in the read buffer
  */
  private int getSemiColonPos() {
    for (tempInt_gSP=0; tempInt_gSP<readBufferSize; tempInt_gSP++) {
      if (readBuffer[tempInt_gSP]==59) { // ;
        return tempInt_gSP;
      }
    }
    return -1;
  }

  /** Checks the read buffer wether it matches a specified int array.
      @param int[] ia: the array to match
      @return boolean: wether the buffer mathes the int array
  */
  private boolean checkReadBuffer(int[] ia) {
    if (readBufferPos==(ia.length+1)) {
      for (tempInt_cRB=0; tempInt_cRB<ia.length; tempInt_cRB++) {
        if (ia[tempInt_cRB]!=readBuffer[tempInt_cRB]) {
          return false;
        }
      }
      return true;
    }
    return false;
  }

  /** Adds a char to the display according to the insertion modes selected by the input stream flow.
      @param char c: the char to be added
  */
  private void addChar(char c) {
    switch (IRMstatus) {
      case IRM_OVR :
        display.addC(c);
	      break;
      case IRM_INS :
        display.insertChar();
	      display.addC(c);
	      break;
    }
  }

}