package com.monica.javacom.transport;

import com.monica.interfaces.*;
import com.monica.javacom.interfaces.*;
import com.monica.javacom.transport.*;
import com.monica.util.*;
import com.monica.util.debug.*;

import java.net.*;
import java.io.*;
import java.util.*;

/** This class provides TCP based connection over IP as a Transport. TCPTransport
    provides both server and client services. This class implements buffering in both
    directions.<br><br>
    Configureable values from the <b>JavaCom.transport</b> Configuration:
    <ul>
      <li><b>transportReadBufferSize</b>: the maximum size of the read buffer
      <li><b>transportWriteBufferSize</b>: the maximum size of the write buffer
    </ul>
*/
public class TCPTransport implements Transport {

  /* ---------------------------------------------------------------------------------- */
  /* Variables */
  /* ---------------------------------------------------------------------------------- */

  // Socket communication
  private ServerSocket server;
  private Socket clientOfServer;
  private Socket serverOfClient;

  // Network information
  private int port;
  private String host;

  // Current status
  private int status;
  private Configuration configuration;
  private String lastError;

  // Thread handling
  private Vector childrenClients;
  private Transport ownerComm;

  // Connection streams
  private OutputStream output;
  private InputStream input;

  // IO
  private com.monica.util.StringBuffer readBuffer;
  private com.monica.util.StringBuffer writeBuffer;
  private Converter converter;

  // Readable
  private com.monica.interfaces.Reader reader;
  private ServerManager serverManager;
  
  // statistics
  
  private int num_sent, num_received, num_readout, num_receivedforsending;
  
  /* ---------------------------------------------------------------------------------- */
  /* Constants */
  /* ---------------------------------------------------------------------------------- */
  
  private static final String DP="TCPTransport";

  private static final String NAME="TCPTransport";
  private static final String VERSION="0.3";
  
  private static final String CONF_STREAM_NAME="JavaCom.transport";

  private static final String CONF_READ_BUFFER_SIZE="transportReadBufferSize";
  private static final String CONF_WRITE_BUFFER_SIZE="transportWriteBufferSize";

  private static final int CONF_DEF_READ_BUFFER_SIZE=50000;
  private static final int CONF_DEF_WRITE_BUFFER_SIZE=50000;
  
  /* ---------------------------------------------------------------------------------- */
  /* Constructors */
  /* ---------------------------------------------------------------------------------- */

  /** Normal constructor. Should be called upon creation.
  */
  public TCPTransport() {
    status=TransportConstants.NOT_CONNECTED;
    childrenClients=new Vector();
    initConfiguration();
    lastError=new String();
    converter=new Converter();
  }

  /** Create a new thread for a new server.
      @param Socket client: the client of this server thread
  */
  private TCPTransport(Socket client) throws IOException {
    this.clientOfServer=client;
    status=TransportConstants.CONNECTED_SERVER;
    initConfiguration();
    initStreams();
    lastError=new String();
    converter=new Converter();
  }

  /* ---------------------------------------------------------------------------------- */
  /* Initializing */
  /* ---------------------------------------------------------------------------------- */

  public void initAsServer(int port) throws IOException {
    if (status==TransportConstants.NOT_CONNECTED) { // not connected
      this.port=port;
      server=new ServerSocket(port);
      status=TransportConstants.ACCEPTING_SOCKETS;
      while (true) {
        clientOfServer=server.accept();
        TCPTransport m=new TCPTransport(clientOfServer);
        m.setConfiguration(configuration);
        clientOfServer=null;
        Thread t=new Thread(m);
        childrenClients.addElement(m);
        serverManager.newClientHasArrivedEvent(this, m);
        //	DebugManager.logOut(DP, "Serverthread starting");
        t.start();
      }
    }
  }

  public void initAsClient(String host, int port) throws IOException {
    if (status==TransportConstants.NOT_CONNECTED) { // not connected
      this.port=port;
      this.host=host;
      serverOfClient=new Socket(host,port);
      Thread clientThread=new Thread(this);
      status=TransportConstants.CONNECTED_CLIENT;
      initStreams();
      clientThread.start();
    }
  }

  /** Runnable run method. Entry point for the read thread.
  */
  public void run() {
    int locstat=status;
    listenForRead();
  }

  /** Method that waits for byte available to be read out from the socket. If those
      arrive, reads them out, sends event to owner and waits again. While waiting
      thread blocks.
  */
  
  private byte lfr_tempRead[];
  
  void listenForRead() {
    int firstByte;
    int t;
    
    lfr_tempRead=new byte[readBuffer.getMaxBufferLength()];

    try {
      while (true) {
        firstByte=input.read();

        if (firstByte==-1) {
          closeConnections();
          return;
        }

        lfr_tempRead[0]=(byte)firstByte;
        addToReadBuffer(lfr_tempRead,1);

        while (input.available()>0) {
          t=input.read(lfr_tempRead);
          addToReadBuffer(lfr_tempRead, t);
        }

        if (readBuffer.getBufferLength()>0) {
	  Thread.currentThread().sleep(10);
          reader.readEvent(this);
        }

      }
    }
    catch (Exception e) {
      lastError=e.toString();
      DebugManager.logException(DP, e);
      closeConnections();
    }
  }

  /** Adds a byte array with the specified length to the read buffer.
      @param byte[] b: the byte array to be added to the read buffer
      @param int l: the length of the byte array
  */
  private void addToReadBuffer(byte[] c, int l) {
    if (!readBuffer.addToBuffer(c, l)) {
      synchronized (this) {
 	try {
  	  Thread.currentThread().sleep(1);
	}
 	catch (Exception e) {
          DebugManager.logException(DP, e);  
	}
      }
      reader.readEvent(this);
      try {
        synchronized (this) {
          while (!readBuffer.addToBuffer(c, l)) {          
            wait();
          }
          num_received+=l;
        }
      }
      catch (Exception e) {
        DebugManager.logException(DP, e);  
      }      
    }
    else {
      num_received+=l;  
    }
  }

  /** Adds a String to the write buffer.
      @param String s: the String to be added to the write buffer.
  */
  private void addToWriteBuffer(String s) {
    if (isWriteableConnected()) {
      num_receivedforsending+=s.length();  
      if (!writeBuffer.addToBuffer(s)) {
        flush();
        if (!writeBuffer.addToBuffer(s)) {
          writeOut(s);
        }
      }
    }
  }


  public void flush() {
    // DebugManager.logOut(DP, "tr.fl");
    if (isWriteableConnected()) {
      writeOut(writeBuffer.emptyBuffer());
    }
  }

  /** Writes out a String directly to the TCP socket without buffering.
      @param String s: the String to be written out
  */
  private void writeOut(String s) {
    if (isWriteableConnected()) {
      try {
          
        // time profiling  
        long TP_start, TP_stop, TP_diff; //--
        int TP_bytelen; //--
              
        TP_start=System.currentTimeMillis(); //--
        TP_bytelen=s.length(); //--
                    
        output.write(converter.getBytes(s));
        output.flush();
        num_sent+=s.length();
        
        TP_stop=System.currentTimeMillis(); //--    
        TP_diff=(TP_stop-TP_start)+1;
        DebugManager.logOut(DP, "NET speed="+((TP_bytelen/TP_diff)*1000)+" at the len="+TP_bytelen); //--        
      }
      catch (Exception e) {
        DebugManager.logException(DP, e);  
      }
    }
    else {
      DebugManager.logOut(DP, "writeable not connected.");
    }
  }

  public synchronized String getInput() {
    String temp;
    if ((isReadableConnected()) || (status==TransportConstants.NOT_CONNECTED_WITH_DATA_WAITING)) {
      temp=readBuffer.emptyBuffer();      
      notify();
      num_readout+=temp.length();
      return temp;
    }
    else {
      return new String();
    }
  }

  /* ---------------------------------------------------------------------------------- */
  /* Transport */
  /* ---------------------------------------------------------------------------------- */

  public void closeConnections() {
     
    DebugManager.logOut(DP, "num_received="+num_received);
    DebugManager.logOut(DP, "num_readout="+num_readout);
    
    DebugManager.logOut(DP, "num_sent="+num_sent);
    DebugManager.logOut(DP, "num_receivedforsending="+num_receivedforsending);
    
    int i;

    switch (status) {
      case TransportConstants.NOT_CONNECTED :
        break;

      case TransportConstants.CONNECTED_CLIENT :
        if (readBuffer.getBufferLength()>0) {
          status=TransportConstants.NOT_CONNECTED_WITH_DATA_WAITING;
        }
        else {
          status=TransportConstants.NOT_CONNECTED;
        }
        try {
          serverOfClient.close();
        }
        catch (Exception e) {
          DebugManager.logException(DP, e);  
        }
        if (reader!=null) {
          reader.connectionClosedEvent(this);
        }
        break;

      case TransportConstants.CONNECTED_SERVER :
        if (readBuffer.getBufferLength()>0) {
          status=TransportConstants.NOT_CONNECTED_WITH_DATA_WAITING;
        }
        else {
          status=TransportConstants.NOT_CONNECTED;
        }
        try {
          clientOfServer.close();
        }
        catch (Exception e) {
          DebugManager.logException(DP, e);  
        }
        toDefaultSettings();
        if (reader!=null) {
          reader.connectionClosedEvent(this);
        }
        break;

      case TransportConstants.ACCEPTING_SOCKETS :
        status=TransportConstants.NOT_CONNECTED;
        for (i=0; i<childrenClients.size(); i++) {
          ((Transport)(childrenClients.elementAt(i))).closeConnections();
        }
        try {
          server.close();
        }
        catch (Exception e) {
          DebugManager.logException(DP, e);  
        }
        break;

      case TransportConstants.NOT_CONNECTED_WITH_DATA_WAITING :
        break;
    }

    toDefaultSettings();
  }

  public String getTransportName() {
    return NAME;
  }

  public String getTransportVersion() {
    return VERSION;
  }
  
  public String getLayerTopViewName() {
    return NAME;
  }
  
  public String getLayerTopViewVersion() {
    return VERSION;
  }
  
  public String getCloseableName() {
    return NAME;
  }
  
  public String getCloseableVersion() {
    return VERSION;
  }                
  
  public void setReader(com.monica.interfaces.Reader reader) {
    this.reader=reader;
  }

  public com.monica.interfaces.Reader getReader() {
    return reader;
  }

  public void setServerManager(ServerManager serverManager) {
    this.serverManager=serverManager;
  }

  public ServerManager getServerManager() {
    return serverManager;
  }

  /* ---------------------------------------------------------------------------------- */
  /* Configureable */
  /* ---------------------------------------------------------------------------------- */

  public void setConfiguration(Configuration configuration) {
      try {
        this.configuration=configuration.getConfiguration(CONF_STREAM_NAME);        
        initConfiguration();
      }
      catch (Exception e) {
        DebugManager.logException(DP, e);  
      }
  }

  public Configuration getConfiguration() {
    return configuration;
  }

  /* ---------------------------------------------------------------------------------- */
  /* Writeable */
  /* ---------------------------------------------------------------------------------- */

  public String getWriteableName() {
    return NAME;
  }

  public String getWriteableVersion() {
    return VERSION;
  }

  public boolean write(char c) {
    char[] charArray;
    if (isWriteableConnected()) {
      charArray=new char[1];
      charArray[0]=c;
      addToWriteBuffer(new String(charArray));
      return true;
    }
    return false;
  }

  public boolean write(String s) {
    if (isWriteableConnected()) {
      addToWriteBuffer(s);
      return true;
    }
    return false;
  }

  public boolean write(char[] c) {
    if (isWriteableConnected()) {
      addToWriteBuffer(new String(c));
      return true;
    }
    return false;
  }

  public String getLastWriteableError() {
    String temp=lastError;
    lastError=new String();
    return temp;
  }

  public boolean isWriteableConnected() {
    switch (status) {
      case TransportConstants.CONNECTED_SERVER :
        return true;
      case TransportConstants.CONNECTED_CLIENT :
        return true;
      default:
        return false;
    }
  }

  /* ---------------------------------------------------------------------------------- */
  /* Readable */
  /* ---------------------------------------------------------------------------------- */

  public String getReadableName() {
    return NAME;
  }

  public String getReadableVersion() {
    return VERSION;
  }

  public int getAvailable() {
    return readBuffer.getBufferLength();
  }

  public boolean isReadableConnected() {
    switch (status) {
      case TransportConstants.CONNECTED_CLIENT :
        return true;
      case TransportConstants.CONNECTED_SERVER :
        return true;
      default:
        return false;
    }
  }

  private void toDefaultSettings() {
    server=null;
    clientOfServer=null;
    serverOfClient=null;
    childrenClients=null;
    input=null;
    output=null;
  }

  private void initStreams() throws IOException {
    num_sent=0;
    num_received=0;
    num_readout=0;
    num_receivedforsending=0;
    switch (status) {
      case TransportConstants.CONNECTED_CLIENT :
        input=serverOfClient.getInputStream();
        output=serverOfClient.getOutputStream();
        break;
      case TransportConstants.CONNECTED_SERVER :
        input=clientOfServer.getInputStream();
        output=clientOfServer.getOutputStream();
        break;
      default:
//        DebugManager.logOut(DP, "no streams initialized");
        break;
    }
  }

  public String checkSum(String s) {
    long tempLong=0;
    String tempString;
    for (int i=0; i<s.length(); i++) {
      tempLong+=((int)s.charAt(i))+1;
    }
    tempString=new String((new Long(tempLong)).toString());
    if (tempString.length()==0) {
      tempString="##";
    }
    if (tempString.length()==1) {
      tempString=tempString.concat("#");
    }
    tempString=tempString.substring(tempString.length()-2,tempString.length());
    return tempString;
  }

  /** Initializes intern variables from the Configuration.
  */
  private void initConfiguration() {
    if (configuration==null) {
      readBuffer=new com.monica.util.StringBuffer(CONF_DEF_READ_BUFFER_SIZE);
      writeBuffer=new com.monica.util.StringBuffer(CONF_DEF_WRITE_BUFFER_SIZE);
    }
    else {
      readBuffer=new com.monica.util.StringBuffer(configuration.getInt(CONF_READ_BUFFER_SIZE,
        CONF_DEF_READ_BUFFER_SIZE));
      writeBuffer=new com.monica.util.StringBuffer(configuration.getInt(CONF_WRITE_BUFFER_SIZE,
        CONF_DEF_WRITE_BUFFER_SIZE));
    }
  }

}


