package com.monica.streams;

import com.monica.interfaces.*;
import com.monica.util.*;
import java.io.*;

/** Class for enabling to write to the TM layer hierarchy using the
    conventional OutputStream methods. Objects of this class act as
    OutputStreams, and they forward the data written on them 
    to Readers.
*/    

public class ReadableIsAOS extends OutputStream implements Readable {

  private static final String NAME="ReadableIsAOS";
  private static final String VERSION="0.3";
  private static final int bufSize=50000;
  
  private com.monica.util.StringBuffer buf;
  private boolean connected;
  private com.monica.interfaces.Reader reader;
  private boolean threadSuspended;
  
  /** Default constructor.
  */    
  public ReadableIsAOS() {
    connected=true;
    buf=new com.monica.util.StringBuffer(bufSize);
    threadSuspended=false;
  }
  
  // --------------------------------------------------------------------------
  // Readable
  // --------------------------------------------------------------------------
  
  public String getReadableName() {
    return NAME;
  }
  
  public String getReadableVersion() {
    return VERSION;
  }
  
  public void setReader(com.monica.interfaces.Reader reader) {
    this.reader=reader;
  }            
  
  public com.monica.interfaces.Reader getReader() {
    return reader;
  }
  
  public int getAvailable() {
    return buf.getBufferLength();
  }
  
  public synchronized String getInput() {
    String temp=buf.emptyBuffer();
    threadSuspended=false;
    notify();
    return temp;
  }

  public boolean isReadableConnected() {
    return connected;
  }    
  
  // --------------------------------------------------------------------------
  // OutputStream
  // --------------------------------------------------------------------------    
  
  /** This method closes the OuputStream. After it, nothing can
      be written on it. If there is still data in the buffer, 
      a readEvent is called for the Reader if specified.
      @exception IOException: if something goes wrong
  */      
  public void close() throws IOException {
    if (connected==false) throw new IOException("Not connected.");
    if (buf.getBufferLength()>0) {
      if (reader==null) throw new IOException("Reader not specified.");
      reader.readEvent(this);
    }
    connected=false;  
  }
  
  /** Flushes the read buffer to the Reader. As reading in the
      TM layer hierarchy is asynchronous, the current thread will
      be suspended after a readEvent is called at the Reader till
      the getInput method of this object is called.
      @exception IOException: if something goes wrong
  */      
  public void flush() throws IOException {
    if (connected==false) throw new IOException("Not connected.");
    if (buf.getBufferLength()>0) {
      if (reader==null) throw new IOException("Reader not specified.");
      threadSuspended=true;
      reader.readEvent(this);
      synchronized (this) {              
	try {
	  if (threadSuspended) {
    	    wait();
	  }    
	}
	catch (Exception e) {
	}  
      }	
    }
  }
  
  /** Writes an int (0...255) to the Reader. Writing is buffered.
      @param int b: the int to be written
      @exception IOException: if something goes wrong
  */      
  public void write(int b) throws IOException {
    if (connected==false) throw new IOException("Not connected.");
    if (reader==null) throw new IOException("Reader not specified.");
    while (!buf.addToBuffer((char)b)) {      
      threadSuspended=true;
      reader.readEvent(this);
      synchronized (this) {        
	try {
	  if (threadSuspended) {
            wait();
	  }    
	}
	catch (Exception e) {
	}
      }
    }  	
  }
  
  /** Writes a byte array to the Reader. Writing is buffered.
      @param byte[] b: the byte array to be written
      @exception IOException: if something goes wrong
  */      
  public void write(byte b[]) throws IOException {
    if (connected==false) throw new IOException("Not connected.");
    if (reader==null) throw new IOException("Reader not specified.");
    int sP=0;
    int bL=0;
    boolean doLoop=true;
    while (doLoop) {
      bL=buf.available();
      if ((b.length-sP)>bL) {
        buf.addToBuffer(b,sP,bL);
	sP+=bL;
	threadSuspended=true;
	reader.readEvent(this);
	synchronized (this) {          
	  try {	  
	    if (threadSuspended) {
	      wait();
	    }
	  }
	  catch (Exception e) {
	  }
	}
	if (sP==b.length) {
	  doLoop=false;
	}  
      }
      else {
        buf.addToBuffer(b,sP,b.length-sP);
	doLoop=false;
      }
    }
  }
  
  /** Writes a part (specified with start offset and length) of a
      byte array to the Reader. Writing is buffered.      
      @param byte[] b: the byte array to be written
      @param int off: the starting offset of the part
      @param int len: the length of the part
      @exception IOException: if something goes wrong
  */      
  public void write(byte b[], int off, int len) throws IOException {
    if (connected==false) throw new IOException("Not connected.");
    if (reader==null) throw new IOException("Reader not specified.");
    int sP=off;
    int bL=0;
    boolean doLoop=true;
    while (doLoop) {
      bL=buf.available();
      if (len>bL) {
        buf.addToBuffer(b,sP,bL);
	sP+=bL;
	len-=bL;
	threadSuspended=true;
	reader.readEvent(this);
	synchronized (this) {	          
	  try {
	    if (threadSuspended) {
       	      wait();
	    }
	  }
	  catch (Exception e) {
	  }    
	}
	if (len==0) {
	  doLoop=false;
	}  
      }
      else {
        buf.addToBuffer(b,sP,len);
	doLoop=false;
      }
    }
  }
  
}