package com.monica.streams;

import com.monica.interfaces.*;
import com.monica.util.*;
import java.io.*;

/** Class for enabling to read from the TM layer hierarchy using the
    conventional InputStream methods. Objects of this class act as
    InputStreams, they read the data from their Readable. 
*/    

public class ReaderIsAIS extends InputStream implements com.monica.interfaces.Reader {

  private static final String NAME="ReaderIsAIS";
  private static final String VERSION="0.3";
  private static final int bufSize=50000;
  
  private Readable readable;
  private com.monica.util.StringBuffer buf;
  
  private boolean rClosed;
  private boolean isClosed;
  private boolean threadSuspendedGI;
  private boolean threadSuspendedR;
  
  /** Default constructor.
  */
  public ReaderIsAIS() {
    buf=new com.monica.util.StringBuffer(bufSize);
    rClosed=false;
    isClosed=false;
    threadSuspendedGI=false;
    threadSuspendedR=false;
  }
  
  // --------------------------------------------------------------------------
  // Reader
  // --------------------------------------------------------------------------
  
  public String getReaderName() {
    return NAME;
  }
  
  public String getReaderVersion() {
    return VERSION;
  }
  
  public void readEvent(Readable readable) {
    String temp=readable.getInput();
    int bL=0;
    boolean doLoop=true;
    int sP=0;
    while (doLoop) {
      bL=buf.available();
      if ((temp.length()-sP)>bL) {
        buf.addToBuffer(temp,sP,bL);
	sP+=bL;
	threadSuspendedGI=true;
	synchronized (this) {
	  threadSuspendedR=false;
	  notify();
	  try {
	    if (threadSuspendedGI) {
  	      wait();
	    }  
	  }
	  catch (Exception e) {
	  }
	}
	if (sP==temp.length()) {
	  doLoop=false;
	}
      }	
      else {
        buf.addToBuffer(temp,sP,temp.length()-sP);
	synchronized (this) {
	  threadSuspendedR=false;
	  notify();
	}  
	doLoop=false;
      }
    }  	
  }
  
  public void connectionClosedEvent(Readable readable) {
    rClosed=true;
    synchronized (this) {
      threadSuspendedR=false;
      notify();
    }  
  }
  
  public void errorEvent(String errorDescription, Readable readable) {
    rClosed=true;
    synchronized (this) {
      threadSuspendedR=false;
      notify();
    }  
  }
  
  public void setReadable(Readable readable) {
    this.readable=readable;
  }
  
  public Readable getReadable() {
    return readable;
  } 
  
  // --------------------------------------------------------------------------
  // InputStream
  // --------------------------------------------------------------------------
  
  /** Reads 1 single byte from the stream. The read blocks till data
      is available. If the stream is closed, -1 will be returned, 
      otherwise the byte read.
      @return int: -1 if stream end, otherwise the byte read
      @exception IOException: if the InputStream has been closed
  */      
  public synchronized int read() throws IOException {
    if (isClosed) throw new IOException("InputStream closed.");
    if ((rClosed) && (buf.getBufferLength()==0)) return -1;
    
    while ((buf.getBufferLength()==0) || (!rClosed)) {
      threadSuspendedR=true;
      if (threadSuspendedR) {
        try {
	  wait();
	}
	catch (Exception e) {
	}
      }
    }  	  
    
    if ((rClosed) && (buf.getBufferLength()==0)) return -1;  
    
    String temp=buf.emptyBuffer(1);
    threadSuspendedGI=false;
    notify();
    return ((int)temp.charAt(0));      
  }
  
  /** Reads b.length bytes from the stream. The read blocks till data
      is available. If the stream is closed, -1 will be returned, 
      otherwise the number of bytes read.
      @param byte[] b: the byte array into the data will be read
      @return int: -1 if stream end, otherwise the number of bytes read
      @exception IOException: if the InputStream has been closed
  */      
  public synchronized int read(byte[] b) throws IOException {
    if (isClosed) throw new IOException("InputStream closed.");
    if ((rClosed) && (buf.getBufferLength()==0)) return -1;    
    
    String temp;
    
    if ((rClosed) && (buf.getBufferLength()<b.length)) {      
      temp=buf.emptyBuffer();
      for (int i=0; i<temp.length(); i++) b[i]=(byte)temp.charAt(i);
      if (temp.length()==bufSize) {
        threadSuspendedGI=false;
	notify();
      }	
      return temp.length();
    }
    
    int len;
    if (b.length>bufSize) len=bufSize; else len=b.length;
    
    while ((buf.getBufferLength()<len) || (!rClosed)) {
      threadSuspendedR=true;
      if (threadSuspendedR) {
        try {
	  wait();
	}
	catch (Exception e) {
	}
      }
    }
    
    if ((rClosed) && (buf.getBufferLength()<b.length)) {
      temp=buf.emptyBuffer();
      for (int i=0; i<temp.length(); i++) b[i]=(byte)temp.charAt(i);
      if (temp.length()==bufSize) {
        threadSuspendedGI=false;
	notify();
      }	
      return temp.length();
    }
    
    temp=buf.emptyBuffer(len);
    threadSuspendedGI=false;
    notify();
    for (int i=0; i<temp.length(); i++) b[i]=(byte)temp.charAt(i);
    return temp.length();
  }
        
  /** Reads len bytes from the stream. The read blocks till data
      is available. If the stream is closed, -1 will be returned, 
      otherwise the number of bytes read.
      @param byte[] b: the byte array into the data will be read
      @param int off: the beginning offset from which the bytes will
        be written into b (inclusive)
      @param int len: the number of bytes to be read	
      @return int: -1 if stream end, otherwise the number of bytes read
      @exception IOException: if the InputStream has been closed
      @exception IndexOutOfBoundsException: if (b.length<(off+len))
  */      
  public synchronized int read(byte[] b, int off, int len) throws IOException {
    if (isClosed) throw new IOException("InputStream closed.");
    if ((rClosed) && (buf.getBufferLength()==0)) return -1;    
    
    String temp;
    
    if ((rClosed) && (buf.getBufferLength()<len)) {      
      temp=buf.emptyBuffer();
      for (int i=0; i<temp.length(); i++) b[off+i]=(byte)temp.charAt(i);
      if (temp.length()==bufSize) {
        threadSuspendedGI=false;
	notify();
      }	
      return temp.length();
    }
    
    int lenR;
    if (b.length>bufSize) lenR=bufSize; else lenR=b.length;
    
    while ((buf.getBufferLength()<lenR) || (!rClosed)) {
      threadSuspendedR=true;
      if (threadSuspendedR) {
        try {
	  wait();
	}
	catch (Exception e) {
	}
      }
    }
    
    if ((rClosed) && (buf.getBufferLength()<len)) {
      temp=buf.emptyBuffer();
      for (int i=0; i<temp.length(); i++) b[off+i]=(byte)temp.charAt(i);
      if (temp.length()==bufSize) {
        threadSuspendedGI=false;
	notify();
      }	
      return temp.length();
    }
    
    temp=buf.emptyBuffer(len);
    threadSuspendedGI=false;
    notify();
    for (int i=0; i<temp.length(); i++) b[off+i]=(byte)temp.charAt(i);
    return temp.length();
  }
  
  /** Skips n bytes in the stream. Skipped data will be lost. For
      some reasons the stream may skip less bytes than n.
      @param long n: number of bytes to be skipped
      @return long: number of bytes actually skipped
      @exception IOException: if the InputStream has been closed
  */      
  public long skip(long n) throws IOException {
    if (isClosed) throw new IOException("InputStream closed.");
  
    int ret;
    if (n>buf.getBufferLength()) {
      ret=buf.getBufferLength();
      buf.emptyBuffer();
    }
    else {
      ret=(int)n;
      buf.emptyBuffer((int)n);  
    }
    synchronized (this) {
      threadSuspendedGI=false;
      notify();
    }
    return (long)ret;
  }        
  
  /** Returns the number of bytes that are currently available
      for reading from the stream.
      @return int: the number of bytes available from the stream
      @exception IOException: if the InputStream has been closed
  */      
  public int available() throws IOException {
    if (isClosed) throw new IOException("InputStream closed.");  
    return (buf.getBufferLength());
  }
  
  /** Closes the InputStream. After close(), no read, close and available
      methods can be called without IOException.
      @exception IOException: if the InputStream has been closed
  */      
  public void close() throws IOException {
    if (isClosed) throw new IOException("InputStream closed.");  
    isClosed=true;
  }
  
  /** ReaderIsAIS does not support mark/reset. This method
      does actually nothing.
  */      
  public synchronized void mark(int readlimit) {
  }
  
  /** ReaderIsAIS does not support mark/reset. This method will
      always throw an IOException.
      @exception IOException: ReaderIsAIS does not support mark/reset
  */      
  public synchronized void reset() throws IOException {
    throw new IOException("Mark not supported.");
  }
  
  /** ReaderIsAIS does not support mark/reset.
      @return boolean: false
  */      
  public boolean markSupported() {
    return false;
  } 
    
}