package com.monica.streams;

import com.monica.interfaces.*;
import com.monica.util.*;
import java.io.*;

/** Class for enabling to read from the TM layer hierarchy using the
    conventional InputStream methods. Objects of this class act as 
    InputStreams, the data readable from the InputStream will be
    generated by the Writer corresponding to this Writeable.
*/

public class WriteableIsAIS extends InputStream implements Writeable {

  private static final String NAME="WriteableIsAIS";
  private static final String VERSION="0.3";
  private static final int bufSize=50000;
  
  private com.monica.util.StringBuffer buf;
  private boolean isClosed;
  private boolean threadSuspendedW;
  private boolean threadSuspendedR;
  private String lastError;
  
  /** Default constructor.
  */  
  public WriteableIsAIS() {
    buf=new com.monica.util.StringBuffer(bufSize);
    isClosed=false;
    threadSuspendedW=false;
    threadSuspendedR=false;
    lastError=new String();
  }
  
  // --------------------------------------------------------------------------
  // Writeable
  // --------------------------------------------------------------------------
  
  public String getWriteableName() {
    return NAME;
  }
  
  public String getWriteableVersion() {
    return VERSION;
  }    
  
  public boolean write(char c) {    
    if (isClosed) {
      lastError="InputSream has been closed.";
      return false;
    }  
    boolean doLoop=true;
    while (doLoop) {
      if (buf.available()>0) {
        buf.addToBuffer(c);
	synchronized (this) {
	  threadSuspendedR=false;
	  notify();
	}  	
	doLoop=false;
      }
      else {	
	threadSuspendedW=true;
	synchronized (this) {
	  try {
	    if (threadSuspendedW) {
  	      wait();
	    }  
	  }
	  catch (Exception e) {
	  }
	}
      }	
    }  	
    return true;
  }

  public boolean write(String s) {
    if (isClosed) {
      lastError="InputSream has been closed.";
      return false;
    }  
    int bL=0;
    boolean doLoop=true;
    int sP=0;
    while (doLoop) {
      bL=buf.available();
      if ((s.length()-sP)>bL) {
        buf.addToBuffer(s,sP,bL);
	sP+=bL;
	threadSuspendedW=true;
	synchronized (this) {
	  threadSuspendedR=false;
	  notify();
	  try {
	    if (threadSuspendedW) {
  	      wait();
	    }  
	  }
	  catch (Exception e) {
	  }
	}
	if (sP==s.length()) {
	  doLoop=false;
	}
      }	
      else {
        buf.addToBuffer(s,sP,s.length()-sP);
	synchronized (this) {
	  threadSuspendedR=false;
	  notify();
	}  
	doLoop=false;
      }
    }  	
    return true;
  }
  
  public boolean write(char[] c) {
    if (isClosed) {
      lastError="InputSream has been closed.";
      return false;
    }  
    int bL=0;
    boolean doLoop=true;
    int sP=0;
    while (doLoop) {
      bL=buf.available();
      if ((c.length-sP)>bL) {
        buf.addToBuffer(c,sP,bL);
	sP+=bL;
	threadSuspendedW=true;
	synchronized (this) {
	  threadSuspendedR=false;
	  notify();
	  try {
	    if (threadSuspendedW) {
  	      wait();
	    }  
	  }
	  catch (Exception e) {
	  }
	}
	if (sP==c.length) {
	  doLoop=false;
	}
      }	
      else {
        buf.addToBuffer(c,sP,c.length-sP);
	synchronized (this) {
	  threadSuspendedR=false;
	  notify();
	}  
	doLoop=false;
      }
    }  	
    return true;
  }
  
  public void flush() {
  }
  
  public String getLastWriteableError() {
    String temp=lastError;
    lastError=new String();
    return temp;
  }
  
  public boolean isWriteableConnected() {
    return (!isClosed);
  }
  
  // --------------------------------------------------------------------------
  // InputStream
   // --------------------------------------------------------------------------

  /** Reads 1 single byte from the stream. The read blocks till data
      is available.
      @return int: the byte read
      @exception IOException: if the InputStream has been closed
  */      
  public synchronized int read() throws IOException {
    if (isClosed) throw new IOException("InputStream closed.");
    
    while (buf.getBufferLength()==0) {
      threadSuspendedR=true;
      if (threadSuspendedR) {
        try {
	  wait();
	}
	catch (Exception e) {
	}
      }
    }  	  
    
    String temp=buf.emptyBuffer(1);
    threadSuspendedW=false;
    notify();
    return ((int)temp.charAt(0));      
  }
  
  /** Reads b.length bytes from the stream. The read blocks till data
      is available. Returns the number of bytes read.
      @param byte[] b: the byte array into the data will be read
      @return int: the number of bytes read
      @exception IOException: if the InputStream has been closed
  */      
  public synchronized int read(byte[] b) throws IOException {
    if (isClosed) throw new IOException("InputStream closed.");
    
    String temp;
    
    int len;
    if (b.length>bufSize) len=bufSize; else len=b.length;
    
    while (buf.getBufferLength()<len) {
      threadSuspendedR=true;
      if (threadSuspendedR) {
        try {
	  wait();
	}
	catch (Exception e) {
	}
      }
    }
    
    temp=buf.emptyBuffer(len);
    threadSuspendedW=false;
    notify();
    for (int i=0; i<temp.length(); i++) b[i]=(byte)temp.charAt(i);
    return temp.length();
  }
        
  /** Reads len bytes from the stream. The read blocks till data
      is available. Returns the number of bytes read.
      @param byte[] b: the byte array into the data will be read
      @param int off: the beginning offset from which the bytes will
        be written into b (inclusive)
      @param int len: the number of bytes to be read	
      @return int: the number of bytes read
      @exception IOException: if the InputStream has been closed
      @exception IndexOutOfBoundsException: if (b.length<(off+len))
  */      
  public synchronized int read(byte[] b, int off, int len) throws IOException {
    if (isClosed) throw new IOException("InputStream closed.");
    
    String temp;
    
    int lenR;
    if (b.length>bufSize) lenR=bufSize; else lenR=b.length;
    
    while (buf.getBufferLength()<lenR) {
      threadSuspendedR=true;
      if (threadSuspendedR) {
        try {
	  wait();
	}
	catch (Exception e) {
	}
      }
    }
    
    temp=buf.emptyBuffer(len);
    threadSuspendedW=false;
    notify();
    for (int i=0; i<temp.length(); i++) b[off+i]=(byte)temp.charAt(i);
    return temp.length();
  }
  
  /** Skips n bytes in the stream. Skipped data will be lost. For
      some reasons the stream may skip less bytes than n.
      @param long n: number of bytes to be skipped
      @return long: number of bytes actually skipped
      @exception IOException: if the InputStream has been closed
  */      
  public long skip(long n) throws IOException {
    if (isClosed) throw new IOException("InputStream closed.");
  
    int ret;
    if (n>buf.getBufferLength()) {
      ret=buf.getBufferLength();
      buf.emptyBuffer();
    }
    else {
      ret=(int)n;
      buf.emptyBuffer((int)n);  
    }
    synchronized (this) {
      threadSuspendedW=false;
      notify();
    }
    return (long)ret;
  }        
  
  /** Returns the number of bytes that are currently available
      for reading from the stream.
      @return int: the number of bytes available from the stream
      @exception IOException: if the InputStream has been closed
  */      
  public int available() throws IOException {
    if (isClosed) throw new IOException("InputStream closed.");  
    return (buf.getBufferLength());
  }
  
  /** Closes the InputStream. After close(), no read, close and available
      methods can be called without IOException.
      @exception IOException: if the InputStream has been closed
  */      
  public void close() throws IOException {
    if (isClosed) throw new IOException("InputStream closed.");  
    isClosed=true;
  }
  
  /** WriteableIsAIS does not support mark/reset. This method
      does actually nothing.
  */      
  public synchronized void mark(int readlimit) {
  }
  
  /** WriteableIsAIS does not support mark/reset. This method will
      always throw an IOException.
      @exception IOException: WriteableIsAIS does not support mark/reset
  */      
  public synchronized void reset() throws IOException {
    throw new IOException("Mark not supported.");
  }
  
  /** WriteableIsAIS does not support mark/reset.
      @return boolean: false
  */      
  public boolean markSupported() {
    return false;
  } 
   
}