package com.monica.streams;

import com.monica.interfaces.*;
import com.monica.util.*;
import java.io.*;

/** Class to enable to write into the TM layer hierarchy using the
    conventional OutputStream methods. Objects of this class act as
    OutputStreams, and they forward the data written on them 
    to Writeables.
*/

public class WriterIsAOS extends OutputStream implements com.monica.interfaces.Writer {

  private static final String NAME="WriterIsAOS";
  private static final String VERSION="0.3";
  private static final int bufSize=50000;
  
  private com.monica.util.StringBuffer buf;
  private Writeable writeable;
  private boolean connected;
  
  /** Default constructor.
  */
  public WriterIsAOS() {
    connected=true;
  }    
  
  // ------------------------------------------------------------------------------------
  // Writer 
  // ------------------------------------------------------------------------------------

  public String getWriterName() {
    return NAME;
  }
  
  public String getWriterVersion() {
    return VERSION;
  }    
  
  public void setWriteable(Writeable writeable) {
    this.writeable=writeable;
    connected=true;
    buf=new com.monica.util.StringBuffer(bufSize);
  }
  
  public Writeable getWriteable() {    
    return writeable;
  }    
  
  // ------------------------------------------------------------------------------------
  // OutputStream
  // ------------------------------------------------------------------------------------
  
  /** Writes an int (0...255) to the Writeable of this object. Writing is buffered.
      @param int b: the int to be written to the MCJWriteable
      @exception IOException: if something goes wrong
  */      
  public void write(int b) throws IOException {
    if (!connected) throw new IOException("Not connected.");
    if (writeable==null) throw new IOException("Writeable not specified.");
    if (!buf.addToBuffer((char)b)) {
      flush();
      if (!buf.addToBuffer((char)b)) {
        connected=false;
        throw new IOException("Buffer too small.");	
      }	
    }
  }
  
  /** Writes a byte array to the Writeable of this object. Writing is buffered.
      @param byte[] b: the byte array to be written to the MCJWriteable
      @exception IOException: if something goes wrong
  */
  public void write(byte b[]) throws IOException {
    if (!connected) throw new IOException("Not connected.");
    if (writeable==null) throw new IOException("Writeable not specified.");
    if (!buf.addToBuffer(b)) {
      flush();
      if (!buf.addToBuffer(b)) {
        writeOut(new String(b));
      }
    }  
  }
  
  /** Writes a part (specified with beginning offset and length) of a byte array 
      to the Writeable of this object. Writing is buffered.
      @param byte[] b: the byte array to be written to the MCJWriteable
      @param int off: the beginning offset (inclusive) of the part to be written out
      @param int len: the length of the part to be written out
      @exception IOException: if something goes wrong
  */
  public void write(byte[] b, int off, int len) throws IOException {
    if (!connected) throw new IOException("Not connected.");
    if (writeable==null) throw new IOException("Writeable not specified.");
    if (!buf.addToBuffer(b, off, len)) {
      flush();
      if (!buf.addToBuffer(b, off, len)) {
        writeOut(new String(b, off, len));
      }
    }  
  }
  
  /** Flushes the contents of the buffer to the Writeable of this object.
      @throw IOException: if something goes wrong
  */      
  public void flush() throws IOException {
    if (!connected) throw new IOException("Not connected.");
    if (writeable==null) throw new IOException("Writeable not specified.");    
    writeOut(buf.emptyBuffer());
  }
  
  /** Closes the OuputStream. After a successful close, nothing can be
      written to it.
      @exception IOException: if something goes wron
  */      
  public void close() throws IOException {
    if (!connected) throw new IOException("Not connected.");
    if (buf.getBufferLength()>0) {
      flush();
    }  
    connected=false;
  }
 
  /** Writes a String directly to the Writeable of this object. Writing is
      not buffered.
      @exception IOException: if something goes wrong
  */      
  private void writeOut(String s) throws IOException {
    if (writeable==null) throw new IOException("Writeable not specified.");
    if (!writeable.write(s)) {
      throw new IOException("Writeable returned error string: "+
        writeable.getLastWriteableError());
    }	
  }
}