package com.monica.util;

import java.util.*;

/** Class for providing buffer functionality over arrays of bytes.
 */

public class ByteArrayBuffer {
    
    private Vector data;
    private int len;
    private int maxLength;

    /** Creates new ByteArrayBuffer */
    public ByteArrayBuffer() {
        maxLength=-1;
        _internalInit();
    }
    
    /** Creates new ByteArrayBuffer with the specified maximum
     *  number of bytes.
     *  @param int mL: the maximum number of bytes allowed in this buffer
     */     
    public ByteArrayBuffer(int mL) {
        maxLength=mL;
        _internalInit();
    }
    
    /** Returns the number of bytes in the buffer.
     *  @return int: the number of bytes in the buffer
     */
    public int length() {
        return len;
    }
    
    /** Adds a byte to the end of the buffer.
     *  @param byte b: the byte to be added to the buffer
     *  @return boolean: whether the byte could be added to the buffer
     */
    public boolean add (byte b) {
        if (_possible(1)) {
            byte ba[];
            ba=new byte[1];
            ba[0]=b;
            data.addElement(ba);
            return true;
        }
        return false;
    }
    
    /** Adds a byte array to the end of the buffer.
     *  @param byte[] b: the byte array to be added to the buffer
     *  @return boolean: whether the byte array could be added to the buffer
     */
    public boolean add (byte b[]) {
        if (_possible(b.length)) {
            data.addElement(b);
            return true;
        }
        return false;
    }
    
    /** Adds the first len bytes of a byte array to the end of the buffer.
     *  @param byte b: the byte to be added to the buffer
     *  @param int len: the number of bytes to add to the buffer
     *  @return boolean: whether the byte array could be added to the buffer
     */    
    public boolean add (byte b[], int len) {
        if (_possible(len)) {
            byte ba[];
            ba=new byte[len];
            System.arraycopy(b,0,ba,0,len);
            data.addElement(ba);
            return true;
        }
        return false;
    }
    
    /** Adds len bytes beginning with offset from a byte array to the buffer.
     *  @param byte[] b: the byte array to be added to the buffer
     *  @param int offset: the starting offset
     *  @param int len: the number of bytes to be added to the buffer
     *  @return boolean: whether the bytes could be added to the buffer
     */    
    public boolean add (byte b[], int offset, int len) throws ArrayIndexOutOfBoundsException {
        if (b.length<(offset+len)) {
            throw new ArrayIndexOutOfBoundsException();
        }
        if (_possible(len)) {
            byte ba[];
            ba=new byte[len];
            System.arraycopy(b,offset,ba,0,len);
            data.addElement(ba);
            return true;
        }
        return false;            
    }

    /** Return the contents of the buffer and clears its content.
     *  @return byte[]: the contents of the buffer
     */
    public byte[] emptyBuffer() {
        byte b[]=getBuffer();
        clear();
        return b;
    }
    
    /** Return the contents of the buffer.
     *  @return byte[]: the contents of the buffer
     */
    public byte[] getBuffer() {
        byte bt[], bs[];
        bt=new byte[len];
        int size=data.size();
        int pos=0;
        for (int i=0; i<size; i++) {
            bs=(byte[])data.elementAt(i);
            System.arraycopy(bs,0,bt,pos,bs.length);
            pos+=bs.length;
        }
        return bt;
    }
    
    /** Clears the contents of the buffer. */
    public void clear() {
        _internalInit();
    }
      
    private void _internalInit() {
        data=new Vector();
        len=0;
    }
    
    private boolean _possible(int len) {
        if (maxLength==-1) {
            this.len+=len;
            return true;
        }
        else {
            if ((this.len+len)<=maxLength) {
                this.len+=len;
                return true;
            }
            return false;
        }
    }     
}
