package com.monica.util;

import com.monica.interfaces.*;

/** Class for implementing a Vector like functionality with using Equality 
    for checking equality between the contained objects.
*/    
public class EqualityVector {

  private static final float REDUCE_FACTOR=0.6f;
  private static final float REDUCE_LIMIT=0.5f;
  private static final float ENLARGE_FACTOR=1.5f;
  private static final int MIN_SIZE=10;

  private Equality equality;
  private boolean synchronizing;
  private boolean erased[];
  private int erasedCount;
  private Object[] content;
  private int count;
  private int maxSize;
  private boolean instantRemoval;

  /** Default constructor. 
      @param Equality e: the Equality to be used for checking equality between contained objects
      @param boolean s: wether using synchronization for all the operations or not, true=synchronization
      @param boolean iR: wether to remove the removed elements from the memory list instantly or not,
        true=instant removal
      @param int m: the maximal number of objects in this Vector
  */      
  public EqualityVector(Equality e, boolean s, boolean iR, int m) {
    equality=e;
    synchronizing=s;
    content=new Object[MIN_SIZE];
    erased=new boolean[MIN_SIZE];
    for (int i=0; i<MIN_SIZE; i++) erased[i]=false;
    count=0;
    maxSize=m;
    erasedCount=0;
    instantRemoval=iR;
  }

  /** Sets the Equality to be used.
      @param Equality e: the Equality to be used for checking equality between objects
  */      
  public synchronized void setEquality(Equality e) {
    equality=e;
  }

  /** Sets wether to use synchronization for all the operations, or not. 
      @param boolean s: true for using synchronization
  */      
  public synchronized void setSynchronizing(boolean s) {
    synchronizing=s;
  }

  /** Adds an object to the vector.
      @param Object o: the Object to be added
      @return boolean: true if adding successful
  */      
  public boolean add(Object o) {
    if (synchronizing) {
      synchronized (this) {
        return internal_add(o);
      }
    }
    return internal_add(o);
  }

  /** Returns the first occurance of the Object o.
      @param Object o: the Object whose first occurance should be returned
      @param boolean useEquality: wether to use the supplied Equality or use Object.equals instead,
        true=use Equality
      @return int: the first index of the Object o, -1 if not in the vector	
  */            
  public int firstIndex(Object o, boolean useEquality) {
    if (synchronizing) {
      synchronized (this) {
        return internal_firstIndex(o, useEquality);
      }
    }
    return internal_firstIndex(o, useEquality);
  }

  /** Returns the last occurance of the Object o.
      @param Object o: the Object whose last occurance should be returned
      @param boolean useEquality: wether to use the supplied Equality or use Object.equals instead,
        true=use Equality
      @return int: the first last of the Object o, -1 if not in the vector	
  */            
  public int lastIndex(Object o, boolean useEquality) {
    if (synchronizing) {
      synchronized (this) {
        return internal_lastIndex(o, useEquality);
      }
    }
    return internal_lastIndex(o, useEquality);
  }
  
  /** Returns the number of occurances of the Object o.
      @param Object o: the Object whose occurances should be counted
      @param boolean useEquality: wether to use the supplied Equality or use Object.equals instead,
        true=use Equality
      @return int: the number of occurances of o	
  */            
  public int count(Object o, boolean useEquality) {
    if (synchronizing) {
      synchronized (this) {
        return internal_count(o, useEquality);
      }
    }
    return internal_count(o, useEquality);
  }

  /** Returns the Object at the position pos.
      @param int pos: the position of the returned Object
      @return Object: the Object at the position pos
      @throws ArrayIndexOutOfBoundsException: if pos>length of the vector
      @throws Exception: if the wanted position is deleted
  */      
  public Object get(int pos) throws Exception {
    if (synchronizing) {
      synchronized (this) {
        return internal_get(pos);
      }
    }
    return internal_get(pos);
  }

  /** Sets the supplied Object o to be the supplied position pos.
      @param Object o: the Object to be set
      @throws ArrayIndexOutOfBoundsException: if pos>=length of the vector
  */      
  public void set(Object o, int pos) throws ArrayIndexOutOfBoundsException {
    if (synchronizing) {
      synchronized (this) {
        internal_set(o, pos);
        return;
      }
    }
    internal_set(o, pos);
  }

  /** Inserts the supplied Object o to be the supplied position pos.
      @param Object o: the Object to be inserted
      @throws ArrayIndexOutOfBoundsException: if pos>length of the vector
  */
  public boolean insert(Object o, int pos) throws ArrayIndexOutOfBoundsException {
    if (synchronizing) {
      synchronized (this) {
        return internal_insert(o, pos);
      }
    }
    return internal_insert(o, pos);
  }
  
  /** Removes the supplied Object o from the vector.
      @param Object o: the Object to be removed
      @param boolean useEquality: wether to use the supplied Equality or use Object.equals instead,
        true=use Equality
      @return boolean: wether to Object could be removed or not
  */      	
  public boolean remove(Object o, boolean useEquality) {
    if (synchronizing) {
      synchronized (this) {
        return internal_remove(o, useEquality);
      }
    }
    return internal_remove(o, useEquality);
  }

  /** Removes the Object from the supplied position pos.
      @param int pos: the position from that the Object has to be removed
      @throws ArrayIndexOutOfBoundsException: if pos>=length of the vector
  */      
  public void remove(int pos) throws ArrayIndexOutOfBoundsException {
    if (synchronizing) {
      synchronized (this) {
        internal_remove(pos);
        return;
      }
    }
    internal_remove(pos);
  }

  /** Removes all preliminarily deleted elements finally from the vector.
  */
  public void pack() {
    if (synchronizing) {
      synchronized (this) {
        internal_pack(content.length);
        return;
      }
    }
    internal_pack(content.length);
  }

  /** Returns the number of elements in the vector.
      @return int: the number of elements in the vector
  */      
  public int length() {
    if (synchronizing) {
      synchronized (this) {
        return internal_length();
      }
    }
    return internal_length();
  }

  // internal functionality

  private void internal_pack(int freePos) {
    int p=0;
    if (freePos==0) p=1;
    for (int i=0; i<count; i++) {
      if (!erased[i]) {
        content[p]=content[i];
        p++;
        if (p==freePos) p++;
      }
      else {
        erased[i]=false;
      }
    }
    count-=erasedCount;
  }

  private boolean internal_enlarge(int pos) {

    if ((pos==count) && (count<content.length)) return true; // inserting at the end without packing

    if (erased[pos]) { // inserting in place of an already removed element
      erased[pos]=false;
      erasedCount--;
      return true;
    }

    int aktSize=count-erasedCount;
    if (aktSize<count) { // packing the array to create places at the end and leaving free place for the element to be inserted
      internal_pack(pos);
      return true;
    }

    // there is no help, the array has to be enlarged

    if (content.length==maxSize) return false;

    int newSize;
    if (maxSize>((int)(content.length*ENLARGE_FACTOR))) newSize=(int)(content.length*ENLARGE_FACTOR);
      else newSize=maxSize;
    Object temp[]=new Object[newSize];

    count=aktSize;

    int p=0;
    if (pos==0) p=1;
    erasedCount=0;

    for (int i=0; i<content.length; i++) {
      if (erased[i]) {
        erased[i]=false;
      }
      else {
        temp[p]=content[i];
        p++;
        if (p==pos) p++;
      }
    }

    content=temp;
    erased=new boolean[newSize];
    for (int i=0; i<newSize; i++) erased[i]=false;

    return true;
  }

  private void internal_reduce() {

    int aktSize=count-erasedCount;

    if (aktSize>((int)(content.length*REDUCE_LIMIT))) {
      if (instantRemoval) {
        internal_pack(content.length);
        return;
      }
    }
    else {
      int newSize;
      if (((int)(content.length*REDUCE_FACTOR))<MIN_SIZE) newSize=MIN_SIZE;
        else newSize=(int)(content.length*REDUCE_FACTOR);
      Object[] temp=new Object[newSize];

      count=aktSize;

      int p=0;
      erasedCount=0;

      for (int i=0; i<content.length; i++) {
        if (erased[i]) {
          erased[i]=false;
        }
        else {
          temp[p]=content[i];
          p++;
        }
      }

      content=temp;
      erased=new boolean[newSize];
      for (int i=0; i<newSize; i++) erased[i]=false;
    }
  }

  private boolean internal_add(Object o) {
    if (internal_enlarge(count)) {
      content[count]=o;
      count++;
      return true;
    }
    return false;
  }

  private int internal_firstIndex(Object o, boolean useEquality) {
    if (useEquality) {
      for (int i=0; i<count; i++) {
        if ((equality.equalsTo(o,content[i]) && (!erased[i]))) return i;
      }
    }
    else {
      for (int i=0; i<count; i++) {
        if ((o.equals(content[i])) && (!erased[i])) return i;
      }
    }
    return -1;
  }

  private int internal_lastIndex(Object o, boolean useEquality) {
    if (useEquality) {
      for (int i=(count-1); i>=0; i--) {
        if ((equality.equalsTo(o,content[i]) && (!erased[i]))) return i;
      }
    }
    else {
      for (int i=(count-1); i>=0; i--) {
        if ((o.equals(content[i])) && (!erased[i])) return i;
      }
    }
    return -1;
  }

  private int internal_count(Object o, boolean useEquality) {
    int c=0;
    if (useEquality) {
      for (int i=0; i<count; i++) {
        if ((equality.equalsTo(o,content[i]) && (!erased[i]))) c++;
      }
    }
    else {
      for (int i=0; i<count; i++) {
        if ((o.equals(content[i])) && (!erased[i])) c++;
      }
    }
    return c;
  }

  private Object internal_get(int pos) throws Exception {
    if (pos>=count) throw new ArrayIndexOutOfBoundsException(pos);
    if (erased[pos]) throw new Exception("Element ["+pos+"] already deleted. Cannot retrieve.");
    return content[pos];
  }

  private void internal_set(Object o, int pos) throws ArrayIndexOutOfBoundsException {
    if (pos>=count) throw new ArrayIndexOutOfBoundsException(pos);
    content[pos]=o;
    erased[pos]=false;
  }

  private boolean internal_insert(Object o, int pos) throws ArrayIndexOutOfBoundsException {
    if (pos>count) throw new ArrayIndexOutOfBoundsException(pos);
    if (internal_enlarge(pos)) {
      if (pos>count) throw new ArrayIndexOutOfBoundsException(pos);
      content[pos]=o;
      return true;
    }
    return false;
  }

  private boolean internal_remove(Object o, boolean useEquality) {
    boolean ret=false;
    if (useEquality) {
      for (int i=0; i<count; i++) {
        if (equality.equalsTo(o,content[i])) {
          if (!erased[i]) {
            erased[i]=true;
            erasedCount++;
            ret=true;
          }
        }
      }
    }
    else {
      for (int i=0; i<count; i++) {
        if (o.equals(content[i])) {
          if (!erased[i]) {
            erased[i]=true;
            erasedCount++;
            ret=true;
          }
        }
      }
    }
    internal_reduce();
    return ret;
  }

  private void internal_remove(int pos) throws ArrayIndexOutOfBoundsException {
    if (pos>count) throw new ArrayIndexOutOfBoundsException(pos);
    if (!erased[pos]) {
      erased[pos]=true;
      erasedCount++;
      internal_reduce();
    }
  }

  private int internal_length() {
    return (count-erasedCount);
  }
}
