package com.monica.util;

import java.math.BigInteger;
import java.io.*;
import java.util.*;

/** Class for implementing some IO processing functionality. */

public class IO {

  /** Creates a four byte String according to the SSH2 (UINT32) specification from an int.
      @param int i: the int to be converted to String
      @return String: the result
  */      
  public static String wUint32(int i) {
    char c[]=new char[4];
    c[0]=(char)((i&0xFF000000)>>>24);
    c[1]=(char)((i&0x00FF0000)>>16);
    c[2]=(char)((i&0x0000FF00)>>8);
    c[3]=(char) (i&0x000000FF);
    return new String(c);
  }

  /** Creates a four byte String from an int by writing the bytes in the low order way.
      @param int i: the int to be converted to String
      @return String: the result
  */      
  public static String wLowOrderUint32(int i) {
    char c[]=new char[4];
    c[3]=(char)((i&0xFF000000)>>>24);
    c[2]=(char)((i&0x00FF0000)>>16);
    c[1]=(char)((i&0x0000FF00)>>8);
    c[0]=(char) (i&0x000000FF);
    return new String(c);
  }

  /** Creates an eight byte String according to the SSH2 (UINT64) specification from a long.
      @param long l: the long to be converted to String
      @return String: the result
  */      
  public static String wUint64(long l) {
    char c[]=new char[8];

    long k, m;

    for (int i=0; i<8; i++) {
      k=l;
      m=l>>8;
      k-=(m<<8);
      c[7-i]=(char)k;
      l=m;
    }

    return new String(c);
  }

  /** Creates a BigInteger from a String s. s is interpreted as
      as an unsigned number in 256 radix. s.charAt(0) is MSB, s.charAt(s.length()-1) is LSB.
      (s.charAt(i)&0x00ff)>>7 is MSB and s.charAt(i)&0x0001 is LSB.
      @param String s: String to be interpreted as an unsigned number
      @return BigInteger: the result
  */      
  public static BigInteger stringAsBigInteger(String s) {
    String r=(new String("0")).concat(IO.print(s));
    return new BigInteger(r,16);
  }

  /** Coverts a String s interpreted as an unsigned number in 256 radix to a String r
      representing an unsigned number in 16 radix with the number signs
      "0"..."9" and "A"..."F". r.length()=s.length()*2. r.charAt(i*2)==HEX((s.charAt(i)&0x00f0)>>4),
      r.charAt(i*2+1)==HEX(s.charAt(i)&0x000f)
      @param String s: the String to be interpreted as an unsigned number in 256 radix
      @return String: the String interpreted as an unsigned number in 16 radix
  */      
  public static String print(String s) {
    int len=s.length();
    char c[]=new char[len*2];
    char cc, cc_low, cc_high;

    for (int i=0; i<len; i++) {
      cc=s.charAt(i);
      cc_low=(char)(cc&0x000F);
      cc_high=(char)((cc&0x00F0)>>4);
      c[i*2]=toHex(cc_high);
      c[i*2+1]=toHex(cc_low);
    }

    return new String(c);
  }
  
  /** Transforms a String in HEX form into the binary representation.
   *  @param String s: the String representation in HEX form
   *  @return String: the String in binary representation
   *  @exception Exception: if there is an error
   */  
  public static String dePrint(String s) throws Exception {
      if (s.length()%2!=0) {
          throw new Exception("Length of the String is not multiple of 2");
      }
      char c[]=new char[s.length()/2];
      char c1,c2;
      for (int i=0; i<c.length; i++) {
          c1=fromHex(s.charAt(i*2));
          c2=fromHex(s.charAt(i*2+1));
          c[i]=(char)((c1<<4)|c2);
      }
      return new String(c);
  }
  
  private static char fromHex(char c) throws Exception {
      if ((((int)c)>47) && (((int)c)<58)) {
          return ((char)(((int)c)-48));
      }
      else if ((((int)c)>64) && (((int)c)<71)) {
          return ((char)(((int)c)-55));
      }
      else {
          throw new Exception("Invalid char in the HEX form");
      }
  }
      
  
  private static char toHex(char c) {
      if (((int)c)<10) return (char)(((int)c)+48);
      return (char)(((int)c)+55);
  }
  
  /** Return the next line from an InputStream.
   *  @param InputStream is: the InputStream whose next line will be returned
   *  @return String: the next line of the InputStream, null if there is nothing else to be read
   */
  
  public static String nextLine(InputStream is) {
      byte b[];        
      String result=null;
      int data;
      boolean start=true;
      while (true) {            
          try {
              data=is.read();
          }
          catch (Exception e) {
              return result;
          }
          if (data==-1) {
              return result;
          }
          if ((data==10) || (data==13)) {
              if (!start) {
                  break;
              }
              else {
                  continue;
              }
          }
          start=false;        
          b=new byte[1];
          b[0]=(byte)data;
          if (result==null) {
              result=new String(b);
          }
          else {
              result=result.concat(new String(b));
          }
      }        
      return result;                
  }  
  
    public static String nl2br(String s) {
        s=tr(s, ""+((char)13)+((char)10), ""+((char)10));
        s=tr(s, ""+((char)13), ""+((char)10));
        s=tr(s, ""+((char)10), "<br>");
        return s;
    }
    
    public static String tr(String s, String from, String to) {
        int pos=s.indexOf(from);
        while (pos!=-1) {
            s=s.substring(0,pos)+to+s.substring(pos+from.length());
            pos=s.indexOf(from);
        }
        return s;
    }            
    
    public static String h2s(Hashtable h) {
        Enumeration keys=h.keys();
        String res=new String();
        String k, v;
        while (keys.hasMoreElements()) {
            k=(String)(keys.nextElement());
            v=(String)(h.get(k));
            res=IO.print(k)+"="+IO.print(v)+" ";
        }
        return res;
    }
    
    public static Hashtable s2h(String s) throws Exception {
        StringTokenizer tok=new StringTokenizer(s);
        Hashtable h=new Hashtable();
        String k, v, i;
        int pos;
        while (tok.hasMoreTokens()) {
            i=tok.nextToken();
            pos=i.indexOf("=");
            if (pos!=-1) {
                k=i.substring(0,pos);
                v=i.substring(pos+1);
                k=IO.dePrint(k);
                v=IO.dePrint(v);
                h.put(k, v);
            }
            else {
                throw new Exception("Invalid format.");
            }
        }
        return h;
    }
}