package com.monica.util;

import com.monica.util.*;
import com.monica.util.debug.*;

import java.sql.*;

/** This class provides functionality for creating SQL statements in an easy manner.
 */

public class SQLHelper {
    
    private static final String DP="SQLHelper";
    
    private static final String DELIM=",";
   
    /** Marks an SQL field that is not surrounded by quotes (') */
    public static final int T_NORMAL=1;
    
    /** Marks an SQL field that is surrounded by quotes (') */
    public static final int T_QNORMAL=2;
    
    /** Marks an SQL field that is surrounded by quotes (') and transformed into HEX representation */
    public static final int T_QHEX=3;
    
    private static final int S_INSERT=1;
    private static final int S_SELECT=2;
    private static final int S_UPDATE=3;
    private static final int S_DELETE=4;
            
    private String tableName;
    private int status;
    private String fieldNameL;
    private String fieldValueL;
    private String orderList;
    private String exprList;        
    
    // -----------------------------------------------------------------
    // Starters
    // -----------------------------------------------------------------

    /** Creates a new instance set up for deleting from a table.
     *  @param String tableName: the name of the table from where to delete
     *  @return SQLHelper: the new instance
     *  @exception Exception: if there was an error
     */
    public static SQLHelper delete(String tableName) throws Exception {
        if (tableName==null) {
            throw new Exception("The table name parameter may not be null.");
        }
        SQLHelper sqlH=new SQLHelper();
        sqlH.doDelete(tableName);
        return sqlH;        
    }
                    
    /** Creates a new instance set up for inserting into a table.
     *  @param String tableName: the name of the table into which to insert
     *  @return SQLHelper: the new instance
     *  @exception Exception: if there was an error
     */
    public static SQLHelper insert(String tableName) throws Exception {
        if (tableName==null) {
            throw new Exception("The table name parameter may not be null.");
        }
        SQLHelper sqlH=new SQLHelper();
        sqlH.doInsert(tableName);
        return sqlH;
    }
    
    /** Creates a new instance set up for querying from a table.
     *  @param String tableName: the name of the table from which to query
     *  @param String fieldList: comma separated list of fields to include in the result
     *  @param String orderList: comma separated list of fields to order upon in the result
     *  @return SQLHelper: the new instance
     *  @exception Exception: if there was an error
     */
    public static SQLHelper select(String tableName, String fieldList, String orderList) throws Exception {
        if (tableName==null) {
            throw new Exception("The table name parameter may not be null.");
        }
        if (fieldList==null) {
            throw new Exception("The field list parameter may not be null.");
        }        
        SQLHelper sqlH=new SQLHelper();
        sqlH.doSelect(tableName, fieldList, orderList);
        return sqlH;
    }
    
    // -----------------------------------------------------------------
    // Misc public
    // -----------------------------------------------------------------
        
    /** Counts the number of lines in a ResultSet.
     *  @param ResultSet rs: where to count the lines
     *  @return int: number of lines
     *  @exception SQLException: if there was an error
     */
    public static int numLines(ResultSet rs) throws SQLException {
        int ret=0;
        if (rs.first()) {
            ret=1;
            while (rs.next()) {
                ret++;
            }
        }
        rs.first();
        return ret;
    }
    
    // -----------------------------------------------------------------
    // Delete
    // -----------------------------------------------------------------
    
    private void doDelete(String tableName) {
        status=S_DELETE;
        this.tableName=tableName;
    }
    
    /** Adds an AND clause to the WHERE part of the SQL statement during deletion.
     *  @param String fieldNameWithOp: name of the field and operand, such as <pre>name=</pre>
     *  @param String fieldValue: the value of the right side of the expression
     *  @param int fieldType: one of T_NORMAL, T_QNORMAL, T_QHEX
     *  @exception Exception: if there is an error
     */
    public void deleteAnd(String fieldNameWithOp, String fieldValue, int type) throws Exception {
        if (status!=S_DELETE) {
            throw new Exception("Not set up for deleting.");
        }        
        _andOr(fieldNameWithOp, fieldValue, type, " AND ");
    }
    
    /** Adds an OR clause to the WHERE part of the SQL statement during deletion.
     *  @param String fieldNameWithOp: name of the field and operand, such as <pre>name=</pre>
     *  @param String fieldValue: the value of the right side of the expression
     *  @param int fieldType: one of T_NORMAL, T_QNORMAL, T_QHEX
     *  @exception Exception: if there is an error
     */
    public void deleteOr(String fieldNameWithOp, String fieldValue, int type) throws Exception {
        if (status!=S_DELETE) {
            throw new Exception("Not set up for deleting.");
        }        
        _andOr(fieldNameWithOp, fieldValue, type, " OR ");
    }
    
    /** Returns the final SQL statement.
     *  @return String: the resulting SQL statement
     *  @exception Exception: if there is an error
     */
    public String deleteFlush() throws Exception {
        if (status!=S_DELETE) {
            throw new Exception("Not set up for deleting.");
        }        
        String sql="DELETE FROM "+tableName;
        if (exprList!=null) {
            sql+=" WHERE "+exprList;
        }
        return sql;
    }
            
    // -----------------------------------------------------------------
    // Select
    // -----------------------------------------------------------------
    
    private void doSelect(String tableName, String fieldList, String orderList) {
        status=S_SELECT;
        this.tableName=tableName;
        this.fieldNameL=fieldList;
        this.orderList=orderList;
    }
    
    /** Adds an AND clause to the WHERE part of the SQL statement during querying.
     *  @param String fieldNameWithOp: name of the field and operand, such as <pre>name=</pre>
     *  @param String fieldValue: the value of the right side of the expression
     *  @param int fieldType: one of T_NORMAL, T_QNORMAL, T_QHEX
     *  @exception Exception: if there is an error
     */
    public void selectAnd(String fieldNameWithOp, String fieldValue, int type) throws Exception {
        if (status!=S_SELECT) {
            throw new Exception("Not set up for selecting.");
        }        
        _andOr(fieldNameWithOp, fieldValue, type, " AND ");
    }
    
    /** Adds an OR clause to the WHERE part of the SQL statement during querying.
     *  @param String fieldNameWithOp: name of the field and operand, such as <pre>name=</pre>
     *  @param String fieldValue: the value of the right side of the expression
     *  @param int fieldType: one of T_NORMAL, T_QNORMAL, T_QHEX
     *  @exception Exception: if there is an error
     */
    public void selectOr(String fieldNameWithOp, String fieldValue, int type) throws Exception {
        if (status!=S_SELECT) {
            throw new Exception("Not set up for selecting.");
        }        
        _andOr(fieldNameWithOp, fieldValue, type, " OR ");
    }
    
    /** Returns the final SQL statement.
     *  @return String: the resulting SQL statement
     *  @exception Exception: if there is an error
     */
    public String selectFlush() throws Exception {
        if (status!=S_SELECT) {
            throw new Exception("Not set up for selecting.");
        }        
        String sql="SELECT "+fieldNameL+" FROM "+tableName;
        if (exprList!=null) {
            sql+=" WHERE "+exprList;
        }
        if (orderList!=null) {
            sql+=" ORDER BY "+orderList;
        }
//        DebugManager.logOut(DP, sql);
        return sql;
    }
    
    /** Converts data of a given type to normal representation.
     *  @param String data: the data as from a ResultSet
     *  @param int type: one of T_NORMAL, T_QNORMAL, T_QHEX
     *  @return String: the resulting data in normal representation
     *  @exception Exception: if there is an error
     */
    public String selectGet(String data, int type) throws Exception {
        if (status!=S_SELECT) {
            throw new Exception("Not set up for selecting.");
        }        
        return _untransform(data, type);
    }
        
    // -----------------------------------------------------------------
    // Insert
    // -----------------------------------------------------------------
        
    private void doInsert(String tableName) {
        status=S_INSERT;
        this.tableName=tableName;
    }
    
    /** Adds a field into the insertion list.
     *  @param String fieldName: name of the field name
     *  @param String fieldValue: the value of the field
     *  @param int fieldType: one of T_NORMAL, T_QNORMAL, T_QHEX
     *  @exception Exception: if there is an error
     */
    public void insertValue(String fieldName, String fieldValue, int fieldType) throws Exception {
        if (status!=S_INSERT) {
            throw new Exception("Not set up for inserting.");
        }
        String t_fieldNameL=_addDelim(fieldNameL, fieldName, DELIM); 
        String tr=_transform(fieldValue, fieldType);
        String t_fieldValueL=_addDelim(fieldValueL, tr, DELIM);
        
        fieldNameL=t_fieldNameL;
        fieldValueL=t_fieldValueL;
    }
    
    /** Returns the final SQL statement.
     *  @return String: the resulting SQL statement
     *  @exception Exception: if there is an error
     */
    public String insertFlush() throws Exception {
        if (status!=S_INSERT) {
            throw new Exception("Not set up for inserting.");
        }
        String sql="INSERT INTO "+tableName+" ("+fieldNameL+") VALUES ("+fieldValueL+")";
//        DebugManager.logOut(DP, sql);
        return sql;
    }
    
    // -----------------------------------------------------------------
    // Private Common
    // -----------------------------------------------------------------

    private void _andOr(String fieldNameWithOp, String fieldValue, int type, String delim) throws Exception {
        if (fieldNameWithOp==null) {
            throw new Exception("The field name with operator parameter may not be null.");
        }
        if (fieldValue==null) {
            throw new Exception("The field value parameter may not be null.");
        }
        String tr=_transform(fieldValue, type);
        exprList=_addDelim(exprList, fieldNameWithOp+tr, delim);        
    }
       
    private String _transform(String data, int fieldType) throws Exception {
        switch (fieldType) {
            case T_NORMAL :
                return data;
            case T_QNORMAL :
                return "'"+data+"'";
            case T_QHEX :
                return "'"+IO.print(data)+"'";
            default :
                throw new Exception("Unknown field type: "+fieldType);                
        }
    }
    
    private String _untransform(String data, int fieldType) throws Exception {
        switch (fieldType) {
            case T_NORMAL :
                return data;
            case T_QNORMAL :
                return data;
            case T_QHEX :
                return IO.dePrint(data);
            default :
                throw new Exception("Unknown field type: "+fieldType);                
        }
    }
                
    private String _addDelim(String list, String item, String delim) {
        if (list==null) {
            return item;
        }
        else {
            if (item==null) {
                return list;
            }
            else {
                return list+delim+item;
            }
        }
    }            
}
