package com.monica.util;

import com.monica.interfaces.*;
import com.monica.util.*;
import com.monica.util.debug.*;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;

/** Class responsible for displaying the Splash window
*/
public class Splash implements Runnable {
    
    // -------------------------------------------------------------------
    // Constants and variables
    // -------------------------------------------------------------------
    
    public static final int ST_NORMAL=1;
    public static final int ST_PROGRESS=2;
    public static final int ST_ABOUT=3;
    
    private static final String DP="Splash";
    
    private static final String CONF_NAME="splash.conf";
    
    private static final int SLEEP_INIT=1000;
    private static final int SLEEP_NEW=250;
    
    private boolean enabled;
    private String imageFile;
    private int duration, i_height, i_width;
    private boolean infinite;
    private String windowTitle;
        
    private Splash_Window myWindow;
    private Splash_Frame myFrame;
    
    private Thread myThread;
    
    int status;
    int progressLength;
    String curProgressItem;
    int progressCount;
    
    private boolean started;
    
    // -------------------------------------------------------------------
    // Constructor
    // -------------------------------------------------------------------
    
    /** Default constructor.
        @param Configuration configuration: the customization information
    */
    public Splash(Configuration topLevel) throws Exception {        
        
        status=ST_NORMAL;
        progressCount=0;
        progressLength=0;
        curProgressItem=new String();
        
        started=false;
        
        Configuration myConfig=topLevel.getConfiguration(CONF_NAME);
        enabled=myConfig.getBoolean("splash.enabled", false);
        if (enabled) {            
            imageFile=myConfig.getString("splash.source");
            duration=myConfig.getInt("splash.duration", 1000);            
            i_width=myConfig.getInt("splash.width", 0);
            i_height=myConfig.getInt("splash.height", 0);
            infinite=myConfig.getBoolean("splash.infinite", false);
            windowTitle=myConfig.getString("splash.windowTitle", "Splash");
        }
    }    
    
    // -------------------------------------------------------------------
    // Public methods
    // -------------------------------------------------------------------
    
    public void setStyle(int s) throws Exception {
        if (started) {
            throw new Exception("Splash already started.");
        }
        if ((s==ST_NORMAL) || (s==ST_PROGRESS) || (s==ST_ABOUT)) {
            status=s;
        }
        else {
            throw new Exception("Invalid style: "+s);
        }
    }
                      
    public void start() throws Exception {
        if (enabled) {
            started=true;
            _start();
        }
    }
    
    public void setProgressLength(int pl) throws Exception {
        if (!enabled) {
            return;
        }
        if (status!=ST_PROGRESS) {
            throw new Exception("Method can only be called with style: ST_PROGRESS.");
        }
        if (pl<progressCount) {
            throw new Exception("Invalid progress count: "+pl);
        }
        progressLength=pl;
        _updateProgress();
    }
    
    public void newProgressItem(String i) throws Exception {
        if (!enabled) {
            return;
        }
        if (status!=ST_PROGRESS) {
            throw new Exception("Method can only be called with style: ST_PROGRESS.");
        }
        if (i==null) {
            throw new Exception("Progress item cannot be null.");
        }
        progressCount++;
        curProgressItem=i;
        _updateProgress();
        Thread.currentThread().sleep(SLEEP_NEW);
    }
    
    public void skipProgress(int num) throws Exception {
        if (!enabled) {
            return;
        }
        if (status!=ST_PROGRESS) {
            throw new Exception("Method can only be called with style: ST_PROGRESS.");
        }
        progressCount+=num;        
        _updateProgress();
    }
                    
    public void endProgress() throws Exception {
        if (!enabled) {
            return;
        }
        if (status!=ST_PROGRESS) {
            throw new Exception("Method can only be called with style: ST_PROGRESS.");
        }
        _endProgress();
    }
    
    // -------------------------------------------------------------------
    // private methods
    // -------------------------------------------------------------------
    
    private void _start() throws Exception {
        
        switch (status) {
            case ST_NORMAL :
                myWindow=new Splash_Window(imageFile, duration, i_width, i_height, infinite, this);        
                myWindow.setSize(new Dimension(i_width+myWindow.getInsets().left+myWindow.getInsets().right, i_height+myWindow.getInsets().top+myWindow.getInsets().bottom));
                myWindow.setVisible(true);                
                Thread.yield();        
                Thread myThread=new Thread(this);
                myThread.start();
                break;                
                
            case ST_PROGRESS :
                myFrame=new Splash_Frame(imageFile, duration, i_width, i_height, infinite, this, windowTitle);        
                myFrame.setStyle(status);
                myFrame.setSize(myFrame.getDim());
                myFrame.setVisible(true);   
                Thread.currentThread().sleep(SLEEP_INIT);
                //Thread.yield();
                break;
                
            case ST_ABOUT :
                myFrame=new Splash_Frame(imageFile, duration, i_width, i_height, infinite, this, windowTitle);                        
                myFrame.setStyle(status);
                myFrame.setSize(myFrame.getDim());
                myFrame.setVisible(true);                
                Thread.currentThread().sleep(SLEEP_INIT);
                break;
        }
    }
    
    public void run() {
        try {
            
            Thread.yield();
            myThread.sleep(duration);
            
            if (!infinite) {
                myWindow.setVisible(false);
            }
            
        }
        catch (Exception e) {
            myWindow.setVisible(false);
            DebugManager.logException(DP, e);
        }  
    }
    
    private void _updateProgress() throws Exception {
        if (myFrame!=null) {
            myFrame.updateProgress();
        }
    }
    
    private void _endProgress() throws Exception {
        myFrame.setVisible(false);
        myFrame.dispose();
    }
            
    // -------------------------------------------------------------------
    // Inner classes
    // -------------------------------------------------------------------
    
    // *** Splash_Action
    
    private class Splash_Action extends AbstractAction {
    
        private Splash_Frame owner;
        private String name;
            
        public Splash_Action(Splash_Frame o, String n) {
            owner=o;
            name=n;
        }
    
        public void actionPerformed(ActionEvent e) {
            owner.runCommand(this, name);
        }
    
        public Object getValue(String key) {
            if (key.equals("Name")) {
                return name;
            }
            return super.getValue(key);
        }                    
    }
    
    // *** Splash_Frame
    
    private class Splash_Frame extends JFrame {
        
        private static final String DP="Splash_Frame";

        private int duration;
        private boolean infinite;
        private int i_width, i_height;
        private Splash owner;
        private String windowTile;
        
        private JPanel contentPane;
        private int status;
        
        private JPanel progressPanel;
        private JProgressBar progressBar;
        private JLabel progressLabel;
        
        private JButton okButton;
        private JPanel okPanel;
        
        private Splash_Action myAction;                
                
        Splash_Frame(String imageFile, int d, int iw, int ih, boolean inf, Splash o, String wT) {                        
            
            super(windowTitle);
            
            duration=d;
            i_width=iw;
            i_height=ih;
            infinite=inf;
            owner=o;
            windowTitle=wT;
            
            DebugManager.logOut(DP, imageFile);
            
            Image im=getContentPane().getToolkit().getImage(imageFile);           
            ImageCanvas ic=new ImageCanvas(im, i_width, i_height);
            
            setLocation(new Point(100,100));
                                   
            contentPane=new JPanel();                        
            contentPane.setLayout(new BorderLayout());
            contentPane.add(ic, BorderLayout.CENTER);            
                        
            this.setContentPane(contentPane);            
            
            setResizable(false);
            
            pack();            
        }
        
        public Dimension getDim() {
            
            // panel
            
            JPanel p=null;
            switch (status) {
                case Splash.ST_ABOUT :
                    p=okPanel;
                    break;
                case Splash.ST_PROGRESS :
                    p=progressPanel;
                    break;
            }
            
            // dimension
            
            Dimension d=p.getPreferredSize();            
            
            // widths
            
            int width_g=i_width;
            int width_p=(int)(d.getWidth());
            
            // width
            
            int width;
            if (width_g>width_p) {
                width=width_g;
            }
            else {
                width=width_p;
            }
            width+=getInsets().left+getInsets().right;
            
            // heights
            
            int height_g=i_height;
            int height_p=(int)(d.getHeight());
            
            // height
            
            int height=height_g+height_p;
                            
            height+=getInsets().top+getInsets().bottom;
            
            return new Dimension(width, height);
        }
        
        public void updateProgress() throws Exception {
            progressBar.setMaximum(owner.progressLength);
            progressBar.setValue(owner.progressCount);
            progressLabel.setText(owner.curProgressItem);
        }
        
        private void setStyle(int s) {
            switch (s) {
                case ST_PROGRESS :
                    status=s;
                
                    progressPanel=new JPanel();
                    progressPanel.setLayout(new BorderLayout());
                    progressBar=new JProgressBar(0,owner.progressLength);
                    progressLabel=new JLabel(owner.curProgressItem);
                    progressLabel.setHorizontalAlignment(SwingConstants.HORIZONTAL);
                
                    progressPanel.add(progressBar, BorderLayout.NORTH);
                    progressPanel.add(progressLabel, BorderLayout.SOUTH);
                                    
                    contentPane.add(progressPanel, BorderLayout.SOUTH);
                    
                    break;
                    
                case ST_ABOUT :
                    status=s;
                
                    myAction=new Splash_Action(this, "OK");
                    JButton okButton=new JButton(myAction);
                    
                    okPanel=new JPanel();
                    okPanel.add(okButton);
                    contentPane.add(okPanel, BorderLayout.SOUTH);
                
                    break;
            }                
        }
        
        public void runCommand(Splash_Action a, String command) {
            if (command.equals("OK")) {
                setVisible(false);
            }
        }
    }
        
    // *** Splash_Window

    private class Splash_Window extends Window {
        
        private static final String DP="Splash_Window";

        private int duration;
        private boolean infinite;
        private int i_width, i_height;
        private Splash owner;
                        
        Splash_Window(String imageFile, int d, int iw, int ih, boolean inf, Splash o) {                        
            super(new Frame());
            
            duration=d;
            i_width=iw;
            i_height=ih;
            infinite=inf;
            owner=o;
            
            DebugManager.logOut(DP, imageFile);
            
            Image im=getToolkit().getImage(imageFile);            
            ImageCanvas ic=new ImageCanvas(im, i_width, i_height);            
            setLocation(new Point(100,100));
            add(ic, BorderLayout.CENTER);
            setLayout(new BorderLayout());
            
            pack();            
        }                      
    }
}
