package com.monica.util.ccdm;

/**
 *  Represents an abstract key object.
 *
 *  <b>Objects implementing this interface should have an empty constructor!</b>
 */
public interface CCDKey {
    
    /**
     *  Return the type of this key as a String.
     *  @return String: the type to this key object
     *  @exception CCDException: signaling if there was an error
     */
    public String getKeyType() throws CCDException;

    /**
     *  Returns the length of key core value in bits.
     *  @return int: the length of this key object's core value in bits
     *  @exception CCDException: signaling if there was an error
     */
    public int getKeyLength() throws CCDException;

    /**
     *  Returns the owner of this key object. If not known, 
     *  an exception will be thrown.
     *  @return CCDOwner: the owner of this object
     *  @exception CCDException: signaling if there was an error
     */
    public CCDOwner getOwner() throws CCDException;

    /**
     *  Initalises this key object with a String type raw key format.
     *  Generally this raw key format is the same as the result
     *  from the getKeyAsString() method.
     *  @param String rawData: the raw key data
     *  @exception CCDException: if the initialization fails
     */    
    public void createFromString(String rawData) throws CCDException;
    
    /**
     *  Initalises this key object with a String type core key format.
     *  Generally this core key format is the same as the result
     *  from the getKeyCore() method.
     *  @param String rawData: the core key data
     *  @exception CCDException: if the initialization fails
     */    
    public void createFromCore(String coreData) throws CCDException;
    
    /**
     *  Returns the raw format of this key as a String.
     *  @return String: the raw format of this key object
     *  @exception: signaling if there was an error
     */
    public String getKeyAsString() throws CCDException;

    /**
     *  Returns an Object represening this key object.
     *  @return Object: an object representing this key object
     *  @exception: signaling if there was an error
     */
    public Object getKeyAsObject() throws CCDException;
    
    /** 
     *  Returns the core value of the key, e.g. that part
     *  that will be used by a cipher algorithm for
     *  the cryptographic operations.
     *  @return String: the core value of the key
     *  @exception CCDException: signaling if there was an error
     */
    public String getKeyCore() throws CCDException;
    
    /**
     *  Sets the designation of this key object.
     *  @param String deignation: the designation of this key object
     *  @exception CCDException: signaling if there was an error
     */
    public void setDesignation(String designation) throws CCDException;
    
    /** 
     *  Returns the designation of this key object.
     *  @return String: the designation of this key object
     *  @exception CCDException: signaling if there was an error
     */
    public String getDesignation() throws CCDException;
}
