package com.monica.util.ccdm;

import com.monica.interfaces.*;

/**
 *  Represents an abstract signature algorithm. 
 */
public interface CCDSignatureAlgorithm extends Configureable {
   
    /** 
     *  Returns the signature algorithm's name.
     *  @return String: the signature algorithm's name
     *  @exception CCDException: signaling if there was an error
     */
    public String getName() throws CCDException;

    /**
     *  Returns the length of the signature's core value in
     *  bits according to the current mode of the signature algorithm.
     *  @return int: the signature core value's length in bits
     *  @exception CCDExcepton: signaling if there was an error
     */
    public int getSignatureLength() throws CCDException;
    
    /**
     *  Returns the supported modes as an array of signature
     *  algorithm parameter descriptors.
     *  @return CCDSignatureParamDesc[]: the supported modes
     *  @exception CCDException: signaling if there was an error
     */
    public CCDSignatureParamDesc[] supportedModes() throws CCDException;
    
    /**
     *  Initalizes this signature algorithm to use the given key
     *  for the sign and verify operations.
     *  @param CCDKey key: the key to be used in the future for the sign
     *  and verify operations
     *  @return CCDSignatureParamDesc: the resulting mode according to the supplied key
     *  @exception CCDException: signaling if there was an error
     */
    public CCDSignatureParamDesc init(CCDKey key) throws CCDException;    

    /**
     *  Signs a supplied String according to the current mode of the
     *  signature algorithm and the initialization key.
     *  @param String data: the String to be signed
     *  @return CCDSignature: the resulting signature for the supplied data
     *  @exception CCDException: signaling if there was an error
     */
    public CCDSignature sign(String data) throws CCDException;
    
    /**
     *  Verifyies the authenticity of a signature according to the current
     *  mode of the signature algorithm.
     *  @param String original: the original String to what the signature should match
     *  @param CCDSignature signature: the to-be-verified signature
     *  @return boolen: whether the signature is authentic
     *  @exception CCDException: signaling if there was an error     
     */
    public boolean verify(String original, CCDSignature signature) throws CCDException;        
}
