package com.monica.util.ccdm.dss;

import com.monica.util.ccdm.*;
import com.monica.util.ccdm.dss.*;
import com.monica.util.ccdm.impl.*;

import com.monica.util.debug.*;

import com.monica.interfaces.*;

import java.math.*;

public class CCDDSS implements CCDSignatureAlgorithm {
    
    public static final String NAME="DSS";
    
    private static final String NL=""+((char)13)+((char)10);
    
    private static final String CONF_NAME="CCD.DSS";
    
    private static final String DP="CCDDSS";
    
    private static final CCDSignatureParamDesc mFull=
        new CCDSimpleSignatureParamDesc("DSS Full Mode", 40*8, true, true);
    private static final CCDSignatureParamDesc mSign=
        new CCDSimpleSignatureParamDesc("DSS Sign Mode", 40*8, true, false);
    private static final CCDSignatureParamDesc mVerify=
        new CCDSimpleSignatureParamDesc("DSS Verify Mode", 40*8, false, true);
    
    private static final CCDSignatureParamDesc[] sModes={mFull, mSign, mVerify};
    
    private static final String ERR_DSS_E="The passed DSS object must not be null.";
    
    private static final String SIG_CLASS="com.monica.util.ccdm.dss.CCDDSSSignature";
    
    private Configuration configuration, myConfig;
    private DSS myDSS;
    private CCDOwner whoAmI;
    
    public CCDDSS() {
        myDSS=new DSS();
    }
    
    public CCDDSS(DSS dss, CCDKey key) throws CCDException {
        if (dss==null) {
            throw new CCDException(ERR_DSS_E);
        }        
        myDSS=dss;
        if (key!=null) {
            init(key);                            
        }
        else {
            _myMode();
        }        
    }
    
    // -----------------------------------------------------------------
    // CCDSignatureAlgorithm
    // -----------------------------------------------------------------
        
    public String getName() {
        return NAME;
    }
    
    public int getSignatureLength() {
        return 40*8;
    }
    
    public CCDSignatureParamDesc[] supportedModes() {
        return sModes;
    }
           
    public CCDSignatureParamDesc init(CCDKey key) throws CCDException {
        try {
            // public key
            CCDDSSPublicKey k=(CCDDSSPublicKey)key;
            
            DSS_common d_c;
            DSS_public d_p;
            
            Object o[]=(Object[]) k.getKeyAsObject();
            
            d_c=(DSS_common)o[0];
            d_p=(DSS_public)o[1];
                        
            myDSS.setCommon(d_c);
            myDSS.setPublic(d_p);
            
            whoAmI=(CCDOwner)o[2];
            
            return _myMode();
        }
        catch (Exception e) {
            try {
                // private key
                CCDDSSPrivateKey k=(CCDDSSPrivateKey)key;
                
                DSS_common d_c;
                DSS_private d_p;
                
                Object o[]=(Object[])k.getKeyAsObject();
                
                d_c=(DSS_common)o[0];
                d_p=(DSS_private)o[1];
                
                myDSS.setCommon(d_c);
                myDSS.setPrivate(d_p);
                
                whoAmI=(CCDOwner)o[3];
                
                return _myMode();
            }    
            catch (Exception ee) {
                try {
                    // key pair
                    CCDDSSKeyPair k=(CCDDSSKeyPair)key;
                    
                    DSS_common d_c;
                    DSS_public d_pu;
                    DSS_private d_pr;
                    
                    Object o[]=(Object[])k.getKeyAsObject();
                    
                    d_c=(DSS_common)o[0];
                    d_pr=(DSS_private)o[1];
                    d_pu=(DSS_public)o[2];
                    
                    myDSS.setCommon(d_c);
                    myDSS.setPrivate(d_pr);
                    myDSS.setPublic(d_pu);
                    
                    whoAmI=(CCDOwner)o[3];
                    return _myMode();
                }
                catch (Exception eee) {
                    throw new CCDException("Invalid key type: "+key.getClass().getName());
                }
            }            
        }
    }
    
    public CCDSignature sign(String data) throws CCDException {        
        if (myDSS.getPrivate()!=null) {
            return new CCDDSSSignature(myDSS.sign(data), whoAmI);
        }
        else {
            throw new CCDException("No key has been set for signing set yet.");
        }
    }
    
    public boolean verify(String original, CCDSignature signature) throws CCDException {
        if (myDSS.getPublic()==null) {
            throw new CCDException("No key has been set for verifying yet.");
        }
        try {
            if (!signature.getClass().getName().equals(SIG_CLASS)) {
                throw new CCDException("Invalid signature class. Cannot verify."+signature.getClass().getName());
            }
            Object oa[]=(Object[])signature.getSignatureAsObject();
            // -- TODO: key owner and signature owner check
            DSS_signature sig=(DSS_signature)oa[0];
            return myDSS.verify(original, sig);                        
        }
        catch (Exception e) {
            throw new CCDException(e.toString());
        }            
    }
    
    // -----------------------------------------------------------------
    // Configureable
    // -----------------------------------------------------------------
    
    public void setConfiguration(Configuration configuration) {
        this.configuration=configuration;
        _rereadConfig();
    }
    
    public Configuration getConfiguration() {
        return configuration;
    }
    
    // -----------------------------------------------------------------
    // Private methods
    // -----------------------------------------------------------------
        
    protected CCDSignatureParamDesc _myMode() {
        if (myDSS.getPublic()!=null) {
            if (myDSS.getPrivate()!=null) {
                return mFull;
            }
            else {
                return mVerify;
            }
        }
        if (myDSS.getPrivate()!=null) {
            if (myDSS.getPublic()!=null) {
                return mFull;
            }
            else {
                return mSign;
            }
        }
        return null;
    }
    
    protected void _rereadConfig() {
        try {
            myConfig=configuration.getConfiguration(CONF_NAME);
        }
        catch (Exception e) {
            DebugManager.logOut(DP, "Exception occured in _rereadConfig: "+e.toString());
        }
    }                                
}
