package com.monica.util.ccdm.dss;

import com.monica.util.ccdm.dss.*;
import com.monica.interfaces.*;
import com.monica.util.configuration.*;
import com.monica.util.hash.*;
import com.monica.util.debug.*;

import java.math.*;
import java.util.*;

/** This class provides functionaliy of the DSS signature algorithm as specified in FIPS 186-2. */

public class DSS {
    
  private static final String DP="DSS";

  DSS_private priv;
  DSS_public pub;
  DSS_common common;
  Random rnd;
  SHA1 sha1;
  
  // +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  // Setting UP
  // +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  
  // Common
  
  /** Returns the common part of the currently active DSS key.
   *  @return DSS_common: the common part of the currently active DSS key
   */
  public DSS_common getCommon() {
      return common;
  }
  
  /** Sets the common part of the active DSS key.
   *  @param DSS_common common: the to be set common part of the active DSS key
   *  @return DSS_common: the set common part of the active DSS key
   */
  public DSS_common setCommon(DSS_common common) {
    this.common=common;
    return common;
  }
  
  /** Sets the common part of the active DSS key according to the specified parameters.
   *  @param BigInteger p: the p parameter
   *  @param BigInteger q: the q parameter
   *  @param BigInteger g: the g parameter
   *  @return DSS_common: the set common part of the active DSS key
   */
  public DSS_common setCommon(BigInteger p, BigInteger q, BigInteger g) {
    common=new DSS_common();
    common.p=p;
    common.q=q;
    common.g=g;
    return common;
  }   
  
  /** Sets the common part of the active DSS key according to the values to be found in the supplied Configuration.
   *  @param Configuration conf: the Configuration from which the p, q, g parameters have to be taken
   *  @return DSS_common: the set common part of the active DSS key
   */
  public DSS_common setCommon(Configuration conf) {
    try {
      common=new DSS_common();
      common.p=new BigInteger(conf.getString("p"), 16);
      common.q=new BigInteger(conf.getString("q"), 16);
      common.g=new BigInteger(conf.getString("g"), 16);
      return common;      
    }
    catch (Exception e) {
      DebugManager.logException(DP, e);  
      common=null;
      return null;
    }  
  }
  
  // public
  
  /** Returns the public part of the currently active DSS key.
   *  @return DSS_public: the public part of the currently active DSS key
   */
  public DSS_public getPublic() {
      return pub;
  }
  
  /** Sets the public part of the active DSS key.
   *  @param DSS_public pub: the to be set public part of the active DSS key
   *  @return DSS_public: the set public part of the active DSS key
   */
  public DSS_public setPublic(DSS_public pub) {
    this.pub=pub;
    return pub;
  }
  
  /** Sets the public part of the active DSS key according to the specified parameters.
   *  @param BigInteger y: the y parameter
   *  @return DSS_public: the set public part of the active DSS key
   */
  public DSS_public setPublic(BigInteger y) {
    pub=new DSS_public();
    pub.y=y;
    return pub;
  }
  
  /** Sets the public part of the active DSS key according to the values to be found in the supplied Configuration.
   *  @param Configuration conf: the Configuration from which the y parameter has to be taken
   *  @return DSS_public: the set public part of the active DSS key
   */
  public DSS_public setPublic(Configuration conf) {
    try {
      pub=new DSS_public();
      pub.y=new BigInteger(conf.getString("y"),16);
      return pub;
    }
    catch (Exception e) {
      DebugManager.logException(DP, e);  
      pub=null;
      return null;
    }
  }
  
  /** Adds the common and public part of the active DSS key to the specified Configuration.
   *  @param Configuration conf: the Configuration wo which the p, q, g, y parameters will be added
   */
  public void addPublicToConfiguration(Configuration conf) {
      conf.set("p", common.p.toString(16));
      conf.set("q", common.q.toString(16));
      conf.set("g", common.g.toString(16));
      conf.set("y", pub.y.toString(16));      
  }
  
  // private
  
  /** Returns the private part of the currently active DSS key.
   *  @return DSS_private: the private part of the currently active DSS key
   */
  public DSS_private getPrivate() {
    return priv;
  }
  
  /** Sets the private part of the active DSS key.
   *  @param DSS_private priv: the to be set private part of the active DSS key
   *  @return DSS_private: the set private part of the active DSS key
   */
  public DSS_private setPrivate(DSS_private priv) {
    this.priv=priv;
    return priv;
  }
  
  /** Sets the private part of the active DSS key according to the specified parameters.
   *  @param BigInteger x: the x parameter
   *  @return DSS_private: the set private part of the active DSS key
   */
  public DSS_private setPrivate(BigInteger x) {
    priv=new DSS_private();
    priv.x=x;
    return priv;
  }
  
  /** Sets the private part of the active DSS key according to the values to be found in the supplied Configuration.
   *  @param Configuration conf: the Configuration from which the x parameter has to be taken
   *  @return DSS_private: the set private part of the active DSS key
   */
  public DSS_private setPrivate(Configuration conf) {
    try {
      priv=new DSS_private();
      priv.x=new BigInteger(conf.getString("x"),16);
      return priv;
    }
    catch (Exception e) {
      DebugManager.logException(DP, e);  
      priv=null;
      return null;
    }    
  }
  
  /** Adds the common and private part of the active DSS key to the specified Configuration.
   *  @param Configuration conf: the Configuration wo which the p, q, g, x parameters will be added
   */
  public void addPrivateToConfiguration(Configuration conf) {
      conf.set("p", common.p.toString(16));
      conf.set("q", common.q.toString(16));
      conf.set("g", common.g.toString(16));
      conf.set("x", priv.x.toString(16));      
  }
    
  // +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  // Constructor
  // +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  
  /** Creates a DSS object. */  
  public DSS() {
      rnd=new Random();
      sha1=new SHA1();
  }

  // +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  // Signature process
  // +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    
  /** Signs a String according to the active DSS key. Only the private and common parts of the active DSS key have to be initialised.
   *  @param String toSign: the String to be signed
   *  @return DSS_signature: the signature ot toSign according to the active DSS key
   */
  public DSS_signature sign(String toSign) {
      DSS_signature sig=new DSS_signature();
      BigInteger k=new BigInteger(159,0,rnd);     
      sig.r=common.g.modPow(k, common.p).mod(common.q);
      BigInteger temp=sha1.hashAsBigInteger(toSign).add(priv.x.multiply(sig.r));
      sig.s=k.modInverse(common.q).multiply(temp).mod(common.q);
      return sig;
  }
  
  /** Verifies a signature on a String according to the active DSS key. Only the public and common parts of the active DSS key have to be initialised.
   *  @param String toSign: the String, whose signature should be checked
   *  @param DSS_signature: the to be checked signature
   *  @return boolean: whether the signature is valid
   */
  public boolean verify(String toSign, DSS_signature signature) {
      BigInteger w,hash,t1,t2,u1,u2,v;
      w=signature.s.modInverse(common.q);
      hash=sha1.hashAsBigInteger(toSign);
      u1=(w.multiply(hash)).mod(common.q);
      u2=(signature.r.multiply(w)).mod(common.q);
      t1=common.g.modPow(u1, common.p);
      t2=pub.y.modPow(u2, common.p);
      v=((t1.multiply(t2)).mod(common.p)).mod(common.q);
      return signature.r.equals(v);
  }  
  
  
}