package com.monica.util.ccdm.rsa;

import com.monica.util.ccdm.*;
import com.monica.util.debug.*;
import com.monica.util.ccdm.impl.*;

import java.math.BigInteger;
import java.util.StringTokenizer;

/**	This class represents a basic RSA key.
*/

public class CCDRSAKey implements CCDKey {
    
    private static final String TYPE="RSAKey";
    
    private static final String DP="RSAKey";
    
    private static final String NL=""+((char)13)+((char)10);     
    
    private static final String ERR_NI="Key not initialized.";

    private BigInteger modulus;
    private BigInteger exponent;
        
    private CCDOwner myOwner;
    
    private String myDesignation;
        
    public CCDRSAKey(BigInteger in_modulus, BigInteger in_exponent, CCDOwner o) throws CCDException {	
        if (in_modulus==null) {
            throw new CCDException("The supplied modulus cannot be null.");
        }
        if (in_exponent==null) {
            throw new CCDException("The supplied exponent cannot be null.");
        }
        modulus=in_modulus;
        exponent=in_exponent;	
        myOwner=o;
    }
    
    public CCDRSAKey() {
    }

    // ---------------------------------------------------------------------
    // CCDKey
    // ---------------------------------------------------------------------
        
    public void setDesignation(String designation) throws CCDException {
        myDesignation=designation;
    }

    public String getDesignation() throws CCDException {
        return myDesignation;
    }
    
    public CCDOwner getOwner() throws CCDException {
        return myOwner;
    }
    
    public String getKeyType() throws CCDException {
        return TYPE;
    }
    
    public int getKeyLength() throws CCDException {
        if ((modulus==null) || (exponent==null)) {
            throw new CCDException(ERR_NI);
        }
        return exponent.bitLength();
    }
    
    public Object getKeyAsObject() throws CCDException {
        if ((modulus==null) || (exponent==null)) {
            throw new CCDException(ERR_NI);
        }
        Object oa[]=new Object[4];
        oa[0]=modulus;
        oa[1]=exponent;
        oa[2]=myOwner;
        oa[3]=myDesignation;
        return oa;
    }
        
    public String getKeyAsString() throws CCDException {
        if ((modulus==null) || (exponent==null)) {
            throw new CCDException(ERR_NI);
        }
        String ret=new String();
        ret+="e="+exponent.toString(16)+NL;
        ret+="m="+modulus.toString(16)+NL;
        if (myOwner!=null) {
            if (myOwner.getCN()!=null) {
                ret+="cn="+myOwner.getCN()+NL;
            }
            if (myOwner.getDN()!=null) {
                ret+="dn="+myOwner.getDN()+NL;
            }
            if (myOwner.getUID()!=null) {
                ret+="uid="+myOwner.getUID()+NL;
            }
        }
        if (myDesignation!=null) {
            ret+="designation="+myDesignation+NL;
        }
        return ret;        
    }
        
    public String getKeyCore() throws CCDException {
        if ((modulus==null) || (exponent==null)) {
            throw new CCDException(ERR_NI);
        }
        return new String(modulus.toString()+" "+exponent.toString());
    }
    
    public void createFromCore(String coreData) throws CCDException {
        try {
            StringTokenizer stringTokenizer=new StringTokenizer(coreData);
            modulus=new BigInteger(stringTokenizer.nextToken());
            exponent=new BigInteger(stringTokenizer.nextToken());
        }
        catch (Exception e) {
            throw new CCDException(e);
        }
    }
    
    public void createFromString(String rawData) throws CCDException {
       
        int oc=0;
        
        BigInteger t_exp=null, t_mod=null;
        CCDBasicOwner t_o=new CCDBasicOwner();
        String t_d=null;
        
        StringTokenizer tok=new StringTokenizer(rawData, NL);
        String[] sa=_parse(tok);
                
        while (sa!=null) {
            if (sa[0].equals("e")) {
                t_exp=new BigInteger(sa[1],16);
            }
            else if (sa[0].equals("m")) {
                t_mod=new BigInteger(sa[1],16);
            }
            else if (sa[0].equals("uid")) {
                t_o.setUID(sa[1]);
                oc++;
            }
            else if (sa[0].equals("cn")) {
                t_o.setCN(sa[1]);
                oc++;
            }
            else if (sa[0].equals("dn")) {
                t_o.setDN(sa[1]);
                oc++;
            }      
            else if (sa[0].equals("designation")) {
                t_d=sa[1];
            }
            sa=_parse(tok);
        }
        if ((t_exp!=null) && (t_mod!=null)) {
            exponent=t_exp;
            modulus=t_mod;
            myDesignation=t_d;                        
            if (oc>0) {
                myOwner=t_o;
            }
        }
        else {
            throw new CCDException("Could not create key from this String.");
        }
    }
    	
    // ---------------------------------------------------------------------
    // Private methods
    // ---------------------------------------------------------------------
        
    private String[] _parse(StringTokenizer tok) {
        if (tok.hasMoreTokens()) {
            String s=tok.nextToken();
            int pos=s.indexOf("=");
            String[] sa=new String[2];
            sa[0]=s.substring(0,pos);
            sa[1]=s.substring(pos+1);
            return sa;            
        }
        return null;
    }
                                                        
}