package com.monica.util.ccdm.rsa;

import com.monica.util.ccdm.*;
import com.monica.util.ccdm.impl.*;
import com.monica.util.*;

import java.util.*;
import java.math.*;

public class CCDRSAKeyPair implements CCDAsymmetricKeyPair {
    
    // -------------------------------------------------------------------------
    // Variables and constants
    // -------------------------------------------------------------------------
        
    private static final String DP="RSAKeyPair";
    
    private static final String TYPE="RSAKeyPair";
    
    private static final String ERR_NI="Key not initialized.";
    
    private static final String NL=""+((char)13)+((char)10);     
    
    private CCDOwner myOwner;
    private String myDesignation;
    
    private CCDRSAPrivateKey priv;
    private CCDRSAPublicKey pub;
    
    // -------------------------------------------------------------------------
    // Constructors
    // -------------------------------------------------------------------------
    
    public CCDRSAKeyPair() {
    }
    
    public CCDRSAKeyPair(CCDRSAPublicKey in_pub, CCDRSAPrivateKey in_priv, CCDOwner o) throws CCDException {
        if (in_pub==null) {
            throw new CCDException("Supplied public key may not be null.");
        }
        if (in_priv==null) {
            throw new CCDException("Supplied private key may not be null.");
        }
        pub=in_pub;
        priv=in_priv;
        myOwner=o;
    }
    
    public CCDRSAKeyPair(BigInteger mod, BigInteger pub_exp, BigInteger priv_exp, CCDOwner o) throws CCDException {
        if (pub_exp==null) {
            throw new CCDException("Supplied public exponent may not be null.");
        }
        if (priv_exp==null) {
            throw new CCDException("Supplied private exponent may not be null.");
        }
        if (mod==null) {
            throw new CCDException("Supplied modulus may not be null.");
        }
        pub=new CCDRSAPublicKey(mod, pub_exp, o);
        priv=new CCDRSAPrivateKey(mod, priv_exp, o);
        myOwner=o;
    }
    
    // -------------------------------------------------------------------------
    // CCDAsymmetricKeyPair
    // -------------------------------------------------------------------------
    
    public CCDPublicKey getPublic() throws CCDException {
        return pub;
    }

    public CCDPrivateKey getPrivate() throws CCDException {
        return priv;
    }
    
    // -------------------------------------------------------------------------
    // CCDKey
    // -------------------------------------------------------------------------
    
    public void setDesignation(String designation) throws CCDException {
        myDesignation=designation;
    }

    public String getDesignation() throws CCDException {
        return myDesignation;
    }
    
    public String getKeyType() throws CCDException {
        return TYPE;
    }

    public CCDOwner getOwner() throws CCDException {
        return myOwner;
    }

    public int getKeyLength() throws CCDException {
        if ((pub==null) || (priv==null)) {
            throw new CCDException(ERR_NI);
        }
        return pub.getKeyLength();
    }

    public Object getKeyAsObject() throws CCDException {
        if ((pub==null) || (priv==null)) {
            throw new CCDException(ERR_NI);
        }
        Object oa[]=new Object[4];
        oa[0]=pub;
        oa[1]=priv;
        oa[2]=myOwner;
        oa[4]=myDesignation;
        return oa;
    }
    
    public String getKeyAsString() throws CCDException {
        if ((pub==null) || (priv==null)) {
            throw new CCDException(ERR_NI);
        }
        String ret=new String();
        ret+="pub="+IO.print(pub.getKeyAsString())+NL;
        ret+="priv="+IO.print(priv.getKeyAsString())+NL;
        if (myOwner!=null) {
            if (myOwner.getCN()!=null) {
                ret+="cn="+myOwner.getCN()+NL;
            }
            if (myOwner.getDN()!=null) {
                ret+="dn="+myOwner.getDN()+NL;
            }
            if (myOwner.getUID()!=null) {
                ret+="uid="+myOwner.getUID()+NL;
            }
        }
        if (myDesignation!=null) {
            ret+="designation="+myDesignation+NL;
        }
        return ret;        
    }
    
    public String getKeyCore() throws CCDException {
        if ((pub==null) || (priv==null)) {
            throw new CCDException(ERR_NI);
        }
        return IO.print(pub.getKeyCore())+" "+IO.print(priv.getKeyCore());
    }

    public void createFromCore(String coreData) throws CCDException {
        try {
            StringTokenizer tok=new StringTokenizer(coreData, " ");
            String pubCore=IO.dePrint(tok.nextToken());
            String privCore=IO.dePrint(tok.nextToken());
            CCDRSAPublicKey t_pub=new CCDRSAPublicKey();
            t_pub.createFromCore(pubCore);
            CCDRSAPrivateKey t_priv=new CCDRSAPrivateKey();
            t_priv.createFromCore(privCore);
            pub=t_pub;
            priv=t_priv;            
        }
        catch (Exception e) {
            throw new CCDException(e);
        }
    }
    
    public void createFromString(String rawData) throws CCDException {
        
        try {
       
            int oc=0;
        
            CCDRSAPrivateKey t_priv=null;
            CCDRSAPublicKey t_pub=null;
            CCDBasicOwner t_o=new CCDBasicOwner();
            String t_d=null;
        
            StringTokenizer tok=new StringTokenizer(rawData, NL);
            String[] sa=_parse(tok);
                
            while (sa!=null) {
                if (sa[0].equals("pub")) {
                    t_pub=new CCDRSAPublicKey();
                    t_pub.createFromString(IO.dePrint(sa[1]));
                }
                else if (sa[0].equals("priv")) {
                    t_priv=new CCDRSAPrivateKey();
                    t_priv.createFromString(IO.dePrint(sa[1]));
                }
                else if (sa[0].equals("uid")) {
                    t_o.setUID(sa[1]);
                    oc++;
                }
                else if (sa[0].equals("cn")) {
                    t_o.setCN(sa[1]);
                    oc++;
                }
                else if (sa[0].equals("dn")) {
                    t_o.setDN(sa[1]);
                    oc++;
                }      
                else if (sa[0].equals("designation")) {
                    t_d=sa[1];
                }
                sa=_parse(tok);
            }
            if ((t_pub!=null) && (t_priv!=null)) {
                pub=t_pub;
                priv=t_priv;
                myDesignation=t_d;            
                if (oc>0) {
                    myOwner=t_o;
                }
            }
            else {
                throw new CCDException("Could not create key from this String.");
            }
        }
        catch (CCDException e) {
            throw e;
        }
        catch (Exception e) {
            throw new CCDException(e);
        }
    }

    // -------------------------------------------------------------------------
    // Private emthods
    // -------------------------------------------------------------------------

    private String[] _parse(StringTokenizer tok) {
        if (tok.hasMoreTokens()) {
            String s=tok.nextToken();
            int pos=s.indexOf("=");
            String[] sa=new String[2];
            sa[0]=s.substring(0,pos);
            sa[1]=s.substring(pos+1);
            return sa;            
        }
        return null;
    }
    
}
