package com.monica.util.configuration;

import com.monica.util.debug.*;

import com.monica.interfaces.Configuration;
import com.monica.util.ByteArrayBuffer;

import java.util.Hashtable;
import java.util.Properties;
import java.io.InputStream;
import java.io.OutputStream;

/** This class implements the Configuration interface to provide customization possibility
    for TM classes. This class reads customization information from a file called TM.conf,
    which has to be located in the current directory.<br><br>
    The configuration file can have the following stucure:<br>
    <ul>
      <li>lines beginning with # are recognized as comments and therefore ignored
      <li>lines not beginning with # are recognized as configuration lines<br><br>
        configuration lines have the following structure:<br>
        <pre>keyword=value</pre>
        Values can be Strings, ints, and booleans. Booleans are represented with the
        strings "true" and "false". These values can be referenced int he get methods
        of the Configuration with the id matching the keyword.
    </ul>
*/
public class InputStreamConfiguration implements Configuration {
    
  private static final String DP="InputStreamConfiguration";  

  // private variables
  private Properties prop;
  private Hashtable hash;
  private int status;
  private ByteArrayBuffer buf;
  
  /** The InputStream from which the configuration settings will be read.
  */
  protected InputStream iS;  

  // constants
  private static final int ST_NO=0;
  private static final int ST_YES=1;

  private static final String NAME="InputStreamConfiguration";
  private static final String VERSION="0.3";
  
  private static final String ERR_NO_SUCH_STREAM="No such stream : ";
  
  /** Method that has to be called first in the constructors of all subclasses.
  */
  protected void pre_init() {
    status=ST_NO;
    hash=new Hashtable();
  }

  /** Method that has to be called last in the constructors of all subclasses.
  */
  protected void post_init() {
    try {
      prop=new Properties();
      prop.load(iS);
      status=ST_YES;
    }
    catch (Exception e) {
      e.printStackTrace();
    }
  }
  
  /** Creates an instance according to the InputStream specified.
   *  @param InputStream i: the specified InputStream
   */
  public InputStreamConfiguration(InputStream i) {
    pre_init();
    iS=i;
    post_init();
  }    
  
  /** Creates a not-set-up instance.
   */
  public InputStreamConfiguration() {
      status=ST_NO;
  }  

  public String getConfigurationName() {
    return NAME;
  }

  public String getConfigurationVersion() {
    return VERSION;
  }

  public int getInt(String id) throws ConfigurationException {
    String temp;
    if (status==ST_YES) {
      temp=prop.getProperty(id);
      if (temp!=null) {
        return Integer.parseInt(temp);
      }
    }
    throw new ConfigurationException("Key not found in conf file - "+id);
  }

  public boolean getBoolean(String id) throws ConfigurationException {
    String temp;
    if (status==ST_YES) {
      temp=prop.getProperty(id);
      if (temp!=null) {
        if (temp.equals("true")) {
          return true;
        }
        if (temp.equals("false")) {
          return false;
        }
      }
    }
    throw new ConfigurationException("Key not found in conf file - "+id);
  }

  public String getString(String id) throws ConfigurationException {
    String temp;
    if (status==ST_YES) {
      temp=prop.getProperty(id);
      if (temp!=null) {
        return temp;
      }
    }
    throw new ConfigurationException("Key not found in conf file - "+id);
  }

  public int getInt(String id, int def) {
    try {
      return getInt(id);
    }
    catch (Exception e) {
      return def;
    }
  }

  public boolean getBoolean(String id, boolean def) {
    try {
      return getBoolean(id);
    }
    catch (Exception e) {
      return def;
    }
  }

  public String getString(String id, String def) {
    try {
      return getString(id);
    }
    catch (Exception e) {
      return def;
    }
  }

  public int getIntFromStrings(String id, String[] possibilities, int[] values, int def) {
    String temp;
    int i;
    try {
      temp=getString(id);
      i=0;
      while (i<possibilities.length) {
        if (possibilities[i].equals(temp)) {
          return values[i];
        }
        i++;
      }
      return def;
    }
    catch (Exception e) {
      return def;
    }
  }
  
  public void setStream(String ID, InputStream stream) {
    try {
      buf=new ByteArrayBuffer();
      byte[] b;
      int r;
      while (stream.available()>0) {
        b=new byte[stream.available()];
        r=stream.read(b);
        if (r==-1) break;
        buf.add(b,r);
      }
      hash.put(ID,buf.emptyBuffer());
    }
    catch (Exception e) {
    }
  }

  public InputStream getStream(String ID) throws ConfigurationException {
    Object stream=hash.get(ID);
    if (stream!=null) {
      return new com.monica.util.StringInputStream(new String((byte[])stream));
    }
    throw new ConfigurationException(ERR_NO_SUCH_STREAM+ID);
  }
  
  public Configuration getConfiguration(String ID) throws ConfigurationException {
      return new InputStreamConfiguration(getStream(ID));
  }
  
  public boolean save(OutputStream os) {
    try {
      if (status==ST_YES) {
        prop.store(os, null);        
        return true;  
      }
    }
    catch (Exception e) {
    }
    return false;
  }
  
  public void set(String ID, String value) {
    prop.setProperty(ID, value);  
  }
  
  public void remove(String ID) {
    prop.remove(ID);
  }
  
  public Object clone() {
      InputStreamConfiguration ret=new InputStreamConfiguration();
      ret.prop=(Properties)(prop.clone());
      ret.hash=(Hashtable)(hash.clone());
      ret.status=ST_YES;
      return ret;
  }            
}