package com.monica.util.debug;

import com.monica.interfaces.*;
import com.monica.util.*;

import java.util.*;
import java.io.*;

/** Facade for providing debugging functionality. <br><br>
 *
 *  Configureable values fromt he <b>debug.conf</b> Configuration:
 *  <ul>
 *      <li><b>debug</b>: true/false, whether to enable debugging
 *      <li><b>debugOutputClass</b>: fully qualified class of a DebugOutput
 *      <li><b>debugDelimiter</b>: delimiter to be added between debug requester name and debug info
 *      <li><b>useErr</b>: true/false, use standard Err or redirect all debug info from Err to Out
 *      <li><b>filtering</b>: true/false, whether to use filtering (not all debug info will appear in the debug output)
 *      <li><b>filterExceptions</b>: true/false, whether to filter exception debugging
 *      <li><b>filterOut</b>: comma separated list of ebug requester names, which will be filtered
 *          out; usage of * wildchar allowed
 *  </ul>
 */
public class DebugManager {
    
    // ------------------------------------------------------------------
    // Variables and constants
    // ------------------------------------------------------------------
    
    private static DebugOutput myOut;
    private static Configuration topLevel, myConfig;
    private static String delim;
    private static boolean useErr;
    private static Class EHC;
    
    private static final String CONF_DEBUG="debug.conf";    
    private static final boolean CONF_DEF_USE_ERR=true;
    
    private static final String DEF_DELIM="> ";
    
    private static final String EXCEPTION_HOLDER_CLASS="com.monica.util.ExceptionHolder";    
    
    private static boolean filterNormal;
    private static boolean filterExceptions;
    private static Vector fn_fullMatch, fn_partialMatch;
    private static Vector fe_fullMatch, fe_partialMatch;
    
    private static Object[] a_fn_fullMatch, a_fn_partialMatch;
    private static Object[] a_fe_fullMatch, a_fe_partialMatch;
    
    private static boolean fullFeature;
    private static boolean doDebug;
            
    // ------------------------------------------------------------------
    // startUp
    // ------------------------------------------------------------------
    
    static {
        String version=System.getProperty("java.version");
        int pos;
        int major, minor;
        pos=version.indexOf(".");
        major=Integer.parseInt(version.substring(0,pos));
        if (major==1) {
            version=version.substring(pos+1);
            pos=version.indexOf(".");
            minor=Integer.parseInt(version.substring(0,pos));
            if (minor>=4) {
                fullFeature=true;
            }
            else {
                fullFeature=false;
            }
        }
        else {
            fullFeature=false;
        }
        doDebug=false;
        
        if (!fullFeature) {
            System.out.println("*****************************************************");
            System.out.println("*    DebugManager is running in restricted mode.    *");
            System.out.println("*****************************************************");
            System.out.println("     Obtain J2SE 1.4 at least to get all features    ");
            System.out.println("*****************************************************");
        }
    }
                    
    /** Sets up the DebugManager.
     *  @param Configuration topLevel: the topLevel Configuration as returned from the ConfigurationLoader
     */
    public static void startUp(Configuration tL) {
        try {
            topLevel=tL;
            myConfig=topLevel.getConfiguration(CONF_DEBUG);            
            if (myConfig.getBoolean("debug")) { // debugging enabled
                doDebug=true;
                EHC=Class.forName(EXCEPTION_HOLDER_CLASS);
                delim=myConfig.getString("debugDelimiter", DEF_DELIM);
                String className=myConfig.getString("debugOutputClass");
                myOut=(DebugOutput)(Class.forName(className).newInstance());
                myOut.setConfiguration(topLevel);                
                useErr=myConfig.getBoolean("useErr", CONF_DEF_USE_ERR);
                _startUpFiltering(myConfig);
            }
        }
        catch (Exception e) {
            e.printStackTrace();
        }
    }
    
    private static void _startUpFiltering(Configuration myConfig) throws Exception {
        
        filterNormal=myConfig.getBoolean("filterNormal", false);
        filterExceptions=myConfig.getBoolean("filterExceptions", false);
        
        if (filterNormal) {
            String filterOutNormal=myConfig.getString("filterOutNormal");
            fn_fullMatch=new Vector();
            fn_partialMatch=new Vector();
            __startUpFiltering(filterOutNormal, fn_fullMatch, fn_partialMatch);
            a_fn_fullMatch=fn_fullMatch.toArray();
            a_fn_partialMatch=fn_partialMatch.toArray();
        }
        
        if (filterExceptions) {
            String filterOutExceptions=myConfig.getString("filterOutExceptions");
            fe_fullMatch=new Vector();
            fe_partialMatch=new Vector();
            __startUpFiltering(filterOutExceptions, fe_fullMatch, fe_partialMatch);
            a_fe_fullMatch=fe_fullMatch.toArray();
            a_fe_partialMatch=fe_partialMatch.toArray();
        }
    }
    
    private static void __startUpFiltering(String items, Vector full, Vector partial) throws Exception {
        
        StringTokenizer tok=new StringTokenizer(items, ",");
        String curToken;
        while (tok.hasMoreTokens()) {
            curToken=tok.nextToken();
            if (curToken.charAt(curToken.length()-1)=='*') {
                partial.addElement(curToken.substring(0,curToken.length()-1));
            }
            else {
                full.addElement(curToken);
            }
        }
    }

    // ------------------------------------------------------------------
    // Public methods
    // ------------------------------------------------------------------
       
    /** Logs an Exception.
     *  @param String pref: the debug requester name
     *  @param Exception e: the Exception to be logged 
     */
    public static void logException(String pref, Exception e) {
        
        if (!doDebug) {
            return;
        }
        
        if (filterExceptions) {
            if (_filter(pref, a_fe_fullMatch, a_fe_partialMatch)) {
                return;
            }
        }
        
        if (fullFeature) {
            _logException(pref, e, "");
        }
        else {
            StringOutputStream sos=new StringOutputStream();
            e.printStackTrace(new PrintStream(sos));
            __l(pref+delim+sos.getString());
        }
    }
    
    /** Logs a message to the standard out.
     *  @param String pref: the debug requested name
     *  @param String msg: the message to be logged
     */
    public static void logOut(String pref, String s) {
        
        if (!doDebug) {
            return;
        }
        
        if (filterNormal) {
            if (_filter(pref, a_fn_fullMatch, a_fn_partialMatch)) {
                return;
            }
        }

        if (myOut!=null) {
            myOut.logOut(pref+delim+s);
        }
    }
    
    /** Logs a message to the standard err.
     *  @param String pref: the debug requested name
     *  @param String msg: the message to be logged
     */
    public static void logErr(String pref, String s) {
        
        if (!doDebug) {
            return;
        }
        
        if (filterNormal) {
            if (_filter(pref, a_fn_fullMatch, a_fn_partialMatch)) {
                return;
            }
        }
        
        if (myOut!=null) {
            __l(pref+delim+s);
        }
    }
        
    // ------------------------------------------------------------------
    // Private methods
    // ------------------------------------------------------------------
        
    private static void _logException(String pref, Exception e, String indent) {
        StackTraceElement ste[]=e.getStackTrace();        
        __l(pref+delim+indent+"!! Exception occured: "+e.getClass().getName()+" - "+e.toString());
        for (int i=0; i<ste.length; i++) {            
            _l(pref, ste[i], i, indent);
        }
        if (EHC.isInstance(e)) {
            Exception ee=((ExceptionHolder)e).getException();
            if (ee!=null) {
                __l(pref+delim+indent+"Containing a new Exception ...");
                _logException(pref, ee, indent+"  ");
            }
        }        
    }
    
    private static void _l(String pref, StackTraceElement ste, int pos, String indent) {        
        if (pos==0) {
            __l(pref+delim+indent+"Occured in "+ste.getClassName()+" in method "+ste.getMethodName()+"["+ste.getLineNumber()+"]");
        }
        else {
            __l(pref+delim+indent+"... from "+ste.getClassName()+" in method "+ste.getMethodName()+"["+ste.getLineNumber()+"]");
        }
    }
    
    private static void __l(String s) {
        if (myOut!=null) {
            if (useErr) {            
                myOut.logErr(s);
            }
            else {
                myOut.logOut(s);
            }
        }
    }
    
    private static boolean _filter(String s, Object[] full, Object[] partial) {
        
        String ss;
        int i;
        
        // full match
        for (i=0; i<full.length; i++) {
            ss=(String)(full[i]);
            if (ss.equals(s)) {
                return true;
            }
        }
                
        // partial match
        
        for (i=0; i<partial.length; i++) {
            ss=(String)(partial[i]);
            if (s.length()>ss.length()) {
                if (s.substring(0,ss.length()).equals(ss)) {
                    return true;
                }
            }
        }
        
        // no match
        
        return false;
    }
}
